/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://ayyi.org               |
* | copyright (C) 2013-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#ifndef __ayyi_debug_h__
#define __ayyi_debug_h__

#include "glib.h"

#ifdef DEBUG
// note that the level check is now outside the print fn in order to
// prevent functions that return print arguments from being run
#  define dbg(A, B, ...) do{if(A <= _debug_) debug_printf(__func__, A, B, ##__VA_ARGS__);} while(FALSE)

#  define PF {if(_debug_) printf("%s()...\n", __func__);}
#  define PF0 {printf("%s()...\n", __func__);}
#else
#  define dbg(A, B, ...)
#  define PF
#  define PF0
#endif
#define gerr(A, ...) g_critical("%s(): "A, __func__, ##__VA_ARGS__)
#define gwarn(A, ...) g_warning("%s(): "A, __func__, ##__VA_ARGS__)

void        debug_printf             (const char* func, int level, const char* format, ...);
void        warnprintf               (char* format, ...);
void        errprintf                (char* fmt, ...);

void        log_handler              (const gchar* log_domain, GLogLevelFlags, const gchar* message, gpointer);

#ifdef __debug_c__
char ayyi_white [16] = "\x1b[0;39m";
char ayyi_grey  [16] = "\x1b[38;5;240m";
char ayyi_warn  [32] = "\x1b[1;33mwarning:\x1b[0;39m";
char ayyi_err   [32] = "\x1b[1;31merror!\x1b[0;39m";
#else
extern int _debug_;                  // debug level. 0=off.

extern char ayyi_white [16];
extern char ayyi_grey  [16];
extern char ayyi_warn  [32];
extern char ayyi_err   [32];
#endif

#endif //__ayyi_debug_h__
