/*
  This file is part of the Ayyi Project. http://ayyi.org
  copyright (C) 2004-2017 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#ifndef __am_collection_h__
#define __am_collection_h__

#include <glib.h>
#include <glib-object.h>
#include <model/ayyi_model.h>

G_BEGIN_DECLS

#define MODEL_TYPE_COLLECTION (model_collection_get_type ())
#define MODEL_COLLECTION(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), MODEL_TYPE_COLLECTION, ModelCollection))
#define MODEL_COLLECTION_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), MODEL_TYPE_COLLECTION, ModelCollectionClass))
#define MODEL_IS_COLLECTION(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), MODEL_TYPE_COLLECTION))
#define MODEL_IS_COLLECTION_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), MODEL_TYPE_COLLECTION))
#define MODEL_COLLECTION_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), MODEL_TYPE_COLLECTION, ModelCollectionClass))

typedef struct _ModelCollection ModelCollection;
typedef struct _ModelCollectionClass ModelCollectionClass;
typedef struct _ModelCollectionPrivate ModelCollectionPrivate;

#define AM_TYPE_ARRAY (am_array_get_type ())
#define AM_ARRAY(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), AM_TYPE_ARRAY, AMArray))
#define AM_ARRAY_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), AM_TYPE_ARRAY, AMArrayClass))
#define AM_IS_ARRAY(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), AM_TYPE_ARRAY))
#define AM_IS_ARRAY_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), AM_TYPE_ARRAY))
#define AM_ARRAY_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), AM_TYPE_ARRAY, AMArrayClass))

typedef struct _AMArrayClass AMArrayClass;
typedef struct _AMArrayPrivate AMArrayPrivate;

#define AM_TYPE_LIST (am_list_get_type ())
#define AM_LIST(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), AM_TYPE_LIST, AMList))
#define AM_LIST_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), AM_TYPE_LIST, AMListClass))
#define AM_IS_LIST(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), AM_TYPE_LIST))
#define AM_IS_LIST_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), AM_TYPE_LIST))
#define AM_LIST_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), AM_TYPE_LIST, AMListClass))

typedef struct _AMListClass AMListClass;
typedef struct _AMListPrivate AMListPrivate;

#define AM_TYPE_TRACK_LIST (am_track_list_get_type ())
#define AM_TRACK_LIST(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), AM_TYPE_TRACK_LIST, AMTrackList))
#define AM_TRACK_LIST_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), AM_TYPE_TRACK_LIST, AMTrackListClass))
#define AM_IS_TRACK_LIST(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), AM_TYPE_TRACK_LIST))
#define AM_IS_TRACK_LIST_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), AM_TYPE_TRACK_LIST))
#define AM_TRACK_LIST_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), AM_TYPE_TRACK_LIST, AMTrackListClass))

typedef struct _AMTrackListClass AMTrackListClass;
typedef struct _AMTrackListPrivate AMTrackListPrivate;

#define TYPE_FILTER_ITERATOR (filter_iterator_get_type ())
#define FILTER_ITERATOR(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), TYPE_FILTER_ITERATOR, FilterIterator))
#define FILTER_ITERATOR_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), TYPE_FILTER_ITERATOR, FilterIteratorClass))
#define IS_FILTER_ITERATOR(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), TYPE_FILTER_ITERATOR))
#define IS_FILTER_ITERATOR_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), TYPE_FILTER_ITERATOR))
#define FILTER_ITERATOR_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), TYPE_FILTER_ITERATOR, FilterIteratorClass))

typedef struct _FilterIteratorClass FilterIteratorClass;
typedef struct _FilterIteratorPrivate FilterIteratorPrivate;

struct _ModelCollection {
	GObject                 parent_instance;
	ModelCollectionPrivate* priv;
	GList*                  selection;
	GList*                  pending;
	void*                   user_data;
};

typedef void* (*ModelCollectionNextFn) (ModelCollection*, struct _am_iter*);

struct _ModelCollectionClass {
	GObjectClass parent_class;
	guint        (*length)     (ModelCollection*);
	gboolean     (*contains)   (ModelCollection*, gconstpointer item);
	gpointer     (*find_by_id) (ModelCollection*, guint64 id);
	void         (*append)     (ModelCollection*, gconstpointer item);
	void         (*remove)     (ModelCollection*, gconstpointer item);
	void         (*iter_init)  (ModelCollection*, AMIter*);
	//void* (*iter_next) (ModelCollection* self, struct _am_iter* iter);
	ModelCollectionNextFn iter_next;
};

struct _AMArray {
	ModelCollection parent_instance;
	AMArrayPrivate* priv;
	GPtrArray*      array;
};

struct _AMArrayClass {
	ModelCollectionClass parent_class;
};

struct _AMList {
	ModelCollection parent_instance;
	AMListPrivate*  priv;
	GList*          list;
};

struct _AMListClass {
	ModelCollectionClass parent_class;
};

struct _AMTrackList {
	ModelCollection     parent_instance;
	AMTrackListPrivate* priv;
	AMTrack*            track[48];
	gint                last;
};

struct _AMTrackListClass {
	ModelCollectionClass parent_class;
};

typedef gboolean (*AMTrackListTrackEach) (AMTrack* track, void* user_data);

struct _FilterIterator {
    GTypeInstance          parent_instance;
    volatile int           ref_count;
    FilterIteratorPrivate* priv;
    ModelCollection*       list;
    struct _am_iter        iter;
    Filter                 accept;
    gpointer               accept_target;
    GDestroyNotify         accept_target_destroy_notify;
};

struct _FilterIteratorClass {
	GTypeClass parent_class;
	void (*finalize) (FilterIterator *self);
};


GType            model_collection_get_type          (void) G_GNUC_CONST;
guint            model_collection_length            (ModelCollection*);
gboolean         model_collection_contains          (ModelCollection*, gconstpointer item);
gpointer         model_collection_find_by_id        (ModelCollection*, guint64 id);
void             model_collection_append            (ModelCollection*, gconstpointer item);
void             model_collection_remove            (ModelCollection*, gconstpointer item);
void             model_collection_iter_init         (ModelCollection*, AMIter*);
void*            model_collection_iter_next         (ModelCollection*, AMIter*);
void             model_collection_selection_reset   (ModelCollection*, void* sender);
void             model_collection_selection_replace (ModelCollection*, GList*, void* sender);
void             model_collection_selection_add     (ModelCollection*, GList*, void* sender);
void             model_collection_selection_remove  (ModelCollection*, gconstpointer item, void* sender);
gpointer         model_collection_selection_first   (ModelCollection*);
ModelCollection* model_collection_construct         (GType object_type, GType t_type, GBoxedCopyFunc, GDestroyNotify);

GType            am_array_get_type                  (void) G_GNUC_CONST;
AMArray*         am_array_new                       (GType t_type, GBoxedCopyFunc, GDestroyNotify, gint n);
AMArray*         am_array_construct                 (GType object_type, GType t_type, GBoxedCopyFunc, GDestroyNotify, gint n);
gconstpointer*   am_array_at                        (AMArray*, gint index);

GType            am_list_get_type                   (void) G_GNUC_CONST;
AMList*          am_list_new                        (GBoxedCopyFunc, GDestroyNotify);
gpointer         am_list_first                      (AMList*);
gpointer         am_list_find_by_ident              (AMList*, AyyiIdent);
gpointer         am_list_find_by_idx                (AMList*, gint idx);
void             am_list_unpend                     (AMList*, gconstpointer item);
gboolean         am_list_is_pending                 (AMList*, AyyiIdx, AyyiMediaType);

AMTrackList*     am_track_list_new                  (GType, GBoxedCopyFunc, GDestroyNotify);
AMTrack*         am_track_list_at                   (AMTrackList*, gint index);
gint             am_track_list_count                (AMTrackList*);
gint             am_track_list_get_visible_count    (AMTrackList*);
AMTrack*         am_track_list_find_by_shm_idx      (AMTrackList*, AyyiIdx, AMTrackType);
AMTrack*         am_track_list_find_by_ident        (AMTrackList*, AyyiIdent);
AMTrack*         am_track_list_find_audio_by_idx    (AMTrackList*, AyyiIdx);
AMTrack*         am_track_list_find_by_name         (AMTrackList*, gchar* name);
gboolean         am_track_list_is_known             (AMTrackList*, AMTrackType, AyyiIdx);
AMTrack*         am_track_list_get_first_visible    (AMTrackList*, AMTrackType);
void             am_track_list_each                 (AMTrackList*, AMTrackListTrackEach fn, void*);
gboolean         am_track_list_verify_track         (AMTrackList*, AMTrack*);
gboolean         am_track_list_verify_tracks        (AMTrackList*);

FilterIterator*  filter_iterator_new                (GType t_type, GBoxedCopyFunc, GDestroyNotify, ModelCollection* _list, Filter, void*);
gpointer         filter_iterator_ref                (gpointer);
void             filter_iterator_unref              (gpointer);
GType            filter_iterator_get_type           (void) G_GNUC_CONST;
FilterIterator*  filter_iterator_construct          (GType object_type, GType t_type, GBoxedCopyFunc, GDestroyNotify, ModelCollection*, Filter, void*);
gconstpointer*   filter_iterator_next               (FilterIterator*);
GParamSpec*      param_spec_filter_iterator         (const gchar* name, const gchar* nick, const gchar* blurb, GType, GParamFlags);
void             value_set_filter_iterator          (GValue*, gpointer v_object);
void             value_take_filter_iterator         (GValue*, gpointer v_object);
gpointer         value_get_filter_iterator          (const GValue*);


G_END_DECLS

#endif
