/*
  This file is part of the Ayyi Project. http://www.ayyi.org
  copyright (C) 2004-2014 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef __model_pool_h__
#define __model_pool_h__
#include <model/am_collection.h>
#include <model/am_pool_item.h>

#define POOL_ITEM_IS_OK(A) (A && A->state == POOL_ITEM_OK)
#define REGION_FULL 0x0fffffff // a magic region_index number indicating a region representing the whole audio file <-- warning! danger!
#define am_pool__iter_init(I) model_collection_iter_init((ModelCollection*)song->pool, I);
#define am_pool__emit(...) ({ if(song->loaded) g_signal_emit_by_name (song->pool, __VA_ARGS__); })
#define am_pool ((ModelCollection*)song->pool)

#ifdef __am_private__
void        am_pool__init                     ();
void        am_pool__free                     ();
void        am_pool__sync                     ();
void        am_pool__add                      (PoolItem*);
PoolItem*   am_pool__add_from_file            (const char* fname);
#endif
PoolItem*   am_pool__add_ayyi_file            (AyyiFilesource*);
bool        am_pool__remove                   (PoolItem*);
void        am_pool__clear                    ();
bool        am_pool__item_exists              (char* fname);
#ifdef __am_private__
PoolItem*   am_pool__find_pending_by_leafname (const char*);
#endif
PoolItem*   am_pool__get_item_from_idx        (AyyiIdx);
PoolItem*   am_pool__get_item_from_leafname   (const char*);
PoolItem*   am_pool__get_item_by_orig_name    (const char*);
GList*      am_pool__get_new_items            ();
PoolItem*   am_pool__get_item_by_id           (AyyiId);

void        am_pool__print                    ();

#endif
