/*
  This file is part of the Ayyi Project. http://www.ayyi.org
  copyright (C) 2004-2017 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef __pool_item_h__
#define __pool_item_h__

#include "model/model_types.h"
#include "waveform/waveform.h"

G_BEGIN_DECLS

#define AM_TYPE_POOL_ITEM (am_pool_item_get_type ())
#define AM_POOL_ITEM(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), AM_TYPE_POOL_ITEM, AMPoolItem))
#define AM_POOL_ITEM_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), AM_TYPE_POOL_ITEM, AMPoolItemClass))
#define AM_IS_POOL_ITEM(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), AM_TYPE_POOL_ITEM))
#define AM_IS_POOL_ITEM_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), AM_TYPE_POOL_ITEM))
#define AM_POOL_ITEM_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), AM_TYPE_POOL_ITEM, AMPoolItemClass))

typedef struct _AMPoolItem      AMPoolItem;
typedef struct _AMPoolItemClass AMPoolItemClass;

struct _AMPoolItem
{
	Waveform           waveform;

	AyyiIdx            pod_index[2];    // -1 if not set.
	guint64            source_id[2];

	char               filename[AYYI_FILENAME_MAX]; // this should be the exact name in the songcore.
	char               leafname[256];   // the filename without directory info.
	int                state;

	int                channels;        // channels can be the combined width of more than one file.

	bool               peakfile_valid;

	void*              gui_ref;         // eg GtkTreeRowReference*
};

struct _AMPoolItemClass {
	WaveformClass parent_class;
};

enum
{
   POOL_ITEM_NEW = 0,
   POOL_ITEM_PENDING,   //waiting for remote confirmation
   POOL_ITEM_BAD,
   POOL_ITEM_INACCESSIBLE,
   POOL_ITEM_OK,
   POOL_ITEM_STATE_MAX
};


PoolItem*            am_pool_item_new              ();
PoolItem*            pool_item_new_from_ayyi       (AyyiFilesource*);
#ifdef __am_private__
void                 pool_item_set_from_ayyi       (PoolItem*, AyyiFilesource*);
#endif
bool                 pool_item_set_file            (PoolItem*, const char* fname, bool fast);
void                 pool_item_add_channel         (PoolItem*, AyyiFilesource*);
bool                 pool_item_file_exists         (PoolItem*);
GList*               pool_item_get_regions         (PoolItem*);
int                  pool_item_count_regions       (PoolItem*);
bool                 pool_item_get_default_region  (PoolItem*, uint32_t* region_idx);
bool                 pool_item_is_raw              (PoolItem*);
int                  pool_item_get_width           (PoolItem*);
char*                pool_item_get_full_path       (PoolItem*, int ch_num);
void                 pool_item_get_display_name    (PoolItem*, char* name);
bool                 pool_item_in_core             (PoolItem*);
void                 pool_item_load_peak           (PoolItem*, int ch_num);
void                 pool_item_print               (PoolItem*);
const char*          pool_item_print_state         (PoolItem*);

bool                 pool_item_is_nonempty         (PoolItem*, gpointer);
bool                 pool_item_is_new              (PoolItem*, gpointer);

G_END_DECLS

#endif
