/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#ifndef __am_song_h__
#define __am_song_h__

#define STOP_ON_UNLOAD // restart the server for each song load

#include "config.h"
#include "ayyi/ayyi_client.h"
#include "ayyi/ayyi_song.h"
#include "model/am_controller.h"
#include "model/am_palette.h"

ClientController* song;


#ifdef __am_private__
void         am_song__add_handlers                ();
#endif
gboolean     am__register_signals                 (AyyiService*);

void         am_song__new                         ();
void         am_song__free                        ();
void         am_song__load                        ();                     // syncs the local song to that on the server.
void         am_song__unload                      ();                     // remove local objects in preparation for loading. probably needs refactoring.

void         am_song__load_new                    (const char* filename, AyyiHandler, gpointer); // requests the server to close the current song and load the one given.
void         am_song__reload                      (AyyiHandler, gpointer);
void         am_song__save                        (AyyiHandler, gpointer);

void         am_song__undo                        (AyyiService*, int n);
void         am_song__enable_meters               (gboolean);

/* Tracks */
void         am_song__add_track                   (AMTrackType, char* name, int n_channels, AyyiHandler, gpointer);
void         am_song__remove_track                (AMTrack*, AyyiHandler, gpointer);
void         am_song__track_select_first          ();
AMTrack*     am_song__get_track_by_playlist       (AyyiPlaylist*, AyyiMediaType);
AMTrack*     am_song__get_track_for_part          (AyyiRegionBase*, AyyiMediaType);
#ifdef __am_private__
AMTrack*    _am_song__add_track                   (AMTrackType, AyyiIdx);
#endif

/* Parts */
AMPart*      am_song__get_part_by_region_index    (AyyiIdx, AyyiMediaType);
AMPart*      am_song__get_part_by_ident           (AyyiIdent);
AMPart*      am_song__part_lookup_by_name         (const char* name);
void         am_song__part_selection_move         (AyyiSongPos*, gboolean forwards, int track_offset);
void         am_song__add_part                    (AyyiMediaType, AyyiIdx src_region, PoolItem*, AyyiIdx track, AyyiSongPos*, uint64_t len, unsigned int colour, char* name, uint32_t inset, AyyiHandler, gpointer);
AyyiAction*  am_song__remove_part                 (AMPart*, AyyiHandler, gpointer);
void         am_song__remove_all_parts            (AyyiHandler, gpointer);
void         am_song__remove_parts_by_track       (AMTrack*, AyyiHandler, gpointer);
void         am_song__remove_parts_by_file        (PoolItem*, AyyiHandler, gpointer);
#ifdef __am_private__
void        _am_song__remove_part                 (AMPart*);
void        _am_song__remove_all_parts            ();
#endif

/* Files */
void         am_song__add_pool_item               (PoolItem*, AyyiHandler, gpointer);
PoolItem*    am_song__add_file                    (const char* filename, AyyiHandler, gpointer);
void         am_song__remove_file                 (PoolItem*, AyyiHandler, gpointer);
PoolItem*    am_song__get_pool_item_from_region   (AyyiIdx);
PoolItem*    am_song__get_pool_item_from_region_without_id(AyyiIdx);


double       am_song__get_tempo                   ();
void         am_song__set_tempo                   (double tempo, AyyiHandler, gpointer);

void         am_song__get_end                     (GPos*);
void         am_song__set_end                     (int);


void         am_song__midi_note_add               (AMPart*, MidiNote*, AyyiHandler, gpointer);

#ifdef __am_private__
void         am_song__sync_tracks                 ();
void         am_song__sync_parts                  ();
#endif

void         am_song__clear_songmap               ();

void         am_song__print_tracks                ();
void         am_song__channels_print              ();

void         am_song__queue_work                  (AyyiCallback, gpointer);

#define am_song__emit(...) ({ if(song->loaded) g_signal_emit_by_name (song, __VA_ARGS__); })
#define am_song__connect(...) { g_signal_connect (song, __VA_ARGS__); }
#define am_song__disconnect(PANEL, FN) {int n; if((n = g_signal_handlers_disconnect_matched (song, G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, 0, 0, NULL, FN, PANEL)) != 1) gwarn("handler disconnection: %s n=%i", #FN, n); }

#endif //__am_song_h__
