/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://ayyi.org               |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#define __am_track_c__
#define __am_track__peer_c__
#define __am_private__
#include <model/common.h>
#include <gdk/gdkkeysyms.h>
#include <ayyi/ayyi_list.h>
#include <ayyi/ayyi_shm.h>
#include <ayyi/ayyi_client.h>
#include <ayyi/ayyi_song.h>
#include <ayyi/ayyi_action.h>

#include <model/am_song.h>
#include <model/am_curve.h>
#include <model/am_channel.h>
#include <model/am_track_peer.h>


/*
 *  Should only ever be called once.
 */
void
am_tracks_init()
{
	if(track_pool) gwarn("already initialised! use am_tracks_reset() instead.");

	track_pool = g_malloc0(sizeof(AMTrack) * AM_MAX_TRK);

	int t;
	for(t=0;t<AM_MAX_TRK;t++){
#if 0
		song->tracks->track[t] = track_pool_get_item(t);
#endif
		am_track__init(t);
	}
}


void
am_tracks_reset()
{
	g_return_if_fail(track_pool);

	int t;
	for(t=0;t<AM_MAX_TRK;t++){
		song->tracks->track[t] = track_pool_get_item(t);
		am_track_reset(t);
	}
}


/*
 *  Returns an internal track index for the first unused track.
 */
TrackNum
_am_tracks_find_unused()
{
	unsigned t;
	for(t=0;t<AM_MAX_TRK;t++){
		dbg (2, "t=%i track=%p type=%i", t, song->tracks->track[t], song->tracks->track[t] ? song->tracks->track[t]->type : 0);
		if(!song->tracks->track[t] || song->tracks->track[t]->type == TRK_TYPE_NULL) return t;
	}
	return -1;
}


void
am_tracks_rebuild_pointers()
{
	// iterating over track[] is easier if the array isnt sparse, so it needs rebuilding after a delete.

	// this means that pointers to tracks do not change, but the track number does.
	// ie the track_num cannot be used as a reliable reference

	int i = 0;
	int t; for(t=0;t<AM_MAX_TRK;t++){
		AMTrack* tt = track_pool_get_item(t);
		if(tt && tt->type){
			dbg(3, "i=%i t=%i ->t=%i %s", i, t, tt->track_num, tt->name);
			song->tracks->track[i] = tt;
			tt->track_num = i;
			i++;
		}
		else if(tt) dbg(3, "    t=%i ! %s type=%i", t, tt->name, tt->type);
	}
	song->tracks->track[i] = NULL;
}



