/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://ayyi.org               |
* | copyright (C) 2007-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include "stdint.h"

#define mono_wav_fixture "../../libs/waveform/test/data/mono_0:10.wav"
#define mono_wav_fixture2 "../../libs/waveform/test/data/mono_10:00.wav"
#define stereo_wav_fixture "../../libs/waveform/test/data/stereo_0:10.wav"
#define vst_plugin_fixture "SaffireEQMono"
#define MIN_TRACKS 1 //server crashes if no tracks?

struct _app
{
	gboolean       dbus;
	int            timeout;
	int            n_passed;
} app;

typedef void (*Test)    ();
typedef void (TestFn)   ();

void am_test_init       (gpointer tests[], int);
void app_on_all_shm     (GError*, gpointer);
void next_test          ();
void test_finished_     ();

bool get_random_boolean ();
int  get_random_int     (int max);

void errprintf4         (char* format, ...);

void am_wait_for        (gboolean (*)(gpointer), AyyiHandler, gpointer);
void reset_timeout      (int ms);

#include "test/modules.h"

#ifndef __am_common_c__
extern int       current_test;
extern int       test_finished;
extern gpointer* current_tests;
extern bool      passed;
extern int       n_failed;
extern GMainLoop* main_loop;
#endif

#define START_TEST \
	static int step = 0;\
	if(!step){ \
		g_strlcpy(current_test_name, __func__, 64); \
		log_print(0, "running %i of %i: %s ...", current_test + 1, G_N_ELEMENTS(tests), __func__); \
	} \
	if(test_finished) return;

#define FINISH_TEST \
	dbg(0, "finish"); \
	test_finished = true; \
	passed = true; \
	test_finished_(); \
	return;

#define FAIL_TEST(msg, ...) \
	{if(!test_finished) n_failed++; else gwarn("test already finished"); \
	test_finished = true; \
	passed = false; \
	errprintf4(msg, ##__VA_ARGS__); \
	test_finished_(); \
	return; }

#define FAIL_TEST_TIMER(msg) \
	{test_finished = true; \
	passed = false; \
	printf("%s%s%s\n", red, msg, white); \
	test_finished_(); \
	return TIMER_STOP;}

#define assert(A, B, ...) \
	{bool __ok_ = (boolp)A; \
	{if(!__ok_) gerr(B, ##__VA_ARGS__); } \
	{if(!__ok_) FAIL_TEST("assertion failed") }}

