/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include "global.h"
#include "arrange.h"
#ifdef USE_GNOMECANVAS
#include "gnome/automation.h"
#endif
#ifdef USE_OPENGL
#include "gl_canvas.h"
#endif


void
automation_view_vol(ArrTrk* atr)
{
	// Bring volume automation curve to the front. If it is empty, add a start and end point.

	Arrange* arrange = atr->arrange;

	Curve* curve = song->tracks->track[atr->track->track_num]->bezier.vol;
	dbg (0, "t=%i", atr->track->track_num);
	if(!curve->path){
		dbg (0, "no path. creating new default...");
		am_automation_make_empty(curve);
#ifdef USE_GNOMECANVAS
		automation_curve_draw(arrange, curve);
#endif
	}

#ifdef USE_GNOMECANVAS
	if(CANVAS_IS_GNOME){
		g_automation_view_vol(atr);
	}
#endif
#ifdef USE_OPENGL
	if(CANVAS_IS_OPENGL){
		arr_gl_canvas_redraw_track(arrange, atr);
	}
#endif
}


void
automation_view_pan(ArrTrk* atr)
{
	Arrange* arrange = atr->arrange;
	Curve* curve = song->tracks->track[atr->track->track_num]->bezier.pan;
	dbg(0, "arr_trk=%p t=%i", atr, atr->track->track_num);

	if(!curve->path){
		dbg(0, "not yet created...");
		am_automation_make_empty(curve);
#ifdef USE_GNOMECANVAS
		automation_curve_draw(arrange, curve);
#endif
	}

#ifdef USE_GNOMECANVAS
	if(CANVAS_IS_GNOME){
		g_automation_view_pan(atr);
	}
#endif
#ifdef USE_OPENGL
	if(CANVAS_IS_OPENGL){
		arr_gl_canvas_redraw_track(arrange, atr);
	}
#endif
}
