/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

/*
 * Currently using straight lines.
 *
 * Although curves are more useful, they are not supported by the mixer
 * so need to be rendered first. They would need to be stored independently
 * by the gui and hence would be lost in a multi-client situation.
 */
#undef ENABLE_CURVES

/*
 *  GL_POINT_SMOOTH, GL_LINE_SMOOTH, GL_POLYGON_SMOOTH not recomended as not hardware supported, so will cause rendering to be done in software.
 */
#define ENABLE_ANTIALIASING \
		glEnable(GL_LINE_SMOOTH); \
		agl_enable(AGL_ENABLE_BLEND); \
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); \
		glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
#define DISABLE_ANTIALIASING \
		glDisable(GL_LINE_SMOOTH); \
		agl_enable(!AGL_ENABLE_BLEND);

typedef struct {GLfloat x, y, z;} Pt3f;
typedef struct {Pt3f pt1, ctl1, ctl2, pt2;} Bp;

static AGlActorClass automation_actor_class = {0, "Automation", (AGlActorNew*)arr_gl_automation};

static void _draw_aa_line(float x1, float y1, float x2, float y2, float r, float g, float b, float a, float xscale);


static AGlActor*
arr_gl_automation(GtkWidget* panel)
{
	bool arr_gl_draw_automation(AGlActor* actor)
	{
	#ifdef GL_AUTOMATION
		Arrange* arrange = actor->root->user_data;

		//GLfloat knots[8] = {0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0};

		float xscale = arr_px2samples(arrange, 1);

		// take 1st volume automation seg
		Curve* cv = NULL;
		AMTrack* trk = NULL;
		ArrTrk* at;
		while((at = track_list__next_visible(arrange->priv->tracks, trk))){
			trk = at->track;
			cv = trk->bezier.vol;
			if(cv && cv->path){
#ifdef ENABLE_CURVES
				agl_enable(0 /* !AGL_ENABLE_TEXTURE_2D */);
#else
				agl_enable(AGL_ENABLE_TEXTURE_2D | AGL_ENABLE_BLEND);
				glBindTexture(GL_TEXTURE_2D, agl->aaline.texture);
				agl->shaders.text->uniform.fg_colour = 0xff0000ff;
				agl_use_material(&agl->aaline);

#endif
				break;
			}
		}

		trk = NULL;
		while((at = track_list__next_visible(arrange->priv->tracks, trk))){
			trk = at->track;
			cv = trk->bezier.vol;
			if(!(cv && cv->path)) continue;

			BPath* path = cv->path;
			dbg(3, "t=%i path=%p path1=%p", trk->track_num, path, &path[1]);

			//ENABLE_ANTIALIASING
			glPushMatrix();
			glTranslatef(0, at->y, 0.0);
			glScalef (arr_gl_samples2px(arrange, 1), at->height * arrange->vzoom / 100.0, 1.0);

			iRange xrange = {0, arr_px2samples(arrange, actor->region.x2)}; // TODO scrolling

			BPath* p = (BPath*)&path[0];

			Bp cpts = (Bp){
				.pt1 = {
					.x = p->x3 - 2.0,
					.y = p->y3,
					.z = 0
				}
			};
			BPath first_prev_path = {CURVETO, 0.0, 0.0, 0.0, 0.0, p->x3 - 20.0, p->y3};
			int i; for(i=0;i<1000 && p->code != BP_END;i++){
				p = (BPath*)&path[i];
				//BPath* next = (BPath*)&path[i+1];
				//BPath* prev = i ? (BPath*)&path[i-2] : &first_prev_path;
				if(p->code == CURVETO) gwarn("curve not supported! straight line expected.");
				if((p->x3 >= xrange.start) && (cpts.pt1.x <= xrange.end)){ // curve segment is visible.

					cpts.pt2 = (Pt3f){
						.x = p->x3,
						.y = p->y3,
						.z = 0
					};

#ifdef ENABLE_CURVES
					// find the average gradient before and after pt 1
					float x0 = prev->x3;
					float x1 = cpts.pt1.x;
					float x2 = cpts.pt2.x;
					float gr1 = (cpts.pt1.y - prev->y3) / (x1 - x0);
					float gr2 = (cpts.pt2.y - cpts.pt1.y) / (x2 - x1);
					float gr = (gr1 + gr2) / 2;
					dbg(2, "%i: gradient1: a=%.2f b=%.2f --> %.2f", i, gr1, gr2, gr);
					float offset_r_x = MIN(20.0, (x2 - x0) / 8);
					if(i < 2) offset_r_x = 0; //FIXME
					float offset_r_y = gr * offset_r_x;

					cpts.ctl1 = (Pt3f){
						.x = cpts.pt1.x + offset_r_x,
						.y = cpts.pt1.y + offset_r_y,
						.z = 0
					};

					// find the average gradient before and after pt 2
					gr1 = gr2;
					gr2 = (next->y3 - cpts.pt2.y) / (next->x3 - cpts.pt2.x);
					gr = (gr1 + gr2) / 2;
					dbg(2, "%i: gradient2=%.2f\n", i, gr);
					offset_r_x *= -1;
					offset_r_y = gr * offset_r_x;

					cpts.ctl2 = (Pt3f){
						.x = cpts.pt2.x + offset_r_x,
						.y = cpts.pt2.y + offset_r_y,
						.z = 0
					};
					//dbg(0, "%.2f %.2f %.2f %.2f", cpts.pt1.x, cpts.pt1.y, cpts.pt2.x, cpts.pt2.y);

					/*
					gluBeginSurface(theNurb);
					gluNurbsSurface(theNurb, 
										8, knots, 8, knots,
										4 * 3, 3, &ctlpoints[0][0][0], 
										4, 4, GL_MAP2_VERTEX_3);
					gluEndSurface(theNurb);
					*/

					gluBeginCurve(theNurb);
					gluNurbsCurve(theNurb, 
										8, knots, // will render 1 bezier section only (4 points).
										3, (GLfloat*)&cpts, 
										4, GL_MAP1_VERTEX_3);
					gluEndCurve(theNurb);
#else
					BPath* pp = i ? (BPath*)&path[i-1] : &first_prev_path;
#if 0

					glBegin(GL_LINES);
					glVertex2f(pp->x3, pp->y3);
					glVertex2f(p->x3, p->y3);
					glEnd();
#else
					float y = at->y * arrange->vzoom;
					_draw_aa_line(pp->x3, y + pp->y3, p->x3, y + p->y3, 1.0, 0.0, 0.0, 1.0, xscale);
#endif
#endif
				}
				// start of next segment becomes the end of this one
				cpts.pt1.x  = p->x3;
				cpts.pt1.y  = p->y3;

				// TODO dont show points that are not visible
				if (1/*showPoints*/) {
					glPointSize(4.0);
					glColor3f(1.0, 1.0, 0.0);
					/*
					glBegin(GL_POINTS);
					int i, j; for (i = 0; i < 4; i++) {
						for (j = 0; j < 4; j++) {
							glVertex3f(ctlpoints[i][j][0], ctlpoints[i][j][1], ctlpoints[i][j][2]);
							//dbg(0, "%.2f %.2f %.2f", ctlpoints[i][j][0], ctlpoints[i][j][1], ctlpoints[i][j][2]);
						}
					}
					glEnd();
					*/

					glBegin(GL_POINTS);
					/*
					int i; for (i = 0; i < 4; i++) {
						glVertex3f(cpts[i][0], cpts[i][1], cpts[i][2]);
						dbg(0, "%.2f %.2f %.2f", ctlpoints[i][j][0], ctlpoints[i][j][1], ctlpoints[i][j][2]);
					}
					*/
					glVertex3f(cpts.pt1.x, cpts.pt1.y, cpts.pt1.z);
					if(0){
						glVertex3f(cpts.ctl2.x, cpts.ctl2.y, 0.0);
					}
					glEnd();
				}
			}
			//DISABLE_ANTIALIASING

			glPopMatrix();
		}
	#endif
		return true;
	}

	void automation_set_size(AGlActor* actor)
	{
		actor->region = (AGliRegion){
			.x1 = 0,
			.y1 = 0,
			.x2 = agl_actor__width(actor->parent),
			.y2 = agl_actor__height(actor->parent),
		};
	}

	void arr_gl_automation_init(AGlActor* actor)
	{
		// TODO make material init
		agl->aaline.shader = (AGlShader*)agl->shaders.text; // alpha shader
	}

	return AYYI_NEW(AGlActor,
		.init = arr_gl_automation_init,
		.paint = arr_gl_draw_automation,
		.set_size = automation_set_size,
		.name = automation_actor_class.name
	);
}


static void
_draw_aa_line(float x1, float y1, float x2, float y2, float r, float g, float b, float a, float xscale)
{
	// xscale is used to reverse the current GlScale setting

	glBegin(GL_QUADS);
	glTexCoord2d(0.0, 0.0); glVertex2f(x1, y1);
	glTexCoord2d(1.0, 0.0); glVertex2f(x2, y2);
	glTexCoord2d(1.0, 1.0); glVertex2f(x2 + xscale, y2 + 2.0);
	glTexCoord2d(0.0, 1.0); glVertex2f(x1 + xscale, y1 + 2.0);
	glEnd();
}
