/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2014-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

#undef BACKGROUND_2

typedef struct {
    AGlActor    actor;
} BackgroundLayer;

static GLuint bg_textures[TEXTURE_MAX + 1/* TODO */] = {0,};

#ifdef BG_TEXTURE
static void
arr_gl_create_bg_texture(AGlActor* actor)
{
	if(bg_textures[TEXTURE_BG]) return;
#ifdef BACKGROUND_2
	if(!pattern_shader.shader.program){
		agl_create_program(&pattern_shader.shader);
	}
#endif

	agl_enable(AGL_ENABLE_TEXTURE_2D);

	glGenTextures(TEXTURE_MAX, bg_textures);
	if(glGetError() != GL_NO_ERROR){ gerr ("failed to generate textures."); return; }
	dbg(2, "bg_texture=%u", bg_textures[TEXTURE_BG]);

	{
		// create an alpha-map gradient for use as background

		int width = 256;
		int height = 256;
		char* pbuf = g_new0(char, width * height);
		int y; for(y=0;y<height;y++){
			int x; for(x=0;x<width;x++){
				*(pbuf + y * width + x) = ((x+y) * 0xff) / (width * 2);
			}
		}

		int pixel_format = GL_ALPHA;
		glBindTexture  (GL_TEXTURE_2D, bg_textures[TEXTURE_BG]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexImage2D   (GL_TEXTURE_2D, 0, GL_ALPHA, width, height, 0, pixel_format, GL_UNSIGNED_BYTE, pbuf);
		gl_warn("bg texture bind");

		g_free(pbuf);
	}

	{
		// create button icons

		void create_icon(const char* name, int icon)
		{
			GdkPixbuf* pixbuf = get_icon(name, 16);
			if(pixbuf){
				dbg(2, "icon: pixbuf=%ix%i %ibytes/px", gdk_pixbuf_get_width(pixbuf), gdk_pixbuf_get_height(pixbuf), gdk_pixbuf_get_n_channels(pixbuf));
				glBindTexture   (GL_TEXTURE_2D, bg_textures[icon]);
				glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
				glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
				glTexImage2D    (GL_TEXTURE_2D, 0, GL_RGBA, gdk_pixbuf_get_width(pixbuf), gdk_pixbuf_get_height(pixbuf), 0, GL_RGBA, GL_UNSIGNED_BYTE, gdk_pixbuf_get_pixels(pixbuf));
				gl_warn("texture bind");
				g_object_unref(pixbuf);
			}
		}

		create_icon("record",    TEXTURE_ICON_REC);
		create_icon("record-on", TEXTURE_ICON_REC_ON);
		create_icon("solo",      TEXTURE_ICON_SOLO);
		create_icon("solo-on",   TEXTURE_ICON_SOLO_ON);
		create_icon("mute",      TEXTURE_ICON_MUTE);
		create_icon("mute-on",   TEXTURE_ICON_MUTE_ON);
	}
}
#endif


static AGlActor*
arr_gl_background(GtkWidget* panel)
{
	bool gl_background_paint(AGlActor* actor)
	{
		Arrange* arrange = actor->root->user_data;
		GlCanvas* c = (GlCanvas*)arrange->canvas->gl;

	#ifdef BACKGROUND_2
		// procedural background texture.

		if(agl->use_shaders) agl_use_program((AGlShader*)&pattern_shader);

		glBegin(GL_QUADS);
		double top = 0.0;
		double bot = c->image->allocation.height;
		double x1 = 0.0;
		double x2 = c->image->allocation.width - RRf;
		glVertex2d(x1, top);
		glVertex2d(x2, top);
		glVertex2d(x2, bot);
		glVertex2d(x1, bot);
		glEnd();
		//agl_texture_box(bg_textures[TEXTURE_BG], get_style_bg_color_rgba(GTK_STATE_NORMAL), 0, 0, c->image->allocation.width, c->image->allocation.height);
	#else
		if(agl->use_shaders){
			agl->shaders.alphamap->uniform.fg_colour = actor->colour;
			agl_use_program((AGlShader*)agl->shaders.alphamap);
		}
		agl_textured_rect(bg_textures[TEXTURE_BG], actor->region.x1, actor->region.y1, actor->region.x2 - RRf, actor->region.y2, NULL);

		gl_warn("gl error");
	#endif // BACKGROUND_2

		// highlight the selected track
		{
			if(agl->use_shaders){
				agl->shaders.plain->uniform.colour = 0xffffff08;
				agl_use_program((AGlShader*)agl->shaders.plain);

				GList* s = arrange->tracks.selection;
				for(;s;s=s->next){
					ArrTrk* atr = track_list__lookup_track(arrange->priv->tracks, (AMTrack*)s->data);

					agl_rect_((AGlRect){
						.x = 0.0,
						.y = c->actors[ACTOR_TYPE_TC]->region.y1 + atr->y * CGL->zoom.v.val.f + c->actors[ACTOR_TYPE_MAIN]->scrollable.y1,
						.w = c->image->allocation.width,
						.h = atr->height * CGL->zoom.v.val.f,
					});
				}
			}
		}
		return true;
	}

	void gl_background_set_size(AGlActor* a)
	{
		a->region.x2 = ((AGlActor*)a->root)->region.x2;
		a->region.y2 = ((AGlActor*)a->root)->region.y2;
	}

	return (AGlActor*)AGL_NEW(BackgroundLayer,
		.actor = {
			.name = "Background",
#ifdef BG_TEXTURE
			.init = arr_gl_create_bg_texture,
#endif
			.paint = gl_background_paint,
			.set_size = gl_background_set_size,
			//actor->program = (AGlShader*)agl->shaders.alphamap;
			.colour = 0xffffff40
		},
	);
}


#ifdef NOT_USED
static void
arr_gl_draw_background_non_shader(Arrange* arrange)
{
	GlCanvas* c = (GlCanvas*)arrange->canvas->gl;

	GdkColor bg;
	get_style_bg_color(&bg, GTK_STATE_NORMAL);

#ifdef BG_TEXTURE
	glColor4f(2 * ((float)bg.red)/0x10000, 2 * ((float)bg.green)/0x10000, 2 * ((float)bg.blue)/0x10000, 1.0);

	glUseProgram(0); // seems to be required, though shouldnt be.
#if 0
	agl_enable(AGL_ENABLE_TEXTURE_2D);
			glActiveTexture(GL_TEXTURE0);
	//agl_enable(AGL_ENABLE_BLEND);
			if(!glIsTexture(bg_textures[TEXTURE_BG])) gwarn("not texture");
	glBindTexture(GL_TEXTURE_2D, bg_textures[TEXTURE_BG]);
														gl_warn("gl error: texture bind");
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			agl_enable(AGL_ENABLE_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
#else
	use_texture(bg_textures[TEXTURE_BG]);
#endif
			glColor4f(2 * ((float)bg.red)/0x10000, 2 * ((float)bg.green)/0x10000, 2 * ((float)bg.blue)/0x10000, 1.0);

	glBegin(GL_QUADS);
	double top = 0;
	double bot = c->image->allocation.height;
	double x1 = 0;
	double x2 = c->image->allocation.width;
	glTexCoord2d(1.0, 1.0); glVertex2d(x1, top);
	glTexCoord2d(0.0, 1.0); glVertex2d(x2, top);
	glTexCoord2d(0.0, 0.0); glVertex2d(x2, bot);
	glTexCoord2d(1.0, 0.0); glVertex2d(x1, bot);
	glEnd();

	gl_warn("gl error");
#endif
}
#endif


