/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://ayyi.org               |
* | copyright (C) 2013-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include "gl_button.h"

#define BG_OPACITY(B) (button->animatables[0]->val.f)
#define RIPPLE(B) (button->animatables[1]->val.f)

struct {
	AGliPt pt;
} GlButtonPress;

static bool gl_button_on_event(AGlActor*, GdkEvent*, AGliPt);


AGlActor*
gl_button(int icon, ButtonAction action, ButtonGetState get_state, gpointer user_data)
{
	bool gl_button_paint(AGlActor* actor)
	{
		ButtonActor* button = (ButtonActor*)actor;

		bool state = button->get_state ? button->get_state(actor, button->user_data) : false;

		AGlRect r = {
			.x = 0.0,
			.y = 0.0,
			.w = agl_actor__width(actor),
			.h = agl_actor__height(actor)
		};

		// background
		if(RIPPLE(button) > 0.0){
			agl->shaders.plain->uniform.colour = state
				? colour_mix_rgba(button->bg_colour, get_style_bg_color_rgba(GTK_STATE_SELECTED), RIPPLE(button) / 64.0)
				: colour_mix_rgba(get_style_bg_color_rgba(GTK_STATE_SELECTED), button->bg_colour, RIPPLE(button) / 64.0);
		}else if(state){
			agl->shaders.plain->uniform.colour = get_style_bg_color_rgba(GTK_STATE_SELECTED);
		}else{
			agl->shaders.plain->uniform.colour = button->bg_colour;
		}

		// hover background
		if(BG_OPACITY(button) > 0.0){ // dont get events if disabled, so no need to check state (TODO turn off hover when disabling).
			float alpha = button->animatables[0]->val.f;
#if 0
			uint32_t fg = _get_style_bg_color_rgba(actor->root->widget, GTK_STATE_PRELIGHT);
#else
			uint32_t fg = colour_lighter_rgba(agl->shaders.plain->uniform.colour, 16);
#endif
			agl->shaders.plain->uniform.colour = button->bg_colour ? colour_mix_rgba(agl->shaders.plain->uniform.colour, fg, alpha) : (fg & 0xffffff00) + (uint32_t)(alpha * 0xff);
		}

		if(!(RIPPLE(button) > 0.0)){
			agl_use_program((AGlShader*)agl->shaders.plain);
			agl_rect_(r);

		}else{
			// click ripple
			//int radius = RIPPLE(button);
#undef RIPPLE_EXPAND
#ifdef RIPPLE_EXPAND
			circle_shader.uniform.radius = radius;
			agl_use_program((AGlShader*)&circle_shader);
			AGlRect rr = r;
			rr.w = rr.h = radius * 2;
			glTranslatef(GlButtonPress.pt.x - actor->region.x1 - radius, GlButtonPress.pt.y - radius, 0.0);
			agl_rect_(rr);
			glTranslatef(-(GlButtonPress.pt.x - actor->region.x1 - radius), -(GlButtonPress.pt.y - radius), -0.0);
#else
			circle_shader.uniform.colour = colour_lighter_rgba(agl->shaders.plain->uniform.colour, 32);
			circle_shader.uniform.bg_colour = agl->shaders.plain->uniform.colour;
			circle_shader.uniform.centre = GlButtonPress.pt;
			circle_shader.uniform.radius = RIPPLE(button);
			agl_use_program((AGlShader*)&circle_shader);
			agl_rect_(r);
#endif
		}

		//icon
		agl->shaders.texture->uniform.fg_colour = 0xffffff00 + (agl_actor__is_disabled(actor) ? 0x77 : 0xff);
		agl_use_program((AGlShader*)agl->shaders.texture);

		agl_textured_rect(bg_textures[button->icon + (state ? 1 : 0)], r.x, r.y + r.h / 2.0 - 8.0, 16.0, 16.0, NULL);
		//texture_box(arrange, bg_textures[button->icon], get_style_bg_color_rgba(GTK_STATE_NORMAL), 10.0, -10.0, 32.0, 32.0);

		//texture_box(arrange, bg_textures[TEXTURE_BG], get_style_bg_color_rgba(GTK_STATE_NORMAL), 200.0, 10.0, 50.0, 50.0);
		//texture_box(arrange, bg_textures[TEXTURE_ICON], 0xffffffff, 200.0, 61.0, w, h);

		return true;
	}

	bool gl_button_paint_gl1(AGlActor* actor)
	{
		ButtonActor* button = (ButtonActor*)actor;

		bool state = button->get_state ? button->get_state(actor, button->user_data) : false;

		AGlRect r = {
			.x = 0.0,
			.y = 0.0,
			.w = agl_actor__width(actor),
			.h = agl_actor__height(actor->parent)
		};

		agl_enable(0/* !AGL_ENABLE_TEXTURE_2D*/);
		if(state){
			agl_colour_rbga(get_style_bg_color_rgba(GTK_STATE_SELECTED));
			agl_rect_(r);
		}
		if(!button->disabled && agl_actor__is_hovered(actor)){
			glColor4f(1.0, 1.0, 1.0, 0.1);
			agl_rect_(r);
		}

		agl_enable(AGL_ENABLE_TEXTURE_2D);
		glColor4f(1.0, 1.0, 1.0, 1.0);
		agl_textured_rect(bg_textures[button->icon + (state ? 1 : 0)], r.x, r.y + r.h / 2.0 - 8.0, 16.0, 16.0, NULL);

		return true;
	}

	void button_set_size(AGlActor* actor)
	{
		actor->region.y2 = actor->parent->region.y2 - actor->parent->region.y1;
	}

	void button_init(AGlActor* actor)
	{
		if(!circle_shader.shader.program) agl_create_program(&circle_shader.shader);

		actor->paint = agl->use_shaders ? gl_button_paint : gl_button_paint_gl1;

		//ButtonActor* button = (ButtonActor*)actor;
		//button->bg_colour = 0xff9933ff;
	}

	void button_free(AGlActor* actor)
	{
		g_free(((ButtonActor*)actor)->animatables[0]);
		g_free(((ButtonActor*)actor)->animatables[1]);
	}

	ButtonActor* button = AGL_NEW(ButtonActor,
		.actor = {
			.name = "Button",
			.init     = button_init,
			.free     = button_free,
			.paint    = agl_actor__null_painter,
			.set_size = button_set_size,
			.on_event = gl_button_on_event,
		},
		.icon      = icon,
		.action    = action,
		.get_state = get_state,
		.user_data = user_data,
	);

	WfAnimatable* bg = button->animatables[0] = g_new0(WfAnimatable, 1);
	*bg = (WfAnimatable){
		.model_val.f = &button->bg_opacity,
		.start_val.f = 0.0,
		.val.f       = 0.0,
		.type        = WF_FLOAT
	};

	WfAnimatable* ripple = button->animatables[1] = g_new0(WfAnimatable, 1);
	*ripple = (WfAnimatable){
		.model_val.f = &button->ripple_radius,
		.start_val.f = 0.0,
		.val.f       = 0.0,
		.type        = WF_FLOAT
	};

	return (AGlActor*)button;
}


static bool
gl_button_on_event(AGlActor* actor, GdkEvent* event, AGliPt xy)
{
	ButtonActor* button = (ButtonActor*)actor;

	if(button->disabled) return NOT_HANDLED;

	void animation_done (WfAnimation* animation, gpointer user_data)
	{
	}

	void ripple_done (WfAnimation* animation, gpointer user_data)
	{
		ButtonActor* button = user_data;
		RIPPLE(button) = 0.0;
	}

	switch (event->type){
		case GDK_ENTER_NOTIFY:
			dbg (1, "ENTER_NOTIFY");
			//set_cursor(actor->root->widget->window, CURSOR_H_DOUBLE_ARROW);

			button->bg_opacity = 1.0;
			agl_actor__start_transition(actor, g_list_append(NULL, button->animatables[0]), animation_done, NULL);
			return HANDLED;
		case GDK_LEAVE_NOTIFY:
			dbg (1, "LEAVE_NOTIFY");
			//set_cursor(actor->root->widget->window, CURSOR_NORMAL);

			button->bg_opacity = 0.0;
			agl_actor__start_transition(actor, g_list_append(NULL, button->animatables[0]), animation_done, NULL);
			return HANDLED;
		case GDK_BUTTON_PRESS:
			return HANDLED;
		case GDK_BUTTON_RELEASE:
			dbg(1, "BUTTON_RELEASE");
			if(event->button.button == 1){
				call(button->action, actor, button->user_data);

				GlButtonPress.pt = xy;
				button->animatables[1]->val.f = 0.001;
				button->ripple_radius = 64.0;
				agl_actor__start_transition(actor, g_list_append(NULL, button->animatables[1]), ripple_done, button);
				return HANDLED;
			}
			break;
		default:
			break;
	}
	return NOT_HANDLED;
}


void
gl_button_set_sensitive(AGlActor* actor, gboolean sensitive)
{
#ifdef AGL_DEBUG_ACTOR
	dbg(0, "%s: %i", actor->name, sensitive);
#endif
	((ButtonActor*)actor)->disabled = !sensitive;
}


#if 0
static uint32_t
_get_style_bg_color_rgba(GtkWidget* widget, GtkStateType state)
{
	GtkStyle* style = gtk_widget_get_style(widget);

	GdkColor bg = style->bg[state];
	bg.red   = MIN((1 << 16) - 1, bg.red);
	bg.green = MIN((1 << 16) - 1, bg.green);
	bg.blue  = MIN((1 << 16) - 1, bg.blue);

	return color_gdk_to_rgba(&bg);
}
#endif

