/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#define __gl_canvas_c__
#define __gl_canvas_priv__
#define __arrange_private__
#include "global.h"
#include <math.h>
#include <sys/time.h>
#define USE_FBO // it is assumed that driver fbo support exists so if defined, will always be used.
#include <GL/gl.h>
#include <GL/glu.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <pango/pangofc-font.h>
#include <pango/pangofc-fontmap.h>

#include "agl/ext.h"
#include "agl/pango_render.h"
#include "waveform/waveform.h"
#include "waveform/actor.h"
#include "model/am_time.h"
#include "model/am_midi_part.h"
#include "model/am_pool_item.h"
#include "icon.h"
#include "windows.h"
#include "shell.h"
#include "support.h"
#include "song.h"
#include "arrange.h"
#include "arrange/part.h"
#include "part_manager.h"
#include "../../shortcuts.h"
#include "toolbar.h"
#include "toolbox.h"
#include "gl_canvas_op.h"
#include "arrange/animator.h"
#include "arrange/shader.h"
#include "arrange/gl_meter.h"
#include "arrange/gl_track_control.h"
#include "arrange/view_object.h"

#include "gl_canvas.h"

#define BG_TEXTURE
#define GL_AUTOMATION
#define SHOW_LOG
#undef SHOW_LOG

extern SMConfig* config;

#define GLX_PIXMAP_WORKS 0
#define DIRECT true
#define ISOMETRIC_HEIGHT 100
#define ISOMETRIC_X0 100
#define TRACK_SPACING_Z -50
#define PART_TRANSITION_BORDER 2.0

#define RR  0  // canvas right margin - just for debugging. normally 0
#define RRf 0.0

static GdkGLConfig* glconfig = NULL;
static gboolean gl_initialised = false;
static CanvasType gl_canvas_type = 0;
#if 0
static int frame_count = 0;
#endif

AGl* agl = NULL;

GtkWidget*      arr_gl_canvas_new          (Arrange*);
static void     arr_gl_class_init          ();
static void     arr_gl_canvas_free         (Arrange*);

// canvas api
#if 0
static void     arr_gl_canvas_redraw       (Arrange*);
#endif
static void     arr_gl_get_scroll_offsets  (Arrange*, int* x, int* y);
static void     arr_gl_px_to_model         (Arrange*, Pti, AyyiSongPos*, ArrTrk**);
static uint32_t arr_gl_get_viewport_left   (Arrange*);
static double   arr_gl_pos2px              (Arrange*, GPos*);
static float    arr_gl_samples2px          (Arrange*, uint32_t);
static void     arr_gl_canvas_pick         (Arrange*, double, double, TrackDispNum*, AMPart**);
static void     arr_gl_get_view_fraction   (Arrange*, double* fx, double* fy);
static void     arr_gl_set_part_selection  (Arrange*);
static bool     arr_gl_part_new            (Arrange*, AMPart*);
static void     arr_gl_on_part_delete      (Arrange*, AMPart*);
static void     arr_gl_get_part_rect       (Arrange*, AMPart*, DRect*);
static void     arr_gl_do_follow           (Arrange*);
static void     arr_gl_scroll_to           (Arrange*, int x, int y);
static void     arr_gl_scroll_to_pos       (Arrange*, GPos*, TrackDispNum);
static void     arr_gl_scroll_left         (Arrange*, int pix);
static void     arr_gl_scroll_right        (Arrange*, int pix);
static void     arr_gl_scroll_up           (Arrange*, int pix);
static void     arr_gl_scroll_down         (Arrange*, int pix);
static void     arr_gl_set_spp             (Arrange*, uint32_t);
static FeatureType arr_gl_pick_feature     (CanvasOp*);
static void     arr_gl_on_song_load        (Arrange*);
static void     arr_gl_on_view_change      (Arrange*, AMChangeType);
static void     arr_gl_on_part_change      (Arrange*, AMPart*);
static void     arr_gl_on_songlength_change(Arrange*);
static void     arr_gl_on_new_track        (Arrange*, ArrTrk*);
static void     arr_gl_on_track_delete     (Arrange*, ArrTrk*);
static void     arr_gl_on_locators_change  (Arrange*);
// end api

static void     arr_gl_canvas_on_realise   (GtkWidget*, gpointer);
static void     arr_gl_canvas_on_unrealise (GtkWidget*, gpointer);
static bool     arr_gl_canvas_expose       (GtkWidget*, GdkEventExpose*, gpointer data);
static void     arr_gl_on_allocate         (GtkWidget*, GtkAllocation*, Arrange*);
static void     arr_gl_on_hscroll          (GtkWidget*, Arrange*);
static void     arr_gl_on_vscroll          (GtkWidget*, Arrange*);
static void     arr_gl_remove_tracks       (Arrange*);
static void     arr_gl_draw_part_bg        (Arrange*, AMPart*, double start, double top, double bot, double right);
static void     arr_gl_canvas_redraw_fast  (Arrange*);

static AGlActorNew
#ifdef DEBUG
                arr_gl_frame,
#endif
#ifdef SHOW_LOG
                log_actor,
#endif
#ifndef USE_FBO
                arr_gl_parts,
#endif
                spp_actor,
                arr_gl_tracks,
                arr_gl_automation;

static void     arr_gl_parts_destroy       (Arrange*);

static void     gl_set_colour_from_part_bg (AMPart*);
static void     gl_set_palette_colour      (int);
#ifdef USE_FBO
#ifdef NOT_USED
static void     fbo_2_png                  (AGlFBO*);
#endif
#endif

static void     arr_gl_setup_view          (Arrange*);
static void     arr_gl_on_iso_button_press (GtkWidget* widget, Arrange*);
static void     arr_gl_on_iso_key_press    (GtkAccelGroup*, gpointer);
static void     arr_gl_enable_isometric    (Arrange*);
static void     arr_gl_disable_isometric   (Arrange*);
static void     rotation_start             (Arrange*);
static gboolean rotation_frame             (Arrange*);
static void     arr_gl_zoom_start          (Arrange*, AMChangeType);
static void     arr_gl_get_viewport        (Arrange*, WfViewPort*);
#if 0
static GPos*    arr_gl_px2pos              (Arrange*, GPos*, double px);
#endif
static GlTrack*_add_new_gl_track           (Arrange*, AMTrack*);
static GlPart* _add_new_gl_part            (Arrange*, AMPart*);
static void    _remove_gl_part             (GlCanvas*, GlPart*);
static void   __arr_gl_remove_glpart       (gpointer key, gpointer value, gpointer _arrange);

static void     arr_gl_draw_part           (Arrange*, GlPart*);

static void     arr_gl_set_tc_width        (Arrange*, int);

static void     menu_arr_set_gl_canvas     (Arrange*);

#define GL_TRACK_HEIGHT(atr) ((atr->height * (1-c->iso_factor) + ISOMETRIC_HEIGHT * c->iso_factor) * arrange->vzoom)

#define GL_TRACK_HEIGHT2(gltrack) ((((AGlActor*)gltrack->actor)->region.y2 - ((AGlActor*)gltrack->actor)->region.y1) * (1-c->iso_factor) \
                  + (ISOMETRIC_HEIGHT * arrange->vzoom) * c->iso_factor)

typedef void (*ViewChange) (AGlActor*, AMChangeType);
ViewChange on_change[ACTOR_TYPE_MAX] = {NULL,};

#include "gl_tracks.c"
#include "gl_ruler.c"
#include "gl_background.c"
#include "gl_button.c"
#include "gl_divider.c"
#include "gl_grid.c"
#include "gl_op.c"
#include "gl_automation.c"
#include "gl_spp.c"
#include "gl_overview.c"
#include "gl_scrollbar.c"
#include "gl_target.c"
#include "gl_iso.c"

extern GHashTable* canvas_types;

AMAccel gl_keys[] = {
	{"Isometric",   {{(char)'i',},}, arr_gl_on_iso_key_press, NULL, NULL},
};

#define ROOT() ((AGlRootActor*)CGL->actor)


void
arr_gl_canvas_init()
{
	gl_canvas_type = arr_register_canvas("gl", arr_gl_canvas_new, menu_arr_set_gl_canvas);
	CanvasClass* cc = g_hash_table_lookup(canvas_types, &gl_canvas_type);

	cc->provides = PROVIDES_METERS | PROVIDES_RULERBAR | PROVIDES_TRACKCTL | PROVIDES_OVERVIEW;
}


CanvasType
arr_gl_canvas_get_type()
{
	return gl_canvas_type;
}


static void
arr_gl_class_init()
{
	if(gl_initialised) return;

	AyyiPanelClass* k = g_type_class_peek(AYYI_TYPE_ARRANGE);

	GimpActionGroup* ag = shortcuts_add_group("Gl Arrange");

	GtkAccelGroup* g = gtk_accel_group_new();
	make_accels(g, ag, gl_keys, G_N_ELEMENTS(gl_keys), NULL);
	k->accels = g_list_append(k->accels, g);

	agl = agl_get_instance();
#ifdef USE_FBO
	agl_get_extensions();
#endif

	on_change[ACTOR_TYPE_RULER] = arr_gl_ruler_on_view_change;
	on_change[ACTOR_TYPE_OVERVIEW] = arr_gl_overview_on_view_change;
	on_change[ACTOR_TYPE_HSCROLLBAR] = arr_gl_scrollbar_on_view_change;
	on_change[ACTOR_TYPE_VSCROLLBAR] = arr_gl_scrollbar_on_view_change;

	gl_initialised = true;

#if 0
	gboolean check_framerate(gpointer _)
	{
		printf("n_frames=%.1f\n", ((float)frame_count) / 10.0);
		frame_count = 0;
		return TIMER_CONTINUE;
	}
	g_timeout_add(10000, check_framerate, NULL);
#endif
}


GtkWidget*
arr_gl_canvas_new(Arrange* arrange)
{
	PF;
	AyyiPanelClass* panel_class = g_type_class_peek(AYYI_TYPE_ARRANGE);
	arr_gl_class_init();

	ArrCanvas* canvas = arrange->canvas = AYYI_NEW(ArrCanvas,
		.type                 = arr_gl_canvas_get_type(),
		.free                 = arr_gl_canvas_free,
		.get_scroll_offsets   = arr_gl_get_scroll_offsets,
		.px_to_model          = arr_gl_px_to_model,
		.get_viewport_left    = arr_gl_get_viewport_left,
		.pick                 = arr_gl_canvas_pick,
		.get_view_fraction    = arr_gl_get_view_fraction,
		.set_part_selection   = arr_gl_set_part_selection,
		.part_new             = arr_gl_part_new,
		.part_delete          = arr_gl_on_part_delete,
		.get_part_rect        = arr_gl_get_part_rect,
		.do_follow            = arr_gl_do_follow,
		.scroll_to            = arr_gl_scroll_to,
		.scroll_to_pos        = arr_gl_scroll_to_pos,
		.scroll_left          = arr_gl_scroll_left,
		.scroll_right         = arr_gl_scroll_right,
		.scroll_up            = arr_gl_scroll_up,
		.scroll_down          = arr_gl_scroll_down,
		.set_spp              = arr_gl_set_spp,
		.pick_feature         = arr_gl_pick_feature,

		.on_song_load         = arr_gl_on_song_load,
		.on_view_change       = arr_gl_on_view_change,
		.on_part_change       = arr_gl_on_part_change,
		.on_songlength_change = arr_gl_on_songlength_change,
		.on_new_track         = arr_gl_on_new_track,
		.on_track_delete      = arr_gl_on_track_delete,
		.on_locator_change    = arr_gl_on_locators_change,

		.op                   = gl_canvas_op__new(arrange, OP_NONE),

		.h_adj = GTK_ADJUSTMENT(gtk_adjustment_new(0.0, 0.0, 200.0, 1.0, 10.0, 10.0)),
		.v_adj = GTK_ADJUSTMENT(gtk_adjustment_new(0.0, 0.0, 200.0, 1.0, 10.0, 10.0)),
	);

	GlCanvas* c = arrange->canvas->gl = AGL_NEW(GlCanvas,
		.is_new = true,
		.isometric = false,
		.part_background_alpha = false, // c->isometric
		.iso_factor = 0,
		.zoom = {
			.h = (WfAnimatable){{.f=arrange->hzoom}, .model_val.f = &arrange->hzoom, .type=WF_FLOAT},
			.v = (WfAnimatable){{.f=arrange->vzoom}, .model_val.f = &arrange->vzoom, .type=WF_FLOAT},
		},
		.parts = g_hash_table_new_full(NULL, NULL, NULL, g_free),
		.tracks = g_hash_table_new_full(NULL, NULL, NULL, g_free),
	);

	void _arr_gl_on_song_unload(GObject* _arrange, gpointer _)
	{
		Arrange* arrange = (Arrange*)_arrange;
		if(arrange->canvas){
			arr_gl_parts_destroy(arrange);
			arr_gl_remove_tracks(arrange);
		}
		arr_gl_canvas_redraw_fast(arrange);
	}
	g_signal_connect(arrange, "song-unload", G_CALLBACK(_arr_gl_on_song_unload), c);

	g_signal_connect(canvas->h_adj, "value-changed", G_CALLBACK(arr_gl_on_hscroll), arrange);
	g_signal_connect(canvas->v_adj, "value-changed", G_CALLBACK(arr_gl_on_vscroll), arrange);

	GtkWidget* table = c->table = canvas->container = gtk_table_new(2, 2, NON_HOMOGENOUS);

	GList* l = panel_class->accels;
	for(;l;l=l->next){
		GtkAccelGroup* g = l->data;
		gtk_window_add_accel_group(GTK_WINDOW(arrange->panel.shell->widget), g);
		g_object_set_data(G_OBJECT(g), "sm_window", arrange);
	}

#ifdef ROTATION_SLIDERS
	// rotate slider 1
	GtkObject* rotate_adj = gtk_adjustment_new (0.0, -10.0, 10.0, 0.1, 1.0, 1.0);
	GtkWidget* rotate_scale = gtk_hscale_new (GTK_ADJUSTMENT (rotate_adj));
	gtk_scale_set_value_pos(GTK_SCALE(rotate_scale), GTK_POS_RIGHT);
	gtk_box_pack_start (GTK_BOX(arrange->sm_window.toolbar_box[1]), rotate_scale, TRUE, TRUE, 0);
	g_signal_connect((gpointer)rotate_scale, "value-changed", G_CALLBACK(gl_on_rotate), arrange);

	// rotate slider 2
	rotate_adj = gtk_adjustment_new (0.0, -10.0, 10.0, 0.1, 1.0, 1.0);
	rotate_scale = gtk_hscale_new (GTK_ADJUSTMENT (rotate_adj));
	gtk_scale_set_value_pos(GTK_SCALE(rotate_scale), GTK_POS_RIGHT);
	gtk_box_pack_start (GTK_BOX(arrange->sm_window.toolbar_box[1]), rotate_scale, TRUE, TRUE, 0);
	g_signal_connect((gpointer)rotate_scale, "value-changed", G_CALLBACK(gl_on_rotate2), arrange);

	// rotate slider 3
	rotate_adj = gtk_adjustment_new (0.0, -10.0, 10.0, 0.1, 1.0, 1.0);
	rotate_scale = gtk_hscale_new (GTK_ADJUSTMENT (rotate_adj));
	gtk_scale_set_value_pos(GTK_SCALE(rotate_scale), GTK_POS_RIGHT);
	gtk_box_pack_start (GTK_BOX(arrange->sm_window.toolbar_box[1]), rotate_scale, TRUE, TRUE, 0);
	g_signal_connect((gpointer)rotate_scale, "value-changed", G_CALLBACK(gl_on_rotate3), arrange);
#endif

	// TODO buttons added with toolbar_add_new_button do not have tooltips
	c->iso_button = toolbar_add_new_button(&arrange->panel, AGL_NEW(ToolbarButton,
		.key         = "zoom_out",
		.icon        = "zoom_out",
		.callback    = G_CALLBACK(arr_gl_on_iso_button_press),
		.action_name = "toggle isometric"
	));

	if(!glconfig){
		if(!(glconfig = gdk_gl_config_new_by_mode(GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH | GDK_GL_MODE_STENCIL | GDK_GL_MODE_DOUBLE))){
			gerr ("Cannot initialise gtkglext."); return NULL;
		}
	}

	int width = 256, height = 128;
	c->image = arrange->canvas->widget = gtk_drawing_area_new();
	gtk_widget_set_can_focus(c->image, true);
	gtk_widget_set_size_request(c->image, width, height);
	dbg(2, "setting gl_capability context=%p", agl_get_gl_context());
	gtk_widget_set_gl_capability(c->image, glconfig, agl_get_gl_context(), DIRECT, GDK_GL_RGBA_TYPE);
	gtk_widget_add_events (c->image, GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);

	c->actor = agl_actor__new_root(c->image);
	SCENE->user_data = arrange;

	bool root_on_event(AGlActor* actor, GdkEvent* event, AGliPt xy)
	{
		Arrange* arrange = ((AGlScene*)actor)->user_data;
		GlCanvas* c = arrange->canvas->gl;

		switch (event->type){
			case GDK_BUTTON_PRESS:
				switch(event->button.button){
					case 3:
						shell__menu_popup(arrange->panel.shell, event);
				}
				return AGL_HANDLED; // prevent dock drag operation starting
			case GDK_ENTER_NOTIFY:
				;int T[] = {ACTOR_TYPE_VSCROLLBAR, ACTOR_TYPE_HSCROLLBAR};
				int t; for(t=0;t<G_N_ELEMENTS(T);t++)
				if(c->actors[T[t]] && c->actors[T[t]]->disabled){
					c->actors[T[t]]->disabled = false;
					agl_actor__invalidate(c->actors[T[t]]);
				}
				break;
			case GDK_LEAVE_NOTIFY:
				;GdkEventCrossing* e = (GdkEventCrossing*)event;
				if(e->detail == GDK_NOTIFY_ANCESTOR && c->actors[ACTOR_TYPE_VSCROLLBAR]){
					c->actors[ACTOR_TYPE_VSCROLLBAR]->disabled = true;
					agl_actor__invalidate(c->actors[ACTOR_TYPE_VSCROLLBAR]);

					c->actors[ACTOR_TYPE_HSCROLLBAR]->disabled = true;
					agl_actor__invalidate(c->actors[ACTOR_TYPE_HSCROLLBAR]);
				}
				break;
		}
		return AGL_NOT_HANDLED;
	}

	c->actor->on_event = root_on_event;

	g_signal_connect((gpointer)c->image, "realize", G_CALLBACK(arr_gl_canvas_on_realise), arrange);
	g_signal_connect((gpointer)c->image, "unrealize", G_CALLBACK(arr_gl_canvas_on_unrealise), arrange);
	g_signal_connect((gpointer)c->image, "size-allocate", G_CALLBACK(arr_gl_on_allocate), arrange);
	g_signal_connect(G_OBJECT(c->image), "expose_event", G_CALLBACK(arr_gl_canvas_expose), arrange);
	g_signal_connect_swapped((gpointer)c->image, "event", G_CALLBACK(agl_actor__on_event), c->actor);

	gtk_table_attach(GTK_TABLE(table), c->image, 0, 1, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);

	gtk_drag_dest_set(GTK_WIDGET(arrange->canvas->widget), GTK_DEST_DEFAULT_ALL, app->dnd.file_drag_types, app->dnd.file_drag_types_count, (GdkDragAction) (GDK_ACTION_MOVE | GDK_ACTION_COPY));
	g_signal_connect(G_OBJECT(arrange->canvas->widget), "drag-data-received", G_CALLBACK(arr_canvas_drag_received), NULL);

	arr_gl_zoom_start(arrange, AM_CHANGE_ZOOM_H | AM_CHANGE_ZOOM_V);

	return arrange->canvas_scrollwin = table;
}


static void
arr_gl_canvas_free(Arrange* arrange)
{
	// Do not free canvas->op here, it is a singleton

	PF;
	GlCanvas* c = arrange->canvas->gl;
	AyyiPanelClass* k = g_type_class_peek(AYYI_TYPE_ARRANGE);

	// should not be neccesary but just in case...
	#define handler_disconnect(FN) if(g_signal_handlers_disconnect_matched (c->image, G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, 0, 0, NULL, FN, arrange) != 1) gwarn("canvas handler disconnection");
	handler_disconnect(agl_actor__on_event);
	#undef handler_disconnect

	g_signal_handlers_disconnect_matched(arrange, G_SIGNAL_MATCH_ID | G_SIGNAL_MATCH_DATA, ((AyyiArrangeClass*)k)->song_unload_signal, 0, NULL, NULL, c);

	int n;
	AMTrack* tr;
	ArrTrk* at = NULL;
	while((at = track_list__next_visible(arrange->priv->tracks, at ? at->track : NULL)) && (tr = at->track)){
		if(tr->meteradj)
			if((n = g_signal_handlers_disconnect_matched (tr->meteradj, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, at)) != 1) gwarn("canvas meter handler disconnection. t=%i n=%i", tr->track_num, n);
	}

	arrange->canvas->op = NULL;

	dbg(1, "destroying hashtables... parts.size=%i", g_hash_table_size(c->parts));
	g_hash_table_foreach(c->parts, __arr_gl_remove_glpart, arrange);
	g_hash_table_destroy(c->parts);
	c->parts = NULL;

	arr_gl_remove_tracks(arrange);
	g_hash_table_destroy(c->tracks);
	c->tracks = NULL;

	if(c->wfc) wf_context_free0(c->wfc);
	agl_actor__free(c->actor);
	gtk_widget_destroy0(arrange->canvas->widget);
	gtk_widget_destroy0(c->table);
	gtk_widget_destroy0(c->iso_button);
	g_free0(arrange->canvas->gl);
	g_free0(arrange->canvas);

	GList* l = k->accels;
	for(;l;l=l->next){
		GtkAccelGroup* g = l->data;
		AyyiShell* shell = lookup_shell_by_widget((GtkWindow*)gtk_widget_get_toplevel((GtkWidget*)arrange));
		if(shell){
			gtk_window_remove_accel_group(GTK_WINDOW(shell->widget), g);
		}
	}

	if(!arr_get_n_gl_windows()){
#ifdef FREE_TEXTURE
		#ifdef USE_FBO
		glDeleteFramebuffersEXT(1, &fbo->texture);
		g_free0(fbo0);
		#endif
#endif
	}
}


static void
arr_gl_remove_tracks(Arrange* arrange)
{
	GlCanvas* c = arrange->canvas->gl;

	void __arr_gl_remove_gltrack(gpointer key, gpointer value, gpointer _arrange)
	{
		Arrange* arrange = _arrange;

		AGlActor* tc = CGL->actors[ACTOR_TYPE_TC];
		while(tc->children) agl_actor__remove_child(tc, tc->children->data);
	}
	g_hash_table_foreach(c->tracks, __arr_gl_remove_gltrack, arrange);
	g_hash_table_remove_all(c->tracks);
}


static void
arr_gl_setup_view(Arrange* arrange)
{
	GlCanvas* c = arrange->canvas->gl;
	g_return_if_fail(c);
	GtkWidget* widget = c->image;

	// setting the viewport affects both the gain and the maximium extent of the screen coords

	int vx = 0/* + (c->isometric ? 100 : 0)*/;
	int vy = 0;
	int vw = widget->allocation.width;
	int vh = widget->allocation.height;
	// viewport defines the *screen coords* that opengl can draw within.
	// note: -the viewport is modified by glOrtho() (except that viewport doesnt deal with screen coords).
	//       -the viewport is centred in the drawing area if smaller.
	glViewport(vx, vy, vw, vh);
	dbg (2, "viewport: %i %i %i %i", vx, vy, vw, vh);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// establish clipping volume (left, right, bottom, top, near, far)
	glOrtho (0.0, c->image->allocation.width, vh, 0.0, 800.0, -1200.0);

	dbg (2, "rot=%.2f %.2f %.2f", rotate[0], rotate[1], rotate[2]);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	if(c->isometric){
		glRotatef(rotate[0], 1.0f, 0.0f, 0.0f);
		glRotatef(rotate[1], 0.0f, 1.0f, 0.0f);
		glScalef(1.0f, 1.0f, -1.0f);
	}
}


static void
add_audio_part_actor_to_wf_canvas(GlCanvas* c, GlPart* glpart, AMPart* part)
{
	g_return_if_fail(PART_IS_AUDIO(part));
	g_return_if_fail(!glpart->wf_actor);

	glpart->wf_actor = wf_canvas_add_new_actor(c->wfc, (Waveform*)part->pool_item);
	wf_actor_set_region(glpart->wf_actor, &(WfSampleRegion){part->region_start, part->ayyi->length});

	uint32_t fg = agl->use_shaders
		? (part->fg_colour & 0xffffff00) + 156                         // set alpha to 65% - TODO we are overriding a user_setting here?
		: (colour_grey_out_rgba(part->fg_colour) & 0xffffff00) + 0xff; // force opaque due to unresolved blending issues TODO

	wf_actor_set_colour(glpart->wf_actor, fg);
	((AGlActor*)glpart->wf_actor)->name = part->name;

	dbg(2, "%s", part->name);
}

static void
arr_gl_create_part_actor(gpointer key, gpointer value, gpointer _canvas)
{
	GlPart* glpart = (GlPart*)value;
	if(PART_IS_AUDIO((AMPart*)key) && !glpart->wf_actor){
		add_audio_part_actor_to_wf_canvas((GlCanvas*)_canvas, glpart, (AMPart*)key);

		//
		// TODO parts are not assigned parent actors!
		//
	}
}


static void
arr_gl_canvas_on_realise(GtkWidget* widget, gpointer _arrange)
{
	// Note that it can happen that we go through multiple realise/unrealise events when creating a new docked panel.

	// Canvas level objects must not be created until their parent objects are created.
	// ie before gl tracks and gl parts are created, the Panel level objects such as ArrTrk have been created by the Panel.

	PF;
	g_return_if_fail(_arrange);
	Arrange* arrange = _arrange;
	GlCanvas* c = arrange->canvas->gl;

	if(!c->wfc){
		c->wfc = wf_context_new((AGlRootActor*)c->actor);

		if(((AyyiPanel*)arrange)->model_is_loaded && SHOW_PARTCONTENTS){
			g_hash_table_foreach(c->parts, arr_gl_create_part_actor, c);
		}

		char font_string[64];
		get_font_string(font_string, -3); // TODO why do we have to set this so small to get a reasonable font size?
		agl_set_font_string(font_string);
	}

	// WaveformActors are currently destroyed in unrealise so need to be recreated here

	// At startup we should not create parts here as they are created via arrange.on_song_load()
	// however it IS needed when the canvas is unrealised then realised again, or when a second window is opened.
	// *** TODO perhaps it is now no longer necc to remove actors on unrealise? ***

	// -if the song is not fully loaded, the arrange tracks may not be created yet, in which case this will be handled later by the song load handler.
	if(((AyyiPanel*)arrange)->model_is_loaded){
		if(!g_hash_table_size(c->parts)){
			part_foreach {
				arrange->canvas->part_new(arrange, gpart);
			} end_part_foreach;

			// this appears to be unnecesary (see above "never get here") - remove if so.
#if 0
			g_hash_table_foreach(c->parts, arr_gl_create_part_actor, c);
#endif
		}

		if(!g_hash_table_size(c->tracks)){
			// iterate over ArrTrk* so that the gl tracks will not be created until after the Panel level objects are done.
			AMTrack* tr;
			ArrTrk* at = NULL;
			while((at = track_list__next_visible(arrange->priv->tracks, at ? at->track : NULL)) && (tr = at->track)){
				arr_gl_on_new_track(arrange, at);
			}
		}
	}

	// connect signals - must be done _before_ actors are instantiated.
	if(c->is_new){

		void _arr_gl_on_track_delete(GObject* _song, AMTrack* tr, Arrange* arrange)
		{
			if(tr->meteradj)
				signal_handler_disconnect_data(tr->meteradj, track_list__lookup_track(arrange->priv->tracks, tr), 1, "canvas meter");

			if(!g_hash_table_remove(CGL->tracks, tr)) gwarn("unable to remove track from hashtable");

			// TODO redraw fast is too heavy - instead, just clear the FBO of the deleted track and queue a redraw
			arr_gl_canvas_redraw_fast(arrange);
		}
		g_signal_connect(song->tracks, "delete", G_CALLBACK(_arr_gl_on_track_delete), arrange);

		// we do not connect to song->tracks "added" - is done via canvas->on_new_track
		// we do not connect to song->pool "delete" - we expect a part-delete signal instead
	}

	if(!c->actor->children){
		typedef struct {
			int          type;
			AGlActorNew* new;
		} Item;

		void add_actors(AGlActor* actor, int n, Item actors[n], GlCanvas* c)
		{
			int i; for(i=0;i<n;i++){
				Item* item = &actors[i];
				if(item->new){
					AGlActor* child = agl_actor__add_child(actor, item->new((GtkWidget*)arrange));
					if(item->type > -1) c->actors[item->type] = child; // TODO remove c arg
				}
			}
		}

		Item actors[] = {
			{ACTOR_TYPE_BACKGROUND, arr_gl_background},
#ifdef SHOW_LOG
			{-1,              log_actor},
#endif
#ifdef DEBUG
			{-1,              arr_gl_frame},
#endif
#ifndef USE_FBO
			{-1,              arr_gl_parts            },
#endif
			{-1,              arr_gl_divider          },
			{ACTOR_TYPE_MAIN, arr_gl_op_actor         },
			{ACTOR_TYPE_TC,   track_control_actor_new },
			{-1,              meter_actor_new         },
			{ACTOR_TYPE_RULER,arr_gl_ruler},
		};

		add_actors(c->actor, G_N_ELEMENTS(actors), actors, c);

		// parts and spp are done as a child of the MAIN actor.
		Item actors2[] = {
			{-1,                arr_gl_grid},
			{-1,                spp_actor},
			{ACTOR_TYPE_TRACKS, arr_gl_tracks},
			{ACTOR_TYPE_FG,     arr_gl_foreground},
			{-1,                arr_gl_automation},
		};
		add_actors(c->actors[ACTOR_TYPE_MAIN], G_N_ELEMENTS(actors2), actors2, c);

		// actors that are initially inactive
		c->actors[ACTOR_TYPE_ISO_FRAME] = arr_gl_iso_frame((GtkWidget*)arrange);
		c->actors[ACTOR_TYPE_ISO_TC] = arr_gl_iso_tc((GtkWidget*)arrange);

		AGlActor* scrollbar;
		agl_actor__add_child(c->actor, (c->actors[ACTOR_TYPE_HSCROLLBAR] = scrollbar = arr_gl_scrollbar((GtkWidget*)arrange, GTK_ORIENTATION_HORIZONTAL), scrollbar->z = 100, scrollbar));
		agl_actor__add_child(c->actor, (c->actors[ACTOR_TYPE_VSCROLLBAR] = scrollbar = arr_gl_scrollbar((GtkWidget*)arrange, GTK_ORIENTATION_VERTICAL), scrollbar->z = 100, scrollbar));

		AGlActor* overview;
		agl_actor__add_child(c->actor, (c->actors[ACTOR_TYPE_OVERVIEW] = overview = arr_gl_song_overview((GtkWidget*)arrange), overview->z = 99, overview));

		CanvasClass* cc = g_hash_table_lookup(canvas_types, &gl_canvas_type);
		if(!(cc->provides & PROVIDES_TRACKCTL)){
			c->actors[ACTOR_TYPE_MAIN]->region.x1 = 0;
			c->actors[ACTOR_TYPE_BACKGROUND]->region.x1 = 0;
		}
	}

	if(gdk_gl_drawable_gl_begin(ROOT()->gl.gdk.drawable, ROOT()->gl.gdk.context)){

		if(!agl->use_shaders) glUseProgram(0); // is needed though shouldnt be

		gdk_gl_drawable_gl_end(ROOT()->gl.gdk.drawable);
	}

#ifdef CURIOUS
	// Check settings for our opengl implementation
	GLfloat sizes[2];   // supported line width range
	GLfloat step;       // supported line width increments
	// Get supported line width range and step size
	glGetFloatv(GL_LINE_WIDTH_RANGE, sizes);
	glGetFloatv(GL_LINE_WIDTH_GRANULARITY, &step);
	dbg (1, "GL_LINE_WIDTH_RANGE: %.2f - %.2f  GL_LINE_WIDTH_GRANULARITY: %.2f", sizes[0], sizes[1], step);
#endif

	c->is_new = false;
}


static void
arr_gl_canvas_on_unrealise(GtkWidget* widget, gpointer _arrange)
{
	PF;
	Arrange* arrange = _arrange;

	arr_gl_parts_destroy(arrange);

#if 0 // it should be unneccesary to free the wf_canvas
	GlCanvas* c = arrange->canvas->gl;
	if(c->wfc) wf_canvas_free0(c->wfc);
#endif
}


void
arr_gl_allocate_part(Arrange* arrange, GlPart* glpart)
{
	// Currently we attempt to render the whole part. This fails for very large parts (perhaps ~300bars)

	GlCanvas* c = arrange->canvas->gl;
	g_return_if_fail(glpart);
	if(!glpart->wf_actor) return;
	AMPart* part = glpart->part;
	g_return_if_fail(part);
	AMTrack* track = part->track;
	g_return_if_fail(track);
	ArrTrk* atr = track_list__lookup_track(arrange->priv->tracks, track);

	WfViewPort viewport; arr_gl_get_viewport(arrange, &viewport);

	WfRectangle rect = {
		.left = arr_gl_pos2px_(arrange, &part->start) - viewport.left,
		.top = atr->y * arrange->vzoom, // note: vscroll position not considered part of the allocation
		.len = arr_gl_pos2px_(arrange, &part->length),
		.height = GL_TRACK_HEIGHT(atr)
	};

	dbg(2, "%s start=%.2f length=%.2f", part->name, rect.left, rect.len);

	wf_actor_set_rect(glpart->wf_actor, &rect);
}


static void
__arr_gl_allocate_part(gpointer key, gpointer value, gpointer _arrange)
{
	arr_gl_allocate_part((Arrange*)_arrange, (GlPart*)value);
}


static void
arr_gl_wf_on_viewport_change(Arrange* arrange)
{
	g_hash_table_foreach(arrange->canvas->gl->parts, __arr_gl_allocate_part, arrange);
}


static void
arr_gl_on_allocate(GtkWidget* widget, GtkAllocation* allocation, Arrange* arrange)
{
#ifdef DEBUG
	if(!arrange_verify(arrange)) return;
#endif
	GlCanvas* c = arrange->canvas->gl;

	if(gtk_widget_get_realized(c->image)){
		arr_gl_wf_on_viewport_change(arrange);

		c->actor->scrollable.x2 = c->actors[ACTOR_TYPE_MAIN]->region.x2 = c->actor->region.x2 = allocation->width;
		c->actor->scrollable.y2 = c->actors[ACTOR_TYPE_MAIN]->region.y2 = c->actor->region.y2 = allocation->height;

		agl_actor__set_size(c->actor);
	}
}


static bool
arr_gl_canvas_expose(GtkWidget* widget, GdkEventExpose* event, gpointer _arrange)
{
	g_return_val_if_fail(_arrange, false);
	Arrange* arrange = _arrange;
	GlCanvas* c = (GlCanvas*)arrange->canvas->gl;

#if 0
	frame_count++;
#endif

	if(gdk_gl_drawable_gl_begin(ROOT()->gl.gdk.drawable, ROOT()->gl.gdk.context)){
		gl_warn("pre");

		arr_gl_setup_view(arrange);

		GdkColor bg;
		get_style_bg_color(&bg, GTK_STATE_NORMAL);
#ifdef BG_TEXTURE
		glClearColor(0.5 * ((float)bg.red)/0x10000, 0.5 * ((float)bg.green)/0x10000, 0.5 * ((float)bg.blue)/0x10000, 1.0);
#else
		glClearColor(((float)bg.red)/0x10000, ((float)bg.green)/0x10000, ((float)bg.blue)/0x10000, 1.0);
#endif
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		//XX no cannot do this, as it removed the rotation set in setup_view
		//glLoadIdentity(); //something in the font rendering is messing with this

		agl_actor__paint(c->actor);

		gdk_gl_drawable_swap_buffers(ROOT()->gl.gdk.drawable);
		gdk_gl_drawable_gl_end(ROOT()->gl.gdk.drawable);
	}
	else gwarn ("gl_begin failed!");
 
	return true;
}


static void
arr_gl_canvas_redraw_fast(Arrange* arrange)
{
	// This excludes any changes to viewport or dimensions but is still
	// relatively slow due to the fbo's being cleared.
	// For fastest redraw, use gtk_widget_queue_draw()

#ifdef USE_FBO
	CGL->actors[ACTOR_TYPE_TRACKS]->invalidate(CGL->actors[ACTOR_TYPE_TRACKS]);
#endif
	gtk_widget_queue_draw(arrange->canvas->gl->image);
}


void
arr_gl_canvas_redraw_part(Arrange* arrange, AMPart* part)
{
	// This assumes that the part has not moved track. If track has changed you need to also redraw the old track (invalidate cache).

	GlCanvas* c = arrange->canvas->gl;
	g_return_if_fail(part);

	GlPart* glpart = g_hash_table_lookup(c->parts, part);
	g_return_if_fail(glpart);
	arr_gl_allocate_part(arrange, glpart);

	GlTrack* gltrack = g_hash_table_lookup(c->tracks, part->track);
	if(gltrack) gltrack->actor->parts.cache.valid = false;
	gtk_widget_queue_draw(c->image);
}


void
arr_gl_canvas_redraw_track(Arrange* arrange, ArrTrk* at)
{
	g_return_if_fail(at);
	GlCanvas* c = arrange->canvas->gl;

	// TODO only really need to re-allocate parts on _this_ track
	g_hash_table_foreach(c->parts, __arr_gl_allocate_part, arrange);

	GlTrack* gltrack = g_hash_table_lookup(c->tracks, at->track);
	if(gltrack) gltrack->actor->parts.cache.valid = false;
	gtk_widget_queue_draw(arrange->canvas->gl->image);
}


void
arr_gl_canvas_paint_local_part(Arrange* arrange, LocalPart* lpart)
{
	GlCanvas* c = (GlCanvas*)arrange->canvas->gl;
	AMPart* part = (AMPart*)lpart;

	AMTrack* track = part->track;
	ArrTrk* atr = track_list__lookup_track(arrange->priv->tracks, track);
	WfRectangle r = {
		.left = arr_gl_pos2px_(arrange, &part->start) - CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1,
		.height = GL_TRACK_HEIGHT(atr),
		.len = arr_gl_pos2px_(arrange, &part->length),
		.top = atr->y * arrange->vzoom - c->actors[ACTOR_TYPE_MAIN]->scrollable.y1
	};
	double bot = r.top + r.height;
	double right = r.left + r.len;

	arr_gl_draw_part_bg(arrange, part, r.left, r.top, bot, right);
}


#if 0
static void
arr_gl_canvas_redraw(Arrange* arrange)
{
	//this is a _full_ redraw. Use arr_gl_canvas_redraw_fast if size or viewport has not changed.

	GlCanvas* c = arrange->canvas->gl;

	g_hash_table_foreach(c->parts, __arr_gl_allocate_part, arrange);
#ifdef USE_FBO
	if(c->actors[ACTOR_TYPE_TRACKS])
		c->actors[ACTOR_TYPE_TRACKS]->invalidate(CGL->actors[ACTOR_TYPE_TRACKS]);
#endif

	gtk_widget_queue_draw(c->image);

	arr_gl_hscrollbar_update(arrange);
	arr_gl_vscrollbar_update(arrange);
}
#endif


#ifdef DEBUG
static AGlActor*
arr_gl_frame(GtkWidget* panel)
{
	bool arr_gl_draw_frame(AGlActor* actor)
	{
	#undef SHOW_ORIGIN
	#ifdef SHOW_ORIGIN
			agl_enable(0, /* !AGL_ENABLE_TEXTURE_2D */);
			glColor3f(1.0, 0.0, 0.0);

			// show origin
			glLineWidth(1);
			glBegin(GL_LINES);
			glVertex3f( 0, 0, 0 );    glVertex3f(   1,    0, 0 );
			glVertex3f( 0, 0, 0 );    glVertex3f(   0,   -1, 0 );

			// box 320x240
			glVertex3f( 320, -240, 0 ); glVertex3f(320,    0, 0 );
			glVertex3f( 0,   -240, 0 ); glVertex3f(320, -240, 0 );
			glEnd();
			agl_enable(AGL_ENABLE_TEXTURE_2D);
	#endif

	#undef SHOW_PLANES
	#ifdef SHOW_PLANES
			agl_enable(0 /* !AGL_ENABLE_TEXTURE_2D */);
			glColor3f(0.0, 1.0, 0.0);
			int w = 10000;

			// plane at z=0
			//glRectf(-w, -w, w, w);

			//2 more planes facing us, but in front and behind
			glColor3f(1.0, 0.5, 0.0);
			glBegin( GL_QUADS );
			glVertex3f(w, -w, 1.0 ); glVertex3f(w,  0, 1.0);
			glVertex3f(0, -w, 1.0 ); glVertex3f(w, -w, 1.0);
			glEnd();

			glColor3f(0.0, 0.5, 1.0);
			glBegin( GL_QUADS );
			glVertex3f(w, -w, -1.0 ); glVertex3f(w,  0, -1.0);
			glVertex3f(0, -w, -1.0 ); glVertex3f(w, -w, -1.0);
			glEnd();

			//pane at x=0: (havnt actually seen this yet)
			glColor3f(1.0, 1.0, 0.0);
			glBegin( GL_QUADS );
			glVertex3f( 0, w, -w ); glVertex3f(0, w,  0);
			glVertex3f( 0, 0, -w ); glVertex3f(0, w, -w);
			glEnd();

			//pane at y=0: (havnt actually seen this yet)
			glColor3f(0.0, 1.0, 1.0);
			glBegin(GL_QUADS);
			glVertex3f( w, 0, -w ); glVertex3f(w, 0,  0);
			glVertex3f( 0, 0, -w ); glVertex3f(w, 0, -w);
			glEnd();

			agl_enable(AGL_ENABLE_TEXTURE_2D);
	#endif
		return true;
	}

	return AYYI_NEW(AGlActor,
		.name = "Frame",
		.paint = arr_gl_draw_frame,
		.region = {0, 0, 1, 1}
	);
}
#endif


#define print_part_label() \
    agl_print(start + 2, bot - 11, part->track->track_num * -5.0 * 0.001, part->fg_is_set ? part->fg_colour : song->palette[1], "%s", part->name);
//                                                             -----                        
//                                                             TODO fbo needs more z-depth

static inline void
arr_gl_draw_note(ArrTrk* at, AMPart* part, MidiNote* note, double part_x, double part_y, uint32_t colour, double part_z)
{
	Arrange* arrange = at->arrange;

	if(agl->use_shaders){
		agl->shaders.plain->uniform.colour = colour;
		agl_use_program((AGlShader*)agl->shaders.plain);
	}else{
		glColor3f(1.0, 1.0, 0.0);
	}
	agl_rect_((AGlRect){
		.x = part_x + arr_samples2px(arrange, note->start),
		.y = part_y + arr_track_note_pos_y(at, note),
		.w = MAX(arr_samples2px(arrange, note->length), ARR_MIN_NOTE_DISPLAY_LENGTH),
		.h = ((ArrTrackMidi*)at)->note_height
	});
}


static inline void
arr_gl_draw_note_velocity_bar(ArrTrk* at, AMPart* part, MidiNote* note, uint32_t colour)
{
	#define velocity_bar_width 6

	ArrTrackMidi* atm = (ArrTrackMidi*)at;
	Arrange* arrange = at->arrange;
	int chart_height = atm->velocity_chart_height;
	int track_height = at->height * arrange->vzoom;

	// border
	if(agl->use_shaders){
		agl->shaders.plain->uniform.colour = 0x000000ff;
		agl_use_program((AGlShader*)agl->shaders.plain);
	}
	agl_rect_((AGlRect){
		.x = arr_samples2px(arrange, note->start),
		.y = track_height,
		.w = velocity_bar_width,
		.h = - (chart_height * note->velocity) / 128
	});

	// main bar
	if(agl->use_shaders){
		agl->shaders.plain->uniform.colour = colour;
		agl_use_program((AGlShader*)agl->shaders.plain);
	}
	agl_rect_((AGlRect){
		.x = arr_samples2px(arrange, note->start),
		.y = track_height,
		.w = velocity_bar_width - 1,
		.h = - (chart_height * note->velocity) / 128
	});
}


static void
arr_gl_draw_part(Arrange* arrange, GlPart* glpart)
{
	void arr_gl_draw_audio_contents(Arrange* arrange, GlTrack* gltrack, GlPart* glpart, WfRectangle r, double llf, double top0, double bot)
	{
		AMPart* part = glpart->part;
		g_return_if_fail(part->pool_item);

		GlCanvas* c = (GlCanvas*)arrange->canvas->gl;
		WfViewPort viewport; arr_gl_get_viewport(arrange, &viewport);
		AMTrack* track = part->track;
		ArrTrk* atr = track_list__lookup_track(arrange->priv->tracks, track);

		if(arrange->view_options[SHOW_PART_CONTENTS].value){
			if(glpart) glpart->wf_actor->z = track->track_num * -50.0; //TODO move to setup_view?

			if(glpart){
				if(agl->use_shaders) glpart->wf_actor->fg_colour = part->fg_colour; // check if this is needed. it should be set only in add_audio_part_actor_to_wf_canvas

				float height0 = atr->height * arrange->vzoom;
				float height = bot - r.top;
				float scale = height / height0;
				glPushMatrix();
				glScalef(1.0f, scale, 1.0f);
#ifdef USE_FBO
				llf += c->actors[ACTOR_TYPE_MAIN]->scrollable.x1 - gltrack->actor->parts.cache_start_px;
#endif
#ifdef USE_FBO
				glTranslatef(llf + (c->isometric ? ISOMETRIC_X0 : 0), 0.0f, 0.0f);
#else
				glTranslatef(llf + (c->isometric ? ISOMETRIC_X0 : 0), -(top0 - r.top) - c->actors[ACTOR_TYPE_MAIN]->scrollable.y1, 0.0f);
#endif
				((AGlActor*)glpart->wf_actor)->paint((AGlActor*)glpart->wf_actor);
				glPopMatrix();
			}
		}

	}

	void arr_gl_draw_midi_contents(ArrTrk* at, AMPart* part, WfRectangle r, double llf, double top0, double bot)
	{
		AMTrack* track = part->track;
		MidiPart* mpart = (MidiPart*)part;

		double z = track->track_num * -50.0;

		// draw midi notes
		MidiNote* note = NULL;
		int safety = 0;
		while((note = am_midi_part__get_next_event(part, note))){
			arr_gl_draw_note(at, part, note, r.left, r.top, 0xff0000ff, z);
			arr_gl_draw_note_velocity_bar(at, part, note, 0x0ffffff);
			if(safety++ > 512){ gerr("bailing out!"); break; }
		}
		if(!safety) dbg(1, "no events");

		GList* l = mpart->note_selection;
		for(;l;l=l->next){
			MidiNote* note = l->data;
			arr_gl_draw_note(at, part, note, r.left, r.top, 0xffffffff, z);
		}
	}

	AMPart* part = glpart->part;
	GlCanvas* c = (GlCanvas*)arrange->canvas->gl;
	ArrTrk* atr = track_list__lookup_track(arrange->priv->tracks, part->track);
	GlTrack* gltrack = g_hash_table_lookup(c->tracks, part->track);
	WfViewPort viewport; arr_gl_get_viewport(arrange, &viewport);

	WfRectangle r = {
#ifdef USE_FBO
		.top    = PART_TRANSITION_BORDER,
		.left   = arr_gl_pos2px_(arrange, &part->start) - gltrack->actor->parts.cache_start_px, // ignore viewport
#else
		.top    = atr->y * arrange->vzoom - c->actors[ACTOR_TYPE_MAIN]->scrollable.y1,
		.left   = arr_gl_pos2px_(arrange, &part->start) - viewport.left,
#endif
		.height = GL_TRACK_HEIGHT2(gltrack),
		.len    = arr_gl_pos2px_(arrange, &part->length),
	};

#ifdef USE_FBO
	double bot = r.top + r.height;
#else
	double bot = r.height;
#endif
	double top0 = r.top;

#ifndef USE_FBO
	if((r.top > c->actors[ACTOR_TYPE_MAIN]->scrollable.y2) || (atr->y * arrange->vzoom + r.height < c->actors[ACTOR_TYPE_MAIN]->scrollable.y1)) return; // outside viewport

	double top_ = r.top * (1 - c->iso_factor);
	bot -= r.top - top_;
	r.top = top_;
#endif
	//dbg(0, "bot=%.2f top=%.2f-->%.2f pos=%i start=%.2f", bot, top0, top, atr->y * arrange->vzoom, r.left);
#ifdef USE_FBO
	double llf = 0.0; // temp - use gltranslate instead

#if 0
	// fbo border line
	{
		g_return_if_fail(!((AGlActor*)glpart->wf_actor)->fbo);
		if(gltrack->actor->parts.cache.fbo){
			float w = gltrack->actor->parts.cache.fbo->width;
			float h = gltrack->actor->parts.cache.fbo->height;
			agl->shaders.plain->uniform.colour = 0xff0000ff;
			agl_use_program((AGlShader*)agl->shaders.plain);
			agl_rect(0.0, 0.0, w, 1.0); // top
			agl_rect(w - 1.0, 0.0, 1.0, h); // rhs
			//agl->shaders.plain->uniform.colour = 0xffff00ff;
			agl_use_program((AGlShader*)agl->shaders.plain);
			agl_rect(0.0, h - 1.0, w, 1.0); // bottom
			agl_rect(0.0, 0.0, 1.0, h); // lhs
		}
	}
#endif

#else
	double llf = arrange->tc_width.val.i;
#endif

	//TODO should we perhaps use WaveformActor->rect here? are they the same size? yes, mostly
#if 0
	{
		if(glpart){
			WaveformActor* actor = glpart->wf_actor;
			if(actor){
				WfRectangle* rect = &actor->rect;
				dbg(0, "  l=%.2f %.2f %.2f %.2f %.2f %.2f", rect->left, r.left, rect->len, r.len, rect->top, r.top); //top is not the same when using fbo (track offset)
			}
		}
	}
#endif

	arr_gl_draw_part_bg(arrange, part, llf + r.left, r.top, bot, llf + r.left + r.len);

	if(r.len > 2.0){
		agl_enable_stencil(llf + r.left, MAX(0.0, r.top), r.len, r.height);

		switch(part->ident.type){
			case AYYI_OBJECT_AUDIO_PART:
				arr_gl_draw_audio_contents(arrange, gltrack, glpart, r, llf, top0, bot);
				break;
			case AYYI_OBJECT_MIDI_PART:
				arr_gl_draw_midi_contents(atr, part, r, llf, top0, bot);
				break;
			default:
				gwarn("bad AMPart type: %i", part->ident.type);
				break;
		}

		if(r.len > 7.0 && r.height > 10.0){
#ifdef USE_FBO
			bot = r.height;
			float start = r.left;
#else
			float start = llf + r.left;
#endif
			print_part_label();
		}

		agl_disable_stencil();
	}

	// outline
	if (!agl->use_shaders) {
		AGlColourFloat cf;
		colour_rgba_to_float(&cf, arr_part_is_selected(arrange, part)
			? config->part_outline_colour_selected
#define IGNORE_CONFIG
#ifdef IGNORE_CONFIG
			: song->palette[part->bg_colour] //testing: use part colour instead
#else
			: config->part_outline_colour
#endif
		);
		dbg(3, "background_colour=%i %x", part->bg_colour, song->palette[part->bg_colour]);
		agl_enable(0 /* !AGL_ENABLE_TEXTURE_2D */);
		glColor4f(cf.r, cf.g, cf.b, 0.5);
		glLineWidth(1);
		glBegin(GL_LINES);
			glVertex3f(r.left,         r.top, 0 );    glVertex3f(r.left + r.len,  r.top, 0 );
			glVertex3f(r.left + r.len, r.top, 0 );    glVertex3f(r.left + r.len,  bot,   0 );
			glVertex3f(r.left,         bot,   0 );    glVertex3f(r.left + r.len,  bot,   0 );
			glVertex3f(r.left,         r.top, 0 );    glVertex3f(r.left,          bot,   0 );
		glEnd();
	}
}


#ifndef USE_FBO
static AGlActor*
arr_gl_parts(AyyiPanel* panel)
{
	bool arr_gl_draw_parts(AGlActor* actor)
	{
		// Each part is drawn as several rectangles, one per waveform texture.

		Arrange* arrange = actor->root->user_data;
		GlCanvas* c = arrange->canvas->gl;

		void _each(gpointer key, gpointer value, gpointer arrange)
		{
			GlPart* glpart = (GlPart*)value;
			arr_gl_draw_part((Arrange*)arrange, glpart);
		}
		g_hash_table_foreach(c->parts, _each, arrange);

		return true;
	}

	return AYYI_NEW(AGlActor,
		.name = "GlParts",
		.paint = arr_gl_draw_parts,
		.region = (AGliRegion){0, ((Arrange*)panel)->ruler_height, 4096, 4096}
	);
}
#endif


static void
arr_gl_draw_part_bg(Arrange* arrange, AMPart* part, double start, double top, double bot, double right)
{
#ifdef USE_FBO
	double z = 0.0;
#else
	double z = part->track->track_num * -50.0;
#endif

	top = MAX(0.0, top);             // dont allow drawing over the rulerbar
	right = MAX(right, start + 1.0); // minimun width of 1px
	double width = right - start;

	glPushMatrix();
	if(agl->use_shaders){
		glTranslatef(start - PART_TRANSITION_BORDER, top - PART_TRANSITION_BORDER, z);

		part_shader.uniform.colour = song->palette[part->bg_colour];
		part_shader.uniform.selected = arr_part_is_selected(arrange, part);
		part_shader.shader.uniforms[PART_SHADER_WIDTH].value[0] = right - start;
		part_shader.shader.uniforms[PART_SHADER_HEIGHT].value[0] = bot - top;
		glGetFloatv(GL_MODELVIEW_MATRIX, part_shader.uniform.modelview);
		glGetFloatv(GL_PROJECTION_MATRIX, part_shader.uniform.projection);
		agl_use_program((AGlShader*)&part_shader);

		agl_rect(0, 0, 2 * PART_TRANSITION_BORDER + right - start, 2 * PART_TRANSITION_BORDER + bot - top);
	}else{
		glTranslatef(start, top, z);

		int bg_width = 256;

		double tpx = 1.0;
		if(width < bg_width) tpx = width / bg_width;
		// if part is large, only a fraction of the Part is textured.
		// So we need to draw 2 objects, one textured, one not.
		agl_enable(AGL_ENABLE_TEXTURE_2D | AGL_ENABLE_BLEND);
		gl_set_colour_from_part_bg(part);
		agl_textured_rect(bg_textures[TEXTURE_BG], 0, 0, right - start, bot - top, &(AGlQuad){0.0, 0.0, tpx, 1.0});

		if(width > bg_width){
			// draw untextured rectangle.
		}
	}
	glPopMatrix();
}


static void
arr_gl_canvas_pick(Arrange* arrange, double x, double y, TrackDispNum* track_idx, AMPart** part)
{
	// x and y should be in post-zoom world-space (ie viewport offsets already applied).
	// ie x=0 should always correspond to 01:01:01:000 ?

	if(track_idx) *track_idx = -1;
	if(part) *part = NULL;
	if(x < 0) return gwarn("coords out of range: %.1f %.1f", x, y);

	CanvasOp* op = arrange->canvas->op;

	if(y - op->yscrolloffset < 0){
		// rulerbar
	} else {
		TrackDispNum d = arr_px2trk(arrange, y); // TODO this probably doesnt work in iso mode.
		if(d < 0) return;
		if(track_idx) *track_idx = d;
		AMTrack* t = track_list__get_trk_from_display_index (arrange->priv->tracks, d);

		if(part){

			AyyiSongPos pos;
			ArrTrk* _ = NULL;
			arrange->canvas->px_to_model(arrange, (Pti){x - CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1, y}, &pos, &_);

			*part = (AMPart*)am_partmanager__find_by_position(t, &pos);
		}
	}
}


static void
arr_gl_get_view_fraction(Arrange* arrange, double* fx, double* fy)
{
	// Get the fraction of the whole canvas that is currently visible in the window

	// This actually works for all canvases, theres no real need for one per canvas type.

	float tot_x_beats = am_object_val(&song->loc[AM_LOC_END]).sp.beat + 4;
	float visible_x_px = arrange->canvas->widget->allocation.width;
	float visible_x_beats = visible_x_px / (PX_PER_BEAT * arrange->hzoom);

	*fx = visible_x_beats / tot_x_beats;

	if(fy){
		ArrTrk* atr = arrange->priv->tracks->track[am_track_list_count(song->tracks) -1];

		float visible_y_px = arrange->canvas->widget->allocation.height;
		float total_y_px = (atr->y + atr->height) * arrange->vzoom + arrange->ruler_height;

		*fy = visible_y_px / total_y_px;
	}
}


static void
arr_gl_px_to_model(Arrange* arrange, Pti px, AyyiSongPos* pos, ArrTrk** track)
{
	/*
	 * Transforms a raw widget pixel coordinate to the song position and track.
	 *
	 * It is intended to be a replacement for get_scroll_offsets which is not adequate for the transformation for some canvas types
	 *
	 */

	px.x += CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1 - arrange->tc_width.val.i;
	px.y += CGL->actors[ACTOR_TYPE_MAIN]->scrollable.y1 - arrange->ruler_height;

	arr_snap_px2spos(arrange, pos, px.x); // TODO probably should _not_ snap

	if(track){
		TrackDispNum d = arr_px2trk(arrange, px.y);
		dbg(1, "d=%i", d);
		if(d > -1 && d < AM_MAX_TRKX) *track = arrange->priv->tracks->display[d];
	}
}


static void
arr_gl_get_scroll_offsets(Arrange* arrange, int* cx, int* cy)
{
		if(CGL->actors[ACTOR_TYPE_MAIN]){
	*cx = -CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1;
	*cy = -CGL->actors[ACTOR_TYPE_MAIN]->scrollable.y1;
		}
}


static uint32_t
arr_gl_get_viewport_left(Arrange* arrange)
{
	GlCanvas* c = (GlCanvas*)arrange->canvas->gl;

	return -c->actors[ACTOR_TYPE_MAIN]->scrollable.x1 - (c->isometric ? ISOMETRIC_X0 : 0);
}


static void
arr_gl_get_viewport(Arrange* arrange, WfViewPort* viewport)
{
	// Get pixel coords (x1,y1,x2,y2) of the current viewport relative to top left of track1.

	GlCanvas* c = arrange->canvas->gl;
															g_return_if_fail(c->actors[ACTOR_TYPE_MAIN]);
	*viewport = (WfViewPort){
		.left   =  c->actors[ACTOR_TYPE_MAIN]->scrollable.x1 - (c->isometric ? ISOMETRIC_X0 : 0),
		.right  =  c->actors[ACTOR_TYPE_MAIN]->scrollable.x1 + c->image->allocation.width,
		.top    =  c->actors[ACTOR_TYPE_MAIN]->scrollable.y1/* - RULERBAR_HEIGHT_GL*/,
																							// TODO negatives cancel out ?
		.bottom = -(-c->actors[ACTOR_TYPE_MAIN]->scrollable.y1 + arrange->ruler_height - c->image->allocation.height),
	};
}

/*
 *  Scroll the arrange canvas to the given absolute canvas position.
 *  -use -1 to keep existing position.
 */
static void
arr_gl_scroll_to(Arrange* arrange, int x, int y)
{
	GlCanvas* c = arrange->canvas->gl;
	g_return_if_fail(c->actors[ACTOR_TYPE_MAIN]);
	dbg(1, "x=%i y=%i", x, y);

	AMChangeType change = 0;

	if(x > -1 && x != -c->actors[ACTOR_TYPE_MAIN]->scrollable.x1){
		arrange->canvas->op->mouse.x = -1;

		GtkAdjustment* adj = GTK_ADJUSTMENT(arrange->canvas->h_adj);
		int page_size = agl_actor__width(c->actors[ACTOR_TYPE_MAIN]); // not using adj->page_size because viewport is different width than the scrollbar
		double x_ = CLAMP((double)x, adj->lower, adj->upper > page_size ? adj->upper - page_size : adj->upper);
		int width = c->actors[ACTOR_TYPE_MAIN]->scrollable.x2 - c->actors[ACTOR_TYPE_MAIN]->scrollable.x1;
		c->actors[ACTOR_TYPE_MAIN]->scrollable.x1 = -x_;
						c->actors[ACTOR_TYPE_MAIN]->scrollable.x2 = -x_ + width;
		gtk_adjustment_set_value(adj, x_);
		change |= AM_CHANGE_POS_X;
	}

	if(y > -1 && y != -c->actors[ACTOR_TYPE_MAIN]->scrollable.y1){
		arrange->canvas->op->mouse.y = -1;

		GtkAdjustment* adj = GTK_ADJUSTMENT(arrange->canvas->v_adj);
		double y_ = MIN((double)y, MAX(0, adj->upper - adj->page_size));
		c->actors[ACTOR_TYPE_MAIN]->scrollable.y1 = -y_;
		c->actors[ACTOR_TYPE_TC]->scrollable.y1 = -y_;
		gtk_adjustment_set_value(adj, y_);
		change |= AM_CHANGE_VIEWPORT_Y;
	}

	arr_gl_on_view_change(arrange, change);
}


static void
arr_gl_hscrollbar_update(Arrange* arrange)
{
	double fx; arr_gl_get_view_fraction(arrange, &fx, NULL);

	float visible_x_px = arrange->canvas->widget->allocation.width;
	float total_x_px = arrange->canvas_width;

	dbg(2, "totalx=%.2f fx=%.2f", total_x_px, fx);

	GtkAdjustment* h_adj = arrange->canvas->h_adj;
	h_adj->lower = 0;
	h_adj->upper = total_x_px;
	h_adj->step_increment = visible_x_px / 100; //just guessing...
	h_adj->page_increment = visible_x_px;
	h_adj->page_size = visible_x_px;
	gtk_adjustment_changed(h_adj);

#if 0
	show_widget_if(c->hscrollbar, fx < 1.0);
#endif
}


static void
arr_gl_vscrollbar_update(Arrange* arrange)
{
	GlCanvas* c = arrange->canvas->gl;

	float visible_y_px = arrange->canvas->widget->allocation.height;
#if 0
	ArrTrk* last_track = arrange->priv->tracks->track[am_track_list_count(song->tracks) - 1];

	float total_y_px = (last_track->y + last_track->height) * arrange->vzoom + arrange->ruler_height;
#else
	float total_y_px = c->actors[ACTOR_TYPE_MAIN]->scrollable.y2 + arrange->ruler_height;
#endif

	GtkAdjustment* v_adj = arrange->canvas->v_adj;
	v_adj->lower = 0;
	v_adj->upper = total_y_px;
	v_adj->step_increment = visible_y_px / 64;
	v_adj->page_increment = visible_y_px;
	v_adj->page_size = visible_y_px;
	gtk_adjustment_changed(v_adj);

	dbg(2, "totaly=%.2f", total_y_px);
}


#if 0
static GPos*
arr_gl_px2pos(Arrange* arrange, GPos* pos, double px)
{
	//converts a horizontal arrange canvas position to a song position.

	float beats = px / (((float)PX_PER_BEAT) * CGL->zoom.h.val.f);
	uint64_t mu = beats * AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT;
	mu2pos(mu, pos);

	return pos;
}
#endif


AyyiSongPos*
arr_gl_px2songpos(Arrange* arrange, AyyiSongPos* pos, double px)
{
	if(px >= 0.0){
		float beats = px / (((float)PX_PER_BEAT) * CGL->zoom.h.val.f);
		uint64_t mu = beats * AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT;
		ayyi_mu2pos(mu, pos);
	}else{
		*pos = (AyyiSongPos){0,};
	}

	return pos;
}


static double
arr_gl_pos2px(Arrange* arrange, GPos* pos)
{
	// Convert a song position to a post-zoom world position.
	// (differs from arr_pos2px in that it takes into account the transient zoom)

	g_return_val_if_fail(arrange, -1.0);
	gpos_is_valid(pos);

	float ticks_per_beat_f = (float)AYYI_TICKS_PER_BEAT;

	double beats_float = pos->beat + (double)pos->sub / 4.0 + ((double)pos->tick / ticks_per_beat_f);

	return /*LLf + */beats_float * PX_PER_BEAT * CGL->zoom.h.val.f;
}


double
arr_gl_pos2px_(Arrange* arrange, AyyiSongPos* pos)
{
	// Convert a song position to a post-zoom world position.
	// (differs from arr_pos2px in that it takes into account the transient zoom)

	g_return_val_if_fail(arrange, -1.0);
	ayyi_pos_is_valid(pos);

	double beats_float = pos->beat + (double)pos->sub / ((double)AYYI_SUBS_PER_BEAT) + ((double)pos->mu) / ((double)AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT);

	return beats_float * PX_PER_BEAT * CGL->zoom.h.val.f;
}


static float
arr_gl_samples2px(Arrange* arrange, uint32_t samples)
{
	float secs = (float)samples / (float)song->sample_rate;
	float beats = secs * BEATS_PER_SECOND;

	return beats * PX_PER_BEAT * CGL->zoom.h.val.f;
}


static void
arr_gl_on_hscroll(GtkWidget* widget, Arrange* arrange)
{
	g_return_if_fail(arrange);

	CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1 = -gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));
	dbg (2, "x=%.2f", -CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1);

	arr_on_view_changed(arrange, AM_CHANGE_POS_X);
}


static void
arr_gl_on_vscroll(GtkWidget* widget, Arrange* arrange)
{
	g_return_if_fail(arrange);
	GlCanvas* c = arrange->canvas->gl;

	GtkAdjustment* scroll_v_adj = GTK_ADJUSTMENT(widget);
	c->actors[ACTOR_TYPE_MAIN]->scrollable.y1 = -floor(gtk_adjustment_get_value(scroll_v_adj));

	arr_gl_on_view_change(arrange, AM_CHANGE_VIEWPORT_Y);
}


static void
arr_gl_on_song_load(Arrange* arrange)
{
	GlCanvas* c = arrange->canvas->gl;

#if 0
	CGL->actors[ACTOR_TYPE_TC] = agl_actor__replace_child(c->actor, c->actors[ACTOR_TYPE_TC], track_control_actor_new((GtkWidget*)arrange));
#else

	// temporarily using a modified version of agl_actor__replace_child so that the old actor gets freed before the new actor is created (otherwise cannot disconnect signals properly)

	AGlActor* _agl_actor__replace_child(AGlActor* actor, AGlActor* child, AGlActor* new_child)
	{
		GList* l = g_list_find(actor->children, child);
		if(l){
			agl_actor__add_child(actor, new_child);

			// update the children list such that the original order is preserved
			actor->children = g_list_remove(actor->children, new_child);
			GList* j = g_list_find(actor->children, child);
			j->data = new_child;

#if 0
			agl_actor__free(child);
#endif

			return new_child;
		}
		return NULL;
	}

	if(c->actors[ACTOR_TYPE_TC]){
		GList* placeholder = g_list_find(c->actor->children, c->actors[ACTOR_TYPE_TC]);
		agl_actor__free(c->actors[ACTOR_TYPE_TC]);
		AGlActor* ph = g_new0(AGlActor, 1);
		placeholder->data = ph;
		CGL->actors[ACTOR_TYPE_TC] = _agl_actor__replace_child(c->actor, ph, track_control_actor_new((GtkWidget*)arrange));
		g_free(ph);
	}
#endif

	// Although zoom is not a song property, zoom changes are not currently
	// done when the model is not loaded as there is a risk that state may be undefined.
	arr_gl_on_view_change(arrange, AM_CHANGE_ZOOM_H | AM_CHANGE_ZOOM_V);
}


static void
arr_gl_on_view_change(Arrange* arrange, AMChangeType change_type)
{
	dbg(1, "%s", am_print_change(change_type));

	GlCanvas* c = arrange->canvas->gl;

	if(!c->actor || !c->actors[ACTOR_TYPE_MAIN]) return;

	if(change_type & (AM_CHANGE_ZOOM_H | AM_CHANGE_ZOOM_V)){
		ArrTrk* last_track = arrange->priv->tracks->track[am_track_list_count(song->tracks) - 1];

		c->actors[ACTOR_TYPE_MAIN]->scrollable.x2 = arr_gl_pos2px_(arrange, &(AyyiSongPos){am_object_val(&song->loc[AM_LOC_END]).sp.beat,}) + ARR_SONG_END_MARGIN;
		c->actors[ACTOR_TYPE_MAIN]->scrollable.y2 = (last_track->y + last_track->height) * arrange->vzoom;

		arr_gl_zoom_start(arrange, change_type);

		// non animated items
		if(change_type & AM_CHANGE_ZOOM_H){
			arr_gl_hscrollbar_update(arrange);
		}
		if(change_type & AM_CHANGE_ZOOM_V){
			arr_gl_vscrollbar_update(arrange);
		}
	}

	if(change_type & AM_CHANGE_WIDTH) arr_gl_hscrollbar_update(arrange);
	if(change_type & AM_CHANGE_HEIGHT) arr_gl_vscrollbar_update(arrange);

	if(change_type & AM_CHANGE_WIDTH || change_type & AM_CHANGE_HEIGHT){
		agl_actor__set_size(c->actor);
	}

#ifdef USE_FBO
	void maybe_invalidate_track_caches(Arrange* arrange)
	{
		void __track_invalidate(gpointer key, gpointer value, gpointer _arrange)
		{
			Arrange* arrange = _arrange;
			GlTrack* gltrack = value;
			g_return_if_fail(gltrack && gltrack->actor);
			ArrTrk* at = gltrack->actor->track;
			GlCanvas* c = arrange->canvas->gl;

			if(gltrack->actor->parts.cache.fbo){
				AyyiSongPos track_end_pos, track_start_pos;
				am_track__get_start_end(at->track, &track_start_pos, &track_end_pos);

				iRange track = {
					.start = arr_gl_pos2px_(arrange, &track_start_pos),
					.end   = arr_gl_pos2px_(arrange, &track_end_pos)
				};
				iRange cache = {
					.start = gltrack->actor->parts.cache_start_px,
					.end   = gltrack->actor->parts.cache_start_px + gltrack->actor->parts.cache.fbo->width
				};
				iRange viewport = {
					.start = -c->actors[ACTOR_TYPE_MAIN]->scrollable.x1,
					.end   = -c->actors[ACTOR_TYPE_MAIN]->scrollable.x1 + agl_actor__width(c->actors[ACTOR_TYPE_MAIN])
				};

				bool fail1 = (MAX(viewport.start, track.start) < cache.start);
				bool fail2 = (MIN(viewport.end,   track.end  ) > cache.end);
				if(fail1) dbg(1, "cache miss (start). invalidating... (%s)", at->track->name);
				if(fail2) dbg(1, "cache miss (end). %i > %i invalidating... (%s)", viewport.end, cache.end, at->track->name);
				if(fail1 || fail2) gltrack->actor->parts.cache.valid = false;
			}
		}
		g_hash_table_foreach(CGL->tracks, __track_invalidate, arrange);
	}

	if(change_type & AM_CHANGE_POS_X){
		arr_gl_wf_on_viewport_change(arrange);
		maybe_invalidate_track_caches(arrange);
	}
#endif

	if(change_type & AM_CHANGE_VIEWPORT_Y){
		// TODO remove dependency on track_list_box
#if defined(USE_GNOMECANVAS) || defined(USE_CLUTTER)
		GtkAdjustment* adj = gtk_viewport_get_vadjustment(GTK_VIEWPORT(arrange->track_list_box));
		adj->value = c->actors[ACTOR_TYPE_MAIN]->scrollable.y1;
		gtk_adjustment_value_changed(adj);
#endif
	}

	if(change_type & AM_CHANGE_BACKGROUND_COLOUR){
		dbg(1, "AM_CHANGE_BACKGROUND_COLOUR unimplemented");
	}

	int i; for(i=0;i<ACTOR_TYPE_MAX;i++){
		if(c->actors[i] && on_change[i]) on_change[i](c->actors[i], change_type);
	}

	if(change_type & AM_CHANGE_BACKGROUND_COLOUR){
		c->actors[ACTOR_TYPE_BACKGROUND]->colour = arrange->bg_colour;
	}

	if(change_type & AM_CHANGE_SHOW_PART_CONTENTS){
		if(((AyyiPanel*)arrange)->model_is_loaded){
			if(SHOW_PARTCONTENTS){
				g_hash_table_foreach(c->parts, arr_gl_create_part_actor, c);
			}
			agl_actor__invalidate(CGL->actors[ACTOR_TYPE_TRACKS]);
		}
	}

	if(change_type & AM_CHANGE_SHOW_SONG_OVERVIEW){
		agl_actor__set_size(CGL->actors[ACTOR_TYPE_OVERVIEW]);

		if(!SHOW_OVERVIEW){
			CGL->actors[ACTOR_TYPE_HSCROLLBAR]->region = (AGliRegion){0, 0, 1, 1};
			agl_actor__set_size(CGL->actors[ACTOR_TYPE_HSCROLLBAR]);
		}else{
			CGL->actors[ACTOR_TYPE_HSCROLLBAR]->region = (AGliRegion){0,};
		}
	}

	if(change_type & AM_CHANGE_PEAK_GAIN){
		if(((AyyiPanel*)arrange)->model_is_loaded){
			if(SHOW_PARTCONTENTS){
				void arr_gl_each_vzoom(gpointer key, gpointer value, gpointer _arrange)
				{
					GlPart* glpart = (GlPart*)value;
					wf_actor_set_vzoom(glpart->wf_actor, ((Arrange*)_arrange)->peak_gain);
				}
				g_hash_table_foreach(c->parts, arr_gl_each_vzoom, arrange);
			}
			agl_actor__invalidate(CGL->actors[ACTOR_TYPE_TRACKS]);
		}
	}

	gtk_widget_queue_draw(CGL->image);
}


static void
arr_gl_on_songlength_change(Arrange* arrange)
{
	arrange->canvas->h_adj->upper = arrange->canvas_width;
}


static void
arr_gl_on_locators_change(Arrange* arrange)
{
	agl_actor__invalidate(CGL->actors[ACTOR_TYPE_RULER]);
	agl_actor__invalidate(CGL->actors[ACTOR_TYPE_OVERVIEW]);
}


static void
arr_gl_on_new_track(Arrange* arrange, ArrTrk* at)
{
	void _meter(GtkAdjustment* a, ArrTrk* at)
	{
		GlCanvas* c = at->arrange->canvas->gl;
		gtk_widget_queue_draw(c->image);
	}

	_add_new_gl_track(arrange, at->track);

	// connect meters. only needed when transport is stopped because otherwise screen will be updated anyway.
	am_track__enable_metering(at->track);
	if(at->track->meteradj) g_signal_connect(at->track->meteradj, "value-changed", (GCallback)_meter, at);
}


static void
arr_gl_on_track_delete(Arrange* arrange, ArrTrk* at)
{
	PF0;
}


static bool
arr_gl_part_new(Arrange* arrange, AMPart* part)
{
	//note: new parts may not yet have any peak data.
	PF;
	GlCanvas* c = arrange->canvas->gl;

	GlPart* glpart = g_hash_table_lookup(c->parts, part);
	g_return_val_if_fail(!glpart, false);

#if 0
	int num_blocks = pi->waveform->num_peaks / WF_PEAK_TEXTURE_SIZE;
	if(pi->waveform->num_peaks % WF_PEAK_TEXTURE_SIZE) num_blocks++;
	if(!num_blocks) gwarn("num_blocks=0");
#endif

#ifdef TEST_TEXTURE_PERF
	GLboolean resident[num_blocks];
	if(glAreTexturesResident(num_blocks, blocks->texture_name, resident)) dbg (0, "all textures are resident.");
	else {
		dbg (0, "not all textures are resident.");
		int i; for(i=0;i<num_blocks;i++) printf("%i ", resident[i]);
		if(i) printf("\n");
	}
#endif

	GlPart* gp = _add_new_gl_part(arrange, part);

	if(PART_IS_AUDIO(part)){
		dbg(2, "n_channels=%i visible=%i", waveform_get_n_channels(gp->wf_actor->waveform), arr_part_is_in_viewport(arrange, part));
	}
														else{
															gwarn("midi part");
														}
	if(arr_part_is_in_viewport(arrange, part)) gtk_widget_queue_draw(c->image);

	return true;
}


static void
arr_gl_on_part_delete(Arrange* arrange, AMPart* part)
{
	GlCanvas* c = arrange->canvas->gl;

	// delete all the textures used by this part.
	FilterIterator* i = filter_iterator_new (0, NULL, NULL, am_parts, (Filter)am_part__has_pool_item, part->pool_item);
	AMPart* other = (AMPart*)filter_iterator_next(i);
	filter_iterator_unref(i);
	if(!other && part->pool_item){
#if 0
		// no, do not do this - we didnt add a reference, so dont remove one.
		g_object_unref(part->pool_item->waveform); 
#endif
	}

	_remove_gl_part(c, g_hash_table_lookup(c->parts, part));

	if(!g_hash_table_remove(c->parts, part)) gwarn("unable to remove part from hashtable");
}


static void
arr_gl_parts_destroy(Arrange* arrange)
{
	GlCanvas* c = arrange->canvas->gl;

	if(c->parts){
		g_hash_table_foreach(c->parts, __arr_gl_remove_glpart, arrange);
		g_hash_table_remove_all(c->parts);
	}
}


static void
arr_gl_set_part_selection(Arrange* arrange)
{
	arr_gl_canvas_redraw_fast(arrange); // need to undraw previous selection TODO how do we know what was the previous selection?

	if(!arrange->part_selection){
		return;
	}

	typedef struct
	{
		Arrange* arrange;
		int      i;
	} C;

	C* c = AYYI_NEW(C,
		.arrange = arrange // TODO when arrange is destroyed, this animation must be stopped.
	);

	bool selection_frame(gpointer _c)
	{
		C* c = _c;
		GList* l = c->arrange->part_selection;
		for(;l;l=l->next){
			arr_gl_canvas_redraw_part(c->arrange, (AMPart*)l->data);
		}
		part_shader.uniform.frame = ++ c->i;
		if(c->i > 5){
			g_free(c);
			return TIMER_STOP;
		}
		return TIMER_CONTINUE;
	}

	animator_add(selection_frame, c);
}


static void
arr_gl_on_part_change(Arrange* arrange, AMPart* part)
{
	PF;
	GlCanvas* c = arrange->canvas->gl;

	GlPart* glpart = g_hash_table_lookup(c->parts, part);
	if(glpart){
		if(glpart->wf_actor) wf_actor_set_region(glpart->wf_actor, &(WfSampleRegion){part->region_start, part->ayyi->length});
		arr_gl_allocate_part(arrange, glpart);
	}

#ifdef USE_FBO
	c->actors[ACTOR_TYPE_TRACKS]->invalidate(CGL->actors[ACTOR_TYPE_TRACKS]);
#endif
}


static void
arr_gl_get_part_rect(Arrange* arrange, AMPart* part, DRect* b)
{
	// Return coords relative to song 0,0 - not current screen coords.
	// TODO This is a generic fn - should be moved elsewhere.

	GlCanvas* c = arrange->canvas->gl;
	ArrTrk* atr = track_list__lookup_track(arrange->priv->tracks, part->track);

	b->x1 = arr_gl_pos2px_(arrange, &part->start);
	b->y1 = atr->y * arrange->vzoom * (1-c->iso_factor) - c->actors[ACTOR_TYPE_MAIN]->scrollable.y1;
	b->x2 = b->x1 + arr_gl_pos2px_(arrange, &part->length);
	b->y2 = (atr->y + atr->height) * arrange->vzoom * (1-c->iso_factor) - c->actors[ACTOR_TYPE_MAIN]->scrollable.y1;
}


inline int
next_p2(int a)
{
	// Get The First Power Of 2 >= The Int That We Pass It.

	int rval=1;
	while(rval<a) rval<<=1;
	return rval;
}


typedef struct _gl_glyph
{
	int width;
	int height;
	unsigned char* buffer;

} glyph;


static void
gl_set_colour_from_part_bg(AMPart* part)
{
	float r, g, b;
	am_palette_get_float(song->palette, part->bg_colour, &r, &g, &b, 0xff);
	glColor4f(r, g, b, 1.0);
}


static void
arr_gl_on_iso_button_press(GtkWidget* widget, Arrange* arrangeX)
{
	PF;
	Arrange* arrange = (Arrange*)windows__get_panel_from_widget(widget);
	GlCanvas* c = arrange->canvas->gl;

	if(c->isometric) arr_gl_disable_isometric(arrange); else arr_gl_enable_isometric(arrange);
	gtk_widget_queue_draw(arrange->canvas->gl->image);
}


static void
arr_gl_on_iso_key_press(GtkAccelGroup* accel_group, gpointer user_data)
{
	PF;
	AyyiPanel* window = app->active_panel;
	char type[AYYI_SHORT_NAME_MAX]; ayyi_panel_print_type(window, type); dbg(0, "active: windowtype=%s", type);
	if(!AYYI_IS_ARRANGE(window)) return;
	Arrange* arrange = (Arrange*)window;
	GlCanvas* c = arrange->canvas->gl;

	if(c->isometric) arr_gl_disable_isometric(arrange); else arr_gl_enable_isometric(arrange);
	gtk_widget_queue_draw(arrange->canvas->gl->image);
}


static void
arr_gl_enable_isometric(Arrange* arrange)
{
	GlCanvas* c = arrange->canvas->gl;

	c->isometric = TRUE;
	agl_actor__add_child(c->actor, c->actors[ACTOR_TYPE_ISO_FRAME]);
	agl_actor__add_child(c->actor, c->actors[ACTOR_TYPE_ISO_TC]);
	rotation_start(arrange);
}


static void
arr_gl_disable_isometric(Arrange* arrange)
{
	GlCanvas* c = arrange->canvas->gl;

	c->isometric = FALSE;
	c->actor->children = g_list_remove(c->actor->children, c->actors[ACTOR_TYPE_ISO_FRAME]);
	c->actor->children = g_list_remove(c->actor->children, c->actors[ACTOR_TYPE_ISO_TC]);
	rotation_start(arrange);
}


static void
arr_gl_zoom_start(Arrange* arrange, AMChangeType changetype)
{
#ifdef DEBUG
	g_return_if_fail(CANVAS_IS_OPENGL);
#endif
	if(!((AyyiPanel*)arrange)->model_is_loaded) return;

	typedef struct {
		Arrange*     arrange;
		AMChangeType changetype;
	} C;

	GList* animatables = (changetype & AM_CHANGE_ZOOM_H) ? g_list_append(NULL, &CGL->zoom.h) : NULL;
	animatables = ((changetype & AM_CHANGE_ZOOM_V) && (CGL->zoom.v.val.f != *CGL->zoom.v.model_val.f)) ? g_list_append(animatables, &CGL->zoom.v) : animatables;

	// TODO if change is too small, dont animate. increase duration for large changes.
	/*
	float min_duration = 0.1;
	float max_duration = 1.0;
	*/
	if(changetype & AM_CHANGE_ZOOM_V){
	}

	if(changetype & AM_CHANGE_ZOOM_H){
		AGlActor* ruler = CGL->actors[ACTOR_TYPE_RULER];
		if(ruler){
#ifdef AGL_ACTOR_RENDER_CACHE
			CGL->actors[ACTOR_TYPE_RULER]->cache.enabled = false;
			CGL->actors[ACTOR_TYPE_RULER]->cache.valid = false;
#endif
		}
	}

	void redraw(Arrange* arrange, AMChangeType changetype) // TODO move (is generic)
	{
#ifdef USE_FBO
		if(CGL->actors[ACTOR_TYPE_TRACKS]) CGL->actors[ACTOR_TYPE_TRACKS]->invalidate(CGL->actors[ACTOR_TYPE_TRACKS]);
#endif
		g_hash_table_foreach(CGL->parts, __arr_gl_allocate_part, arrange);
		track_control_actor_on_view_change(arrange, changetype);
		gtk_widget_queue_draw(CGL->image);
	}

	void on_frame(WfAnimation* animation, int time)
	{
		C* c = animation->user_data;

		redraw(c->arrange, c->changetype);
	}

	void _on_animation_finished(WfAnimation* animation, gpointer user_data)
	{
		C* c = animation->user_data;
		Arrange* arrange = c->arrange;
		dbg(1, "%.3f-->%.3f", CGL->zoom.h.val.f, arrange->hzoom);

		redraw(arrange, c->changetype);

#ifdef AGL_ACTOR_RENDER_CACHE
						if(CGL->actors[ACTOR_TYPE_RULER]) // TODO ensure is always present
		CGL->actors[ACTOR_TYPE_RULER]->cache.enabled = true;
#endif

		g_free(c);
	}

	if(animatables){
		WfAnimation* animation = wf_animation_new(_on_animation_finished, AYYI_NEW(C,
			.arrange = arrange,
			.changetype = changetype
		)); // automatically freed when the animation finishes.
		animation->length = 250;
		animation->on_frame = on_frame;

		wf_transition_add_member(animation, animatables);
		wf_animation_start(animation);
	}else{
		redraw(arrange, changetype);
	}
}


static void
gl_set_palette_colour(int c_idx)
{
	float r = (song->palette[c_idx] & 0xff000000) >> 24;
	float g = (song->palette[c_idx] & 0x00ff0000) >> 16;
	float b = (song->palette[c_idx] & 0x0000ff00) >>  8;

	glColor3f(r / 0xff, g / 0xff, b / 0xff);
}


static void
arr_gl_do_follow(Arrange* arrange)
{
	GtkAdjustment* adj = arrange->canvas->h_adj;

	WfViewPort viewport;
	arr_gl_get_viewport(arrange, &viewport);
	double spp = arr_get_spp_px(arrange);
	if(spp < viewport.left){
		dbg(2, "paging left...");
		gtk_adjustment_set_value(adj, gtk_adjustment_get_value(adj) - gtk_adjustment_get_page_size(adj));
	}else if(spp > viewport.right){
		dbg(2, "paging right...");
		gtk_adjustment_set_value(adj, gtk_adjustment_get_value(adj) + gtk_adjustment_get_page_size(adj));
	}
}


static void
arr_gl_scroll_to_pos(Arrange* arrange, GPos* pos, TrackDispNum t)
{
	arr_gl_scroll_to(arrange, arr_pos2px(arrange, pos), t > -1 ? arrange->priv->tracks->track[t]->y * arrange->vzoom : -1);
}


static void
arr_gl_scroll_left(Arrange* arrange, int pix)
{
	arr_gl_scroll_to(arrange, MAX(0, -CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1 - pix), -1);
}


static void
arr_gl_scroll_right(Arrange* arrange, int pix)
{
	if(pix)
		arr_gl_scroll_to(arrange, -CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1 + pix, -1);
}


static void
arr_gl_scroll_up(Arrange* arrange, int pix)
{
	// Scroll up by the number of pixels given.
	// -the scrollbar should move upwards, but objects inside the scrollwin move downwards.

	arr_gl_scroll_to(arrange, -1, MAX(0, -CGL->actors[ACTOR_TYPE_MAIN]->scrollable.y1 - pix));
}


static void
arr_gl_scroll_down(Arrange* arrange, int pix)
{
	arr_gl_scroll_to(arrange, -1, -CGL->actors[ACTOR_TYPE_MAIN]->scrollable.y1 + pix);
}


static void
menu_arr_set_gl_canvas(Arrange* arrange)
{
	CanvasType arr_gl_canvas_get_type();
	arr_set_canvas(arrange, arr_gl_canvas_get_type());
}


#ifdef SHOW_LOG
static Actor*
log_actor(Arrange* arrange)
{
	static AGlFBO* fbo0 = NULL;

	#ifdef USE_FBO
	static void create_fbo()
	{
		AGlFBO* fbo = agl_fbo_new(256, 256, 0, 0);
		fbo0 = fbo;
		return;
	}
	#endif

	void log_actor__init(Actor* actor, Arrange* arrange)
	{
	#ifdef USE_FBO
		if(!fbo0) create_fbo();
	#endif
	}

	void log_actor__paint(Actor* actor, Arrange* arrange)
	{
		GlCanvas* c = (GlCanvas*)arrange->canvas->gl;

		GdkColor bg;
		get_style_bg_color(&bg, GTK_STATE_NORMAL);
		uint32_t colour = colour_get_offset_rgba(color_gdk_to_rgba(&bg), 10);

		GtkTextBuffer* log = ayyi.log.txtbuf;
		int n_lines = gtk_text_buffer_get_line_count(log);

	#if 0 // draw all text in one go -- is this more efficient than doing separate operations per line?

		int n_visible_lines = 22;
		GtkTextIter start, end;
		gtk_text_buffer_get_iter_at_line(log, &start, MAX(0, n_lines - n_visible_lines));
		gtk_text_buffer_get_end_iter(log, &end);
		gchar* text = gtk_text_buffer_get_text(log, &start, &end, FALSE);

		agl_draw_to_fbo(fbo0) {
			glClearColor(1.0, 1.0, 1.0, 0.0); //background colour must be same as foreground for correct antialiasing
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
			agl_print(10, 10, 0, 0xffffffff, text);
		} agl_end_draw_to_fbo;
		g_free(text);

		#ifdef USE_FBO
		// put the log fbo onto the screen
		{
			glColor4f(1.0, 1.0, 1.0, 0.10);
			agl_enable(AGL_ENABLE_TEXTURE_2D | AGL_ENABLE_BLEND);
			glBindTexture(GL_TEXTURE_2D, fbo0->texture);
			glBegin(GL_QUADS);
			double log_scale = 2;
			double top = c->image->allocation.height - fbo0->height * log_scale;
			double bot = c->image->allocation.height;
			double x1 = 0;
			double x2 = fbo0->width * log_scale;
			glTexCoord2d(1.0, 1.0); glVertex2d(x2, top);
			glTexCoord2d(0.0, 1.0); glVertex2d(x1, top);
			glTexCoord2d(0.0, 0.0); glVertex2d(x1, bot);
			glTexCoord2d(1.0, 0.0); glVertex2d(x2, bot);
			glEnd();
		}
		#endif
	#else

		void fbo_print(int x, int y, double scale, uint32_t colour, int alpha, const char* text)
		{
			#ifdef USE_FBO
			agl_draw_to_fbo(fbo0) {
				WfColourFloat fc; colour_rgba_to_float(&fc, colour);

				glClearColor(fc.r, fc.g, fc.b, 0.0); //background colour must be same as foreground for correct antialiasing
				glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
				agl_print(0, 0, 0.0, colour, text);

			} agl_end_draw_to_fbo;

			// put the log fbo onto the screen
			{
				glColor4f(1.0, 1.0, 1.0, alpha / 256.0);
				agl_enable(AGL_ENABLE_TEXTURE_2D | AGL_ENABLE_BLEND);
				glBindTexture(GL_TEXTURE_2D, fbo0->texture);
				glBegin(GL_QUADS);
				double top = y;
				double bot = y + fbo0->height * scale;
				double x1 = x;
				double x2 = x + fbo0->width * scale;
				glTexCoord2d(1.0, 1.0); glVertex2d(x2, top);
				glTexCoord2d(0.0, 1.0); glVertex2d(x1, top);
				glTexCoord2d(0.0, 0.0); glVertex2d(x1, bot);
				glTexCoord2d(1.0, 0.0); glVertex2d(x2, bot);
				glEnd();
			}
			#endif
		}

		int alpha = 0x7f;
		double scale = 1.5;
		int line_height = 14; //TODO
		int l; for(l=0;l<MIN(20, n_lines - 1);l++){
			GtkTextIter l1, l2;
			gtk_text_buffer_get_iter_at_line(log, &l1, n_lines -l -2);
			gtk_text_buffer_get_iter_at_line(log, &l2, n_lines -l -1);
			gchar* line = gtk_text_buffer_get_text(log, &l1, &l2, FALSE);
			int y = c->image->allocation.height - 70 - l * line_height * scale;
			if(y < 0) break;
			fbo_print(arrange->tc_width.val.i + 10, y, scale, colour, alpha = alpha - 7, line);
			g_free(line);
		}
	#endif // draw all lines in one go
	}

	return AGL_NEW(Actor,
		.name = "Log"
		.init = log_actor__init,
		.paint = log_actor__paint,
	);
}
#endif


static GlPart*
_add_new_gl_part(Arrange* arrange, AMPart* part)
{
	PF;
	GlCanvas* c = arrange->canvas->gl;

	g_return_val_if_fail(!g_hash_table_lookup(c->parts, part), NULL);

	GlPart* glpart = g_new0(GlPart, 1); // freed automatically when the hashtable is destroyed.
	glpart->part = part;
	glpart->arrange = arrange;

	if(c->wfc && PART_IS_AUDIO(part)) add_audio_part_actor_to_wf_canvas(c, glpart, part);

	g_hash_table_insert(c->parts, part, glpart);

	if(part->pool_item){
		arr_gl_allocate_part(arrange, glpart);

		void _glpart_on_peakdata_available(Waveform* waveform, int block, gpointer _gl_part)
		{
			PF2;
			GlPart* gp = _gl_part;
			g_return_if_fail(windows__verify_pointer((AyyiPanel*)gp->arrange, AYYI_TYPE_ARRANGE));
			GlCanvas* c = gp->arrange->canvas->gl;
			if(arr_part_is_in_viewport(gp->arrange, gp->part)){
				// this is the correct actor to invalidate
				agl_actor__invalidate((AGlActor*)gp->wf_actor);
				// but for now we need to invalidate the track because the parts are not properly parented. TODO
				agl_actor__invalidate(c->actors[ACTOR_TYPE_TRACKS]);
			}
		}

		g_signal_connect (part->pool_item, "peakdata-ready", (GCallback)_glpart_on_peakdata_available, glpart);
		g_signal_connect (part->pool_item, "hires-ready", (GCallback)_glpart_on_peakdata_available, glpart);
	}

	return glpart;
}


static void
_remove_gl_part(GlCanvas* c, GlPart* glpart)
{
	// g_free is called on the part by g_hash_table_destroy or g_hash_table_remove.
	// -it would be nice to pass this fn instead of g_free to g_hash_table_new, but we need the GlCanvas context.

	PoolItem* pi = glpart->part->pool_item;
	if(pi){
		signal_handler_disconnect_data(pi, glpart, 2, "glpart");
		wf_canvas_remove_actor(c->wfc, glpart->wf_actor);
	}
}


static GlTrack*
_add_new_gl_track(Arrange* arrange, AMTrack* tr)
{
	GlCanvas* c = arrange->canvas->gl;

	GlTrack* gltrack = g_hash_table_lookup(c->tracks, tr);
	g_return_val_if_fail(!gltrack, NULL);

	gltrack = g_new0(GlTrack, 1); // freed automatically when the hashtable is destroyed.
	g_hash_table_insert(c->tracks, tr, gltrack);

#if 0 // the fbo is now created on demand (if track is empty, it doesnt need an fbo)
	int height = GL_TRACK_HEIGHT(tr) + 2 * PART_TRANSITION_BORDER;
	gltrack->cache.fbo = agl_fbo_new(512, height, 0, AGL_FBO_HAS_STENCIL);
#endif

	return gltrack;
}


static void
__arr_gl_remove_glpart(gpointer key, gpointer value, gpointer _arrange)
{
	PF;
	GlCanvas* c = ((Arrange*)_arrange)->canvas->gl;

	_remove_gl_part(c, (GlPart*)value);
}


static void
arr_gl_set_spp(Arrange* arrange, uint32_t samples)
{
	gtk_widget_queue_draw(arrange->canvas->gl->image);
}


static void
arr_gl_set_tc_width(Arrange* arrange, int width)
{
	GlCanvas* c = arrange->canvas->gl;

	arrange->tc_width.val.i = width;
	c->actors[ACTOR_TYPE_TC]->region.x2 = width;
	c->actors[ACTOR_TYPE_MAIN]->region.x1 = width;
	agl_actor__set_size(c->actor);

	gtk_widget_queue_draw(CGL->image);
}


double
arr_gl_get_spp_px(Arrange* arrange)
{
	GPos pos;
	am_transport_get_pos(&pos);
	return arr_gl_pos2px(arrange, &pos);
}


static FeatureType
arr_gl_pick_feature(CanvasOp* op)
{
	// Note that at present this is not canvas specific and is essentially the same as the implementations for the other canvas types.

	Arrange* arrange = op->arrange;

	ArrTrk* at = track_list__track_by_index(op->arrange->priv->tracks, op->part->track->track_num);
	ArrTrackMidi* atm = (ArrTrackMidi*)at;

	int part_height = at->height * arrange->vzoom;
	int vchart_height = atm->velocity_chart_height;
	if(op->diff.y < part_height - vchart_height){
		dbg(2, "note area... diff=%.2f", op->diff.y);
		gboolean at_start, at_end;
		if(arr_part_pick_note(arrange, op->part, op->diff, &at_start, &at_end)){
			dbg(2, "over note! at_start=%i at_end=%i", at_start, at_end);
			if(at_start){
				return FEATURE_NOTE_LEFT;
			} else if(at_end){
				return FEATURE_NOTE_RIGHT;
			} else {
				return FEATURE_NOTE;
			}
		}
		else {
			dbg(2, "not note. diff=%.2fx%.2f px1=%.2f", op->diff.x, op->diff.y, op->p1.x);
			return FEATURE_NOTE_AREA;
		}
	}else if(op->diff.y < part_height - vchart_height + 2){
		dbg(2, "FEATURE_VEL_DIVIDER diff=%.2f", op->diff.y);
		return FEATURE_VEL_DIVIDER;
	}else{
		dbg(2, "velocity zone...");
																			#if 0
		gboolean at_top = FALSE;
		Ptd p = {op->diff.x, op->diff.y - (part_height - vchart_height)};
		if(gnome_canvas_part_pick_velocity_bar(GNOME_CANVAS_PART(((GnomeCanvasOp*)op)->item), p, &at_top)){
			if(at_top){
				return FEATURE_VEL_BAR_TOP;
			}
		}
																			#endif
		return FEATURE_VEL_AREA;
	}

	return FEATURE_NONE;
}


/* temporary */
#ifdef DEBUG
int
arr_gl_parts_count(Arrange* arrange)
{
	GlCanvas* c = arrange->canvas->gl;
	return g_hash_table_size(c->parts);
}
#endif


#ifdef NOT_USED
static void
fbo_2_png(AGlFBO* fbo)
{
	int width = agl_power_of_two(fbo->width);
	int height = agl_power_of_two(fbo->height);

	unsigned char buffer[width * height * 4];
	glBindTexture(GL_TEXTURE_2D, fbo->texture);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, buffer);

	GdkPixbuf* pixbuf = gdk_pixbuf_new_from_data(buffer, GDK_COLORSPACE_RGB, HAS_ALPHA_TRUE, 8, width, height, width*4, NULL, NULL);
	gdk_pixbuf_save(pixbuf, "test.png", "png", NULL, NULL);
	g_object_unref(pixbuf);
}
#endif


