/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2013-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

AGlActor*
arr_gl_grid(GtkWidget* panel)
{
	Arrange* arrange = (Arrange*)panel;

	bool arr_gl_grid_draw(AGlActor* actor)
	{
		Arrange* arrange = actor->root->user_data;

		double vp_left = CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1;
		double width = actor->region.x2 - actor->region.x1 - RR;
		double top = CGL->actors[ACTOR_TYPE_MAIN]->scrollable.y1;
		double height = actor->region.y2 + 48;
		double song_end = arr_gl_pos2px_(arrange, &song->loc[9].vals[0].val.sp);
		double stop = MIN(width - 2.0, song_end);

		// vertical grid lines
		dotted_shader.uniform.colour = 0xaaaaaa66;
		agl_use_program((AGlShader*)&dotted_shader);

		int q = agl_power_of_two(24.0 / arrange->hzoom);
		GPos increment = {q, 0, 0};
		GPos pos = {((int)arr_px2beat(arrange, vp_left)) / q * q,};
		double x;
		while((x = arr_gl_pos2px(arrange, &pos)) < stop){
			agl_rect(x, top, 1.0, height);

			pos_add(&pos, &increment);
		}

		// locators
		dotted_shader.uniform.colour = 0x6677ff66;
		agl_use_program((AGlShader*)&dotted_shader);
		int i; for(i=1;i<AYYI_MAX_LOC - 1;i++){
			if(song->loc[i].vals[0].val.sp.beat > -1){
				agl_rect(arr_gl_pos2px_(arrange, &song->loc[i].vals[0].val.sp), top, 1.0, height);
			}
		}

		// grey out after song end
		agl->shaders.plain->uniform.colour = 0xaaaaaa2f;
		agl_use_program((AGlShader*)agl->shaders.plain);
		agl_rect(song_end, top - 40, 2048.0, height);

		return true;
	}

	void arr_gl_grid_set_size(AGlActor* actor)
	{
		actor->region = (AGliRegion){
			.x1 = 0,
			.y1 = 0,
			.x2 = ((AGlActor*)actor->root)->region.x2,
			.y2 = ((AGlActor*)actor->root)->region.y2,
		};
	}

	void arr_gl_grid_init(AGlActor* actor)
	{
		actor->paint = agl->use_shaders ? arr_gl_grid_draw : agl_actor__null_painter;

		if(agl->use_shaders && !dotted_shader.shader.program){
			agl_create_program(&dotted_shader.shader);
			dotted_shader.uniform.colour = 0xaaaaaa44;
		}
	}

	return AGL_NEW(AGlActor,
		.name     = "Grid",
		.init     = arr_gl_grid_init,
		.region   = (AGliRegion){arrange->tc_width.val.i, 0, 4096, arrange->ruler_height},
		.set_size = arr_gl_grid_set_size
	);
}

