/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://ayyi.org               |
* | copyright (C) 2010-2016 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#define __gl_canvas_priv__
#include "global.h"
#include <stdio.h>
#include <string.h>
#include <GL/gl.h>
#include <gtk/gtk.h>
#include "agl/utils.h"
#include "agl/ext.h"
#include "agl/fbo.h"
#include "waveform/actor.h"
#include "model/ayyi_model.h"
#include "arrange.h"
#include "arrange/gl_canvas.h"
#include "arrange/shader.h"
#include "arrange/gl_meter.h"

extern AGlShaderText meter_text;
static void meter_set_uniforms();
static void meter_set_state(AGlActor*);

typedef struct {
	AGlShader    shader;
	struct {
		float    width;
		float    height;
		float    level;
		float    peak;
		float    peak2;
	}            uniform;
} MeterShader;

static MeterShader meter_shader = {{NULL, NULL, 0, agl_null_uniforms, meter_set_uniforms, &meter_text}};

static AGl* agl = NULL;


static void
meter_set_uniforms()
{
	AGlShader* shader = &meter_shader.shader;

	glUniform1f(glGetUniformLocation(shader->program, "width"), ((MeterShader*)shader)->uniform.width);
	glUniform1f(glGetUniformLocation(shader->program, "peak"), ((MeterShader*)shader)->uniform.peak);
}


static void
meter_set_size(AGlActor* actor)
{
	Arrange* arrange = actor->root->user_data;

	meter_shader.uniform.width = 4.0f;

	actor->region = (AGliRegion){
		.x1 = 0,
		.y1 = arrange->ruler_height,
		.x2 = meter_shader.uniform.width,
		.y2 = ((AGlActor*)actor->root)->region.y2,
	};
}


static void
meter_set_state(AGlActor* actor)
{
	AGlShader* shader = &meter_shader.shader;

	meter_shader.uniform.peak = 0.0;

	agl_use_program(shader);

	gl_warn("");
}


AGlActor*
meter_actor_new(GtkWidget* panel)
{
	Arrange* arrange = (Arrange*)panel;

	agl = agl_get_instance();

	float db_xlate(float db) { return MAX(0.0, (db + 70.0) / 70.0); }

	bool meter_paint(AGlActor* actor)
	{
		Arrange* arrange = actor->root->user_data;
		TrackList* tl = arrange->priv->tracks;

		AMTrack* track = NULL;
		ArrTrk* atr = NULL;
		while((atr = track_list__next_visible(tl, atr ? atr->track : NULL)) && (track = atr->track)){
			float height = atr->height * arrange->vzoom;
			if(agl->use_shaders){
				AGlShader* shader = &meter_shader.shader;

				meter_shader.uniform.height = height;
				meter_shader.uniform.level = db_xlate(am_track__get_meterlevel(track));
				meter_shader.uniform.peak2 = height * atr->peak;

				//levels.max[i] = MAX(levels.levels[i], levels.max[i]);
				glUniform1f(glGetUniformLocation(shader->program, "height"), ((MeterShader*)shader)->uniform.height);
				glUniform1f(glGetUniformLocation(shader->program, "level"), ((MeterShader*)shader)->uniform.level);
				glUniform1f(glGetUniformLocation(shader->program, "peak2"), ((MeterShader*)shader)->uniform.peak2);
			}

			glPushMatrix();
			glScalef(1.0, -1.0, 1.0);
			glTranslatef(0.0, -((atr->y + atr->height) * arrange->vzoom), 0.0);
			agl_rect(/*(100.0 + 3.0 * 16.0) + */1.0, 0.0, meter_shader.uniform.width, height);
			glPopMatrix();
		}
		return true;
	}

	AGlActor* a = AGL_NEW(AGlActor,
		.name = "Meter",
		.region.y1 = arrange->ruler_height,
		.set_size = meter_set_size,
		.set_state = agl->use_shaders ? meter_set_state : NULL,
		.paint = meter_paint
	);

	if(agl->use_shaders){
		agl_create_program(&meter_shader.shader);
		gl_warn("meter create");
		dbg(1, "program=%i", meter_shader.shader.program);
		meter_shader.uniform.width = 2.0;
	}

	return a;
}


