/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2013-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

#define R 2
#define Rf 2.0
#define SCROLLBAR_H_PADDING 2
#define SCROLLBAR_V_PADDING 3

static void        arr_gl_scrollbar_set_size      (AGlActor*);
static bool        arr_gl_scrollbar_on_event      (AGlActor*, GdkEvent*, AGliPt);
static void        arr_gl_hscrollbar_bar_position (AGlActor*, iRange*);
static void        arr_gl_vscrollbar_bar_position (AGlActor*, iRange*);


typedef struct {
    AGlActor       actor;
    GtkOrientation orientation;
    AGliPt         grab_offset; // TODO this should probably be part of the actor_context
    float          opacity;
    WfAnimatable   animation;
} ScrollbarActor;

static struct Press {
    AGliPt     pt;
    AGliRegion viewport;
} press = {{0,}};


AGlActor*
arr_gl_scrollbar(GtkWidget* panel, GtkOrientation orientation)
{
	bool arr_gl_scrollbar_draw_gl1(AGlActor* actor)
	{
		return true;
	}

	bool arr_gl_scrollbar_draw_h(AGlActor* actor)
	{
		iRange bar;
		arr_gl_hscrollbar_bar_position(actor, &bar);

		if(!actor->disabled){
			if(((ScrollbarActor*)actor)->animation.val.f > 0.61){
				agl->shaders.plain->uniform.colour = 0xffffff66;
				agl_use_program((AGlShader*)agl->shaders.plain);
				agl_rect(0, 1, agl_actor__width(actor), agl_actor__height(actor) - 1);
			}

			h_scrollbar_shader.uniform.colour = (get_style_bg_color_rgba(GTK_STATE_SELECTED) & 0xffffff00) + (int)(((ScrollbarActor*)actor)->animation.val.f * 0xff);
			h_scrollbar_shader.uniform.bg_colour = v_scrollbar_shader.uniform.colour & 0xffffff00;
			h_scrollbar_shader.uniform.centre1 = (AGliPt){bar.start + 4, 6};
			h_scrollbar_shader.uniform.centre2 = (AGliPt){bar.end   - 4, 6};
			agl_use_program((AGlShader*)&h_scrollbar_shader);

			agl_rect(SCROLLBAR_H_PADDING + bar.start, SCROLLBAR_V_PADDING, bar.end - bar.start, 11);
		}
		return true;
	}

	bool arr_gl_scrollbar_draw_v(AGlActor* actor)
	{
		if(!actor->disabled){
			iRange bar;
			arr_gl_vscrollbar_bar_position(actor, &bar);

			//v_scrollbar_shader.uniform.colour = (get_style_bg_color_rgba(GTK_STATE_SELECTED) & 0xffffff00) + (agl_actor__is_hovered(actor) ? 0xff : 0x8f);
			v_scrollbar_shader.uniform.colour = (get_style_bg_color_rgba(GTK_STATE_SELECTED) & 0xffffff00) + (int)(((ScrollbarActor*)actor)->animation.val.f * 0xff);
			v_scrollbar_shader.uniform.bg_colour = v_scrollbar_shader.uniform.colour & 0xffffff00;
			v_scrollbar_shader.uniform.centre1 = (AGliPt){6, bar.start + 4};
			v_scrollbar_shader.uniform.centre2 = (AGliPt){6, bar.end - 4};
			agl_use_program((AGlShader*)&v_scrollbar_shader);

			agl_rect(SCROLLBAR_V_PADDING, bar.start, 10, bar.end - bar.start);
		}
		return true;
	}

	void arr_gl_scrollbar_init(AGlActor* actor)
	{
		if(((ScrollbarActor*)actor)->orientation == GTK_ORIENTATION_VERTICAL){
			if(!v_scrollbar_shader.shader.program){
				agl_create_program(&v_scrollbar_shader.shader);
				v_scrollbar_shader.uniform.radius = 3;
			}

			actor->paint = agl->use_shaders ? arr_gl_scrollbar_draw_v : agl_actor__null_painter;
		}else{
			if(!h_scrollbar_shader.shader.program){
				agl_create_program(&h_scrollbar_shader.shader);
				h_scrollbar_shader.uniform.radius = 3;
			}
			actor->paint = agl->use_shaders ? arr_gl_scrollbar_draw_h : arr_gl_scrollbar_draw_gl1;
		}
	}

	ScrollbarActor* scrollbar = AGL_NEW(ScrollbarActor,
		.actor = {
			.name = "Scrollbar",
			.init = arr_gl_scrollbar_init,
			.set_size = arr_gl_scrollbar_set_size,
			.paint = agl_actor__null_painter,
			.on_event = arr_gl_scrollbar_on_event,
			.region = (AGliRegion){0, 0, 1, 1},
		},
		.orientation = orientation,
		// TODO fade in and out when leaving/entering window
		.opacity = 0.5,
		.animation = {
			.start_val.f = 0.6,
			.val.f       = 0.6,
			.type        = WF_FLOAT
		}
	);
	scrollbar->animation.model_val.f = &scrollbar->opacity;

	return (AGlActor*)scrollbar;
}


static void
arr_gl_scrollbar_set_size(AGlActor* actor)
{
	Arrange* arrange = actor->root->user_data;
	AGlActor* root = (AGlActor*)actor->root;

	if(root->region.x2 > 0 && actor->region.x2 > 0){
		if(((ScrollbarActor*)actor)->orientation == GTK_ORIENTATION_VERTICAL){
			actor->region = (AGliRegion){
				.x1 = root->region.x2 - SCROLLBAR_H_PADDING - 2 * R - 5,
				.x2 = root->region.x2 - SCROLLBAR_H_PADDING,
				.y1 = SCROLLBAR_H_PADDING,
				.y2 = agl_actor__height(root) - SCROLLBAR_H_PADDING
			};
		}else{
			int offset = SHOW_OVERVIEW ? -OVERVIEW_HEIGHT : 0;

			actor->region = (AGliRegion){
				.x1 = 0,
				.x2 = root->region.x2,
				.y1 = root->region.y2 - SCROLLBAR_V_PADDING - 2 * R - 5 + offset,
				.y2 = root->region.y2                                   + offset
			};
		}
	}
}


static void
arr_gl_hscrollbar_bar_position (AGlActor* actor, iRange* pos)
{
	Arrange* arrange = actor->root->user_data;

	double songlen = arr_gl_pos2px_(arrange, &am_object_val(&song->loc[AM_LOC_END]).sp) + 20;
	double vp_width = agl_actor__width(((AGlActor*)actor->root));
	double pct = vp_width / songlen;
	int len = MAX(28, pct * agl_actor__width(actor));

	if(pct < 1.0){
		pos->start = 1 - CGL->actors[ACTOR_TYPE_MAIN]->scrollable.x1 * pct;
		pos->end = pos->start + len;
	}else{
		*pos = (iRange){0,}; // no scrollbar needed
	}
}


static void
arr_gl_vscrollbar_bar_position (AGlActor* actor, iRange* pos)
{
	Arrange* arrange = actor->root->user_data;
	GtkAdjustment* adj = arrange->canvas->v_adj;

	double scaling = agl_actor__height(actor) / adj->upper;
	if(scaling < 1.0){
		pos->start = 1 - CGL->actors[ACTOR_TYPE_MAIN]->scrollable.y1 * scaling;
		pos->end = pos->start + adj->page_size * scaling;
	}else{
		*pos = (iRange){0,}; // no scrollbar needed
	}
}


static bool
arr_gl_scrollbar_on_event(AGlActor* actor, GdkEvent* event, AGliPt xy)
{
	void animation_done (WfAnimation* animation, gpointer user_data)
	{
	}

	Arrange* arrange = actor->root->user_data;
	ScrollbarActor* scrollbar = (ScrollbarActor*)actor;

	int x = xy.x;

	bool handled = false;

	switch (event->type){
		case GDK_MOTION_NOTIFY:
			if(actor_context.grabbed == actor){
				if(((ScrollbarActor*)actor)->orientation == GTK_ORIENTATION_VERTICAL){
					int dy = xy.y - press.pt.y;
					double scale = arrange->canvas->v_adj->upper / (double)agl_actor__height(actor);
					arrange->canvas->scroll_to(arrange, -1, ((int)-press.viewport.y1) + dy * scale);
				}else{
					// would it simplify to use h_adj here?
					int dx = x - press.pt.x;
					double vp_size = arr_gl_pos2px_(arrange, &am_object_val(&song->loc[AM_LOC_END]).sp) + 20.0;
					double scale = vp_size / (double)agl_actor__width(actor);
					arrange->canvas->scroll_to(arrange, ((int)-press.viewport.x1) + dx * scale, -1);
				}
			}
			break;
		case GDK_ENTER_NOTIFY:
			scrollbar->opacity = 1.0;
			agl_actor__start_transition(actor, g_list_append(NULL, &scrollbar->animation), animation_done, NULL);
			break;
		case GDK_LEAVE_NOTIFY:
			scrollbar->opacity = 0.6;
			agl_actor__start_transition(actor, g_list_append(NULL, &scrollbar->animation), animation_done, NULL);
			break;
		case GDK_BUTTON_PRESS:
			{
				press = (struct Press){.pt = xy, .viewport = CGL->actors[ACTOR_TYPE_MAIN]->scrollable};

				iRange bar;
				if(((ScrollbarActor*)actor)->orientation == GTK_ORIENTATION_VERTICAL){
					arr_gl_vscrollbar_bar_position (actor, &bar);
					if(xy.y > bar.start && xy.y < bar.end){
						actor_context.grabbed = actor;
						((ScrollbarActor*)actor)->grab_offset = (AGliPt){x - bar.start, xy.y - bar.start};
						handled = true;
					}
				}else{
					arr_gl_hscrollbar_bar_position (actor, &bar);
					if(x > bar.start && x < bar.end){
						actor_context.grabbed = actor;
						((ScrollbarActor*)actor)->grab_offset = (AGliPt){x - bar.start, xy.y - bar.start};
						handled = true;
					}
				}
			}
			break;
		case GDK_BUTTON_RELEASE:
			{
				actor_context.grabbed = NULL;
				handled = true;
			}
			break;
		default:
			break;
	}
	return handled;
}


static void
arr_gl_scrollbar_on_view_change(AGlActor* actor, AMChangeType change)
{
	if(change & (AM_CHANGE_POS_X | AM_CHANGE_VIEWPORT_Y)){

		arr_gl_scrollbar_set_size(actor);
	}
}


