/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2015-2016 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include "global.h"
#include <GL/gl.h>
#include "agl/ext.h"
#include "waveform/utils.h"
#include "waveform/waveform.h"
#include "arrange/shader.h"
#include "arrange/shaders/shaders.c"

static void _part_set_uniforms();
static void _pattern_set_uniforms();
static void _dotted_set_uniforms();
static void _circle_set_uniforms();
static void _h_scrollbar_set_uniforms();
static void _v_scrollbar_set_uniforms();


static AGlUniformInfo uniforms1[] = {
   {"width", 1, GL_FLOAT, {0,}, -1},
   {"height", 1, GL_FLOAT, {0,}, -1},
   END_OF_UNIFORMS
};
PartShader part_shader = {{NULL, NULL, 0, uniforms1, _part_set_uniforms, &part_text}};

PatternShader pattern_shader = {{NULL, NULL, 0, NULL, _pattern_set_uniforms, &pattern_text}, {0}, {END_OF_UNIFORMS}};
DottedShader dotted_shader = {{NULL, NULL, 0, NULL, _dotted_set_uniforms, &dotted_text}, {0}, {END_OF_UNIFORMS}};
CircleShader circle_shader = {{NULL, NULL, 0, NULL, _circle_set_uniforms, &circle_text}, {}, {END_OF_UNIFORMS}};
ScrollbarShader h_scrollbar_shader = {{NULL, NULL, 0, NULL, _h_scrollbar_set_uniforms, &hscrollbar_text}, {}, {END_OF_UNIFORMS}};
ScrollbarShader v_scrollbar_shader = {{NULL, NULL, 0, NULL, _v_scrollbar_set_uniforms, &vscrollbar_text}, {}, {END_OF_UNIFORMS}};


static inline void
set_uniform_f(AGlShader* shader, int u, float* prev)
{
	AGlUniformInfo* uniform = &shader->uniforms[u];
	if(uniform->value[0] != *prev){
		glUniform1f(uniform->location, uniform->value[0]);
		*prev = uniform->value[0];
	}
}


static void
_part_set_uniforms()
{
	AGlShader* shader = &part_shader.shader;

	static float prev[PART_SHADER_MAX] = {0,};

	float colour[4] = {0.0, 0.0, 0.0, ((float)(part_shader.uniform.colour & 0xff)) / 0x100};
	agl_rgba_to_float(part_shader.uniform.colour, &colour[0], &colour[1], &colour[2]);
	glUniform4fv(glGetUniformLocation(part_shader.shader.program, "colour"), 1, colour);

	float border_colour[4] = {0.0, 0.0, 0.0, ((float)(part_shader.uniform.border_colour & 0xff)) / 0x100};
	agl_rgba_to_float(part_shader.uniform.border_colour, &border_colour[0], &border_colour[1], &border_colour[2]);
	glUniform4fv(glGetUniformLocation(part_shader.shader.program, "border_colour"), 1, border_colour);

	glUniform1f(glGetUniformLocation(shader->program, "selected"), ((PartShader*)shader)->uniform.selected);
	glUniform1f(glGetUniformLocation(shader->program, "frame"), ((PartShader*)shader)->uniform.frame);
	int i; for(i=0;i<PART_SHADER_MAX;i++){
		set_uniform_f(shader, i, &prev[i]);
	}

	glUniformMatrix4fv(glGetUniformLocation(shader->program, "modelview_matrix"), 1, GL_FALSE, part_shader.uniform.modelview);
	glUniformMatrix4fv(glGetUniformLocation(shader->program, "projection_matrix"), 1, GL_FALSE, part_shader.uniform.projection);
}


static void
_pattern_set_uniforms()
{
}


static void
_dotted_set_uniforms()
{
	float colour[4] = {0.0, 0.0, 0.0, ((float)(dotted_shader.uniform.colour & 0xff)) / 0x100};
	agl_rgba_to_float(dotted_shader.uniform.colour, &colour[0], &colour[1], &colour[2]);
	glUniform4fv(glGetUniformLocation(dotted_shader.shader.program, "colour"), 1, colour);
}

#define SET_COLOUR(SHADER) \
	float colour[4] = {0.0, 0.0, 0.0, ((float)((SHADER)->uniform.colour & 0xff)) / 0x100}; \
	agl_rgba_to_float((SHADER)->uniform.colour, &colour[0], &colour[1], &colour[2]); \
	glUniform4fv(glGetUniformLocation((SHADER)->shader.program, "colour"), 1, colour);
	
static void
_circle_set_uniforms()
{
	float centre[2] = {circle_shader.uniform.centre.x, circle_shader.uniform.centre.y};
	glUniform2fv(glGetUniformLocation(circle_shader.shader.program, "centre"), 1, centre);
	glUniform1f(glGetUniformLocation(circle_shader.shader.program, "radius"), circle_shader.uniform.radius);

	SET_COLOUR(&circle_shader);

	GLint location = 0;
	if(!location){
		location = glGetUniformLocation(circle_shader.shader.program, "bg_colour");
	}
	float bg_colour[4] = {0.0, 0.0, 0.0, ((float)(circle_shader.uniform.bg_colour & 0xff)) / 0x100};
	agl_rgba_to_float(circle_shader.uniform.bg_colour, &bg_colour[0], &bg_colour[1], &bg_colour[2]);
	glUniform4fv(location, 1, bg_colour);
}


static void
_h_scrollbar_set_uniforms()
{
	float centre1[2] = {h_scrollbar_shader.uniform.centre1.x, h_scrollbar_shader.uniform.centre1.y};
	float centre2[2] = {h_scrollbar_shader.uniform.centre2.x, h_scrollbar_shader.uniform.centre2.y};
	glUniform2fv(glGetUniformLocation(h_scrollbar_shader.shader.program, "centre1"), 1, centre1);
	glUniform2fv(glGetUniformLocation(h_scrollbar_shader.shader.program, "centre2"), 1, centre2);
	glUniform1f(glGetUniformLocation(h_scrollbar_shader.shader.program, "radius"), h_scrollbar_shader.uniform.radius);

	SET_COLOUR(&h_scrollbar_shader);

	GLint location = 0;
	if(!location){
		location = glGetUniformLocation(h_scrollbar_shader.shader.program, "bg_colour");
	}
	float bg_colour[4] = {0.0, 0.0, 0.0, ((float)(h_scrollbar_shader.uniform.bg_colour & 0xff)) / 0x100};
	agl_rgba_to_float(h_scrollbar_shader.uniform.bg_colour, &bg_colour[0], &bg_colour[1], &bg_colour[2]);
	glUniform4fv(location, 1, bg_colour);
}


static void
_scrollbar_set_uniforms(ScrollbarShader* scrollbar_shader)
{
	float centre1[2] = {scrollbar_shader->uniform.centre1.x, scrollbar_shader->uniform.centre1.y};
	float centre2[2] = {scrollbar_shader->uniform.centre2.x, scrollbar_shader->uniform.centre2.y};
	glUniform2fv(glGetUniformLocation(scrollbar_shader->shader.program, "centre1"), 1, centre1);
	glUniform2fv(glGetUniformLocation(scrollbar_shader->shader.program, "centre2"), 1, centre2);
	glUniform1f(glGetUniformLocation(scrollbar_shader->shader.program, "radius"), scrollbar_shader->uniform.radius);

	SET_COLOUR(scrollbar_shader);

	GLint location = 0;
	if(!location){
		location = glGetUniformLocation(scrollbar_shader->shader.program, "bg_colour");
	}
	float bg_colour[4] = {0.0, 0.0, 0.0, ((float)(scrollbar_shader->uniform.bg_colour & 0xff)) / 0x100};
	agl_rgba_to_float(scrollbar_shader->uniform.bg_colour, &bg_colour[0], &bg_colour[1], &bg_colour[2]);
	glUniform4fv(location, 1, bg_colour);
}


static void
_v_scrollbar_set_uniforms()
{
	_scrollbar_set_uniforms(&v_scrollbar_shader);
}


