/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2014-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#define __gl_mixer_c__

#include "global.h"
#include <GL/glx.h>
#include <gtk/gtkgl.h>
#include "agl/ext.h"
#include "agl/utils.h"
#include "agl/shader.h"
#include "agl/actor.h"
#include "support.h"
#include "gl_mixer.h"

static AGl* agl = NULL;

#include "mixer/shaders/shaders.c"
#include "mixer/gl_strip.c"

#define DIRECT true

static GObject*  gl_mixer__constructor                 (GType, guint n_construct_properties, GObjectConstructParam*);
static void      gl_mixer__add_accels                  (AyyiGlMixerWinClass*);
static void      gl_mixer__on_ready                    (AyyiPanel*);
static void      gl_mixer__on_song_load                (GObject*, gpointer);
static void      gl_mixer__on_song_unload              (GObject*, gpointer);
static void      gl_mixer__on_channel_add              (GObject*, AMChannel*, GlMixerWin*);
static void      gl_mixer__on_channel_change           (GObject*, gpointer);
static void      gl_mixer__on_channel_selection_change (GObject*, AyyiPanel* sender, gpointer);
static void      gl_mixer__destroy                     (GtkObject*);
static void      gl_mixer__on_resize                   (GtkWidget*, GtkAllocation*);
static void      gl_mixer__on_realise                  (GtkWidget*, gpointer);
static gint      gl_mixer__on_event                    (GtkWidget*, GdkEvent*, gpointer);
static bool      gl_mixer__on_expose                   (GtkWidget*, GdkEventExpose*, gpointer);
static void      gl_mixer__on_channel_delete           (GObject*, AMChannel*, gpointer);
static void      gl_mixer__on_plugin_changed           (GObject*, AMChannel*, gpointer);
static void      gl_mixer__on_tracks_change_by_list    (GObject*, GList*, AMChangeType, AyyiPanel*, gpointer);
static void      gl_mixer__on_track_selection_change   (GObject*, gpointer);
static AGlActor* gl_mixer__insert_new_strip            (GlMixerWin*, AMChannel*, int pos);

static void      gl_mixer__select_all                  (AyyiPanel*);
static void      gl_mixer__zoom_in                     (AyyiPanel*, ZoomType);
static void      gl_mixer__zoom_out                    (AyyiPanel*, ZoomType);

static void      gl_mixer__periodic_update             (GlMixerWin*);

G_DEFINE_TYPE (AyyiGlMixerWin, ayyi_gl_mixer_win, AYYI_TYPE_PANEL)

static GdkGLConfig* glconfig = NULL;


void
gl_mixer_init()
{
	g_type_class_unref (g_type_class_ref (AYYI_TYPE_GL_MIXER_WIN));
}


static void
ayyi_gl_mixer_win_class_init (AyyiGlMixerWinClass* klass)
{
	AyyiPanelClass* panel = AYYI_PANEL_CLASS(klass);
	GtkWidgetClass* widget_class = GTK_WIDGET_CLASS (klass);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS (klass);
	GObjectClass* g_object_class = G_OBJECT_CLASS (klass);

	g_object_class->constructor = gl_mixer__constructor;

	object_class->destroy = gl_mixer__destroy;

	widget_class->size_allocate = gl_mixer__on_resize;

	gl_mixer__add_accels(klass);

	panel->name           = "Gl Mixer";
	panel->has_link       = true;
	panel->has_follow     = true;
	panel->has_toolbar    = true;
	panel->default_size.x = 300;
	panel->default_size.y = 640;
	panel->config         = (ConfigParam**)NULL;
	panel->select_all     = gl_mixer__select_all;
	panel->zoom_in        = gl_mixer__zoom_in;
	panel->zoom_out       = gl_mixer__zoom_out;
	panel->on_ready       = gl_mixer__on_ready;

	g_signal_connect(song->tracks, "selection-change", G_CALLBACK(gl_mixer__on_track_selection_change), NULL);
	g_signal_connect(song->channels, "change", G_CALLBACK(gl_mixer__on_channel_change), NULL);
	g_signal_connect(song->channels, "delete", G_CALLBACK(gl_mixer__on_channel_delete), NULL);

	am_song__connect("plugin-change", G_CALLBACK(gl_mixer__on_plugin_changed), NULL);
	am_song__connect("tracks-change-by-list", G_CALLBACK(gl_mixer__on_tracks_change_by_list), NULL);

	void _mixer__periodic_update(gpointer _){ gl_mixer_foreach { gl_mixer__periodic_update(mixer); } end_gl_mixer_foreach }
	am_song__connect("periodic-update", (GCallback)_mixer__periodic_update, NULL);
}


static void
ayyi_gl_mixer_win_init(AyyiGlMixerWin* mixer)
{
}


static GObject*
gl_mixer__constructor (GType type, guint n_construct_properties, GObjectConstructParam* construct_param)
{
	if(_debug_ > -1) printf("%s%s%s\n", constructor_colour, __func__, white);

	GObject* g_object = G_OBJECT_CLASS(ayyi_gl_mixer_win_parent_class)->constructor(type, n_construct_properties, construct_param);

	ayyi_panel_set_name((AyyiPanel*)g_object, AYYI_TYPE_GL_MIXER_WIN);

	agl = agl_get_instance();

	return g_object;
}


static void
gl_mixer__destroy(GtkObject* object)
{
	AyyiPanel* panel = (AyyiPanel*)object;
	GlMixerWin* mixer = (GlMixerWin*)panel;

	int n = 0;
	#define channels_handler_disconnect(FN) if((n = g_signal_handlers_disconnect_matched (song->channels, G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA, 0, 0, NULL, FN, mixer)) != 1) gwarn("channel handler disconnection n=%i", n);
	channels_handler_disconnect(gl_mixer__on_channel_add);
	channels_handler_disconnect(gl_mixer__on_channel_change);
	channels_handler_disconnect(gl_mixer__on_channel_selection_change);
}


static void
gl_mixer__on_ready(AyyiPanel* panel)
{
	GlMixerWin* mixer = (GlMixerWin*)panel;

	if(mixer->area) return;

	am_song__connect("song-load", G_CALLBACK(gl_mixer__on_song_load), mixer);
	am_song__connect("song-unload", G_CALLBACK(gl_mixer__on_song_unload), mixer);
	g_signal_connect(song->channels, "add", G_CALLBACK(gl_mixer__on_channel_add), mixer);
	g_signal_connect(song->channels, "change", G_CALLBACK(gl_mixer__on_channel_change), mixer);

	if(!glconfig){
		if(!(glconfig = gdk_gl_config_new_by_mode(GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH | GDK_GL_MODE_STENCIL | GDK_GL_MODE_DOUBLE))){
			gerr ("Cannot initialise gtkglext."); return;
		}
	}

	int width = 256, height = 128;
	mixer->area = gtk_drawing_area_new();
	gtk_widget_set_can_focus(mixer->area, true);
	gtk_widget_set_size_request(mixer->area, width, height);
	dbg(2, "setting gl_capability context=%p", agl_get_gl_context());
	gtk_widget_set_gl_capability(mixer->area, glconfig, agl_get_gl_context(), DIRECT, GDK_GL_RGBA_TYPE);
	gtk_widget_add_events(mixer->area, GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
	panel_pack(mixer->area, EXPAND_TRUE);

	mixer->actor = agl_actor__new_root((GtkWidget*)mixer->area);

	g_signal_connect((gpointer)mixer->area, "realize", G_CALLBACK(gl_mixer__on_realise), mixer);
	g_signal_connect((gpointer)mixer->area, "expose_event", G_CALLBACK(gl_mixer__on_expose), mixer);
	g_signal_connect((gpointer)mixer->area, "event", G_CALLBACK(gl_mixer__on_event), mixer);

	if(song->loaded){
		gl_mixer__on_song_load((GObject*)song, mixer);
	}
}


static void
gl_mixer__on_song_load(GObject* _song, gpointer user_data)
{
	// TODO do unload first once we know what to unload

	GlMixerWin* mixer = user_data;

	AMIter iter;
	channel_iter_init(&iter);
	AMChannel* channel;
	while((channel = channel_next(&iter))){
		switch(channel->type){
			case AM_CHAN_INPUT:
				gl_mixer__insert_new_strip(mixer, channel, -1);
				break;
			default:
				break;
		}
	}
}


static void
gl_mixer__on_resize(GtkWidget* widget, GtkAllocation* allocation)
{
	GlMixerWin* mixer = (GlMixerWin*)widget;

	((GtkWidgetClass*)ayyi_gl_mixer_win_parent_class)->size_allocate(widget, allocation);

	mixer->actor->region = (AGliRegion){0, 0, mixer->area->allocation.width, mixer->area->allocation.height};
}


static void
gl_mixer__on_song_unload(GObject* _song, gpointer _mixer)
{
}


static void
gl_mixer__on_channel_add(GObject* _song, AMChannel* new_channel, GlMixerWin* mixer)
{
}


static void
gl_mixer__on_channel_change(GObject* _song, gpointer user_data)
{
}


static void
gl_mixer__on_channel_selection_change(GObject* _song, AyyiPanel* sender, gpointer user_data)
{
	GlMixerWin* mixer = user_data;
	AyyiPanel* panel = (AyyiPanel*)mixer;
	g_return_if_fail(AYYI_IS_GL_MIXER_WIN(mixer));

	if((panel == sender) || panel->link){

		GList* l = mixer->selectionlist;
		for(;l;l=l->next){
  			GlMixerStrip* strip = g_list_nth_data(mixer->actor->children, GPOINTER_TO_INT(mixer->selectionlist->data));
			if(strip){
				strip->selected = true;
				agl_actor__invalidate((AGlActor*)strip);
			}
		}
	}
}


static int
gl_mixer__get_n_strips(GlMixerWin* mixer)
{
	return g_list_length(mixer->actor->children);
}


static AGlActor*
gl_mixer__insert_new_strip(GlMixerWin* mixer, AMChannel* channel, int pos)
{
	// Create a new Strip widget, and add it to the mixer window.

	// @param pos - set to -1 to insert at the default position.

	if(pos == -1){
		// default position is at the end, but before the master.
		pos = MAX(0, gl_mixer__get_n_strips(mixer));
	}
	dbg(1, "inserting new strip at pos=%i. n_strips=%i", pos, gl_mixer__get_n_strips(mixer));

	AGlActor* actor = agl_actor__add_child(mixer->actor, gl_strip_new((GtkWidget*)mixer));
	#define STRIP_WIDTH 40
	#define PADDING 4
	actor->region = (AGliRegion){PADDING + STRIP_WIDTH * pos, 0, STRIP_WIDTH * (pos + 1) - PADDING, mixer->actor->region.y2};

	agl_actor__set_size(mixer->actor);

	return actor;
}


static void
gl_mixer__on_realise(GtkWidget* widget, gpointer _mixer)
{
	AyyiGlMixerWin* mixer = _mixer;

	if(!mixer->gl_drawable){
		mixer->gl_drawable = gtk_widget_get_gl_drawable(mixer->area);
		GdkGLContext* share_list = agl_get_gl_context();
		mixer->gl_context = gtk_widget_create_gl_context(mixer->area, share_list, DIRECT, GDK_GL_RGBA_TYPE);
		if(!share_list) share_list = mixer->gl_context;
	}
}


static gint
gl_mixer__on_event(GtkWidget* widget, GdkEvent* event, gpointer _mixer)
{
	return NOT_HANDLED;
}


static bool
gl_mixer__on_expose(GtkWidget* widget, GdkEventExpose* event, gpointer _mixer)
{
	AyyiGlMixerWin* mixer = _mixer;

	return agl_actor__on_expose(widget, event, mixer->actor);
}


static void
gl_mixer__on_track_selection_change(GObject* object, gpointer user_data)
{
}


static void
gl_mixer__on_channel_delete(GObject* _song, AMChannel* channel, gpointer user_data)
{
}


static void
gl_mixer__on_plugin_changed(GObject* _song, AMChannel* ch, gpointer user_data)
{
}


static void
gl_mixer__on_tracks_change_by_list(GObject* _song, GList* tracks, AMChangeType change_type, AyyiPanel* sender_win, gpointer user_data)
{
}


static void
gl_mixer__periodic_update(GlMixerWin* mixer)
{
}


static void
gl_mixer__select_all(AyyiPanel* panel)
{
}


static void
gl_mixer__zoom_in(AyyiPanel* panel, ZoomType zoom_type)
{
}


static void
gl_mixer__zoom_out(AyyiPanel* panel, ZoomType zoom_type)
{
}


static void
gl_mixer__add_accels(AyyiGlMixerWinClass* klass)
{
	AyyiPanelClass* panel = AYYI_PANEL_CLASS(klass);

	void k_zoom_in_hor(GtkAccelGroup* accel_group, gpointer user_data)
	{
		AyyiPanel* panel = app->active_panel;

		if(panel) gl_mixer__zoom_in(panel, ZOOM_H);
	}

	void k_zoom_out_hor(GtkAccelGroup* accel_group, gpointer user_data)
	{
		PF;
		AyyiPanel* panel = app->active_panel;
		if (!panel) return;

		gl_mixer__zoom_out(panel, ZOOM_H);
	}

	AMAccel keys[] = {
		{"Zoom In Hor",  {{(char)'d', 0, }, {0, 0}}, k_zoom_in_hor,  NULL},
		{"Zoom Out Hor", {{(char)'a', 0, }, {0, 0}}, k_zoom_out_hor, NULL},
	};

	panel->accel_group = gtk_accel_group_new();
	make_accels(panel->accel_group, panel->action_group, keys, G_N_ELEMENTS(keys), NULL);
}


