/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2014-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/

typedef struct {
    AGlActor actor;
    bool     selected;
} GlMixerStrip;

static void gl_strip_set_size(AGlActor*);
static bool gl_strip_on_event(AGlActor*, GdkEvent*, AGliPt);

static AGlActorNew gl_fader_new, gl_aux_new;


AGlActor*
gl_strip_new(GtkWidget* panel)
{
	AGlActor* actor = (AGlActor*)AYYI_NEW(GlMixerStrip,
		.actor = {
			.name     = "Strip",
			.region   = (AGliRegion){0,}, // size is set by parent
			.set_size = gl_strip_set_size,
			.paint    = agl_actor__null_painter,
			.on_event = gl_strip_on_event
		}
	);

	agl_actor__add_child(actor, gl_fader_new(panel));
	agl_actor__add_child(actor, gl_aux_new(panel));

	return actor;
}


static void
gl_strip_set_size(AGlActor* actor)
{
}


static bool
gl_strip_on_event(AGlActor* actor, GdkEvent* event, AGliPt xy)
{
	return NOT_HANDLED;
}


typedef struct {
	AGlShader    shader;
	struct {
		float    dummy;
	}            uniform;
} FaderShader;


static void
fader_set_uniforms()
{
}

static FaderShader fader_shader = {{NULL, NULL, 0, agl_null_uniforms, fader_set_uniforms, &fader_text}};


static AGlActor*
gl_fader_new(GtkWidget* panel)
{
	bool gl_fader__paint(AGlActor* actor)
	{
		agl_use_program((AGlShader*)&fader_shader);
		agl_rect(0, 0, agl_actor__width(actor), agl_actor__height(actor));

		return true;
	}

	void gl_fader_set_size(AGlActor* actor)
	{
		actor->region.x2 = agl_actor__width(actor->parent);
	}

	bool gl_fader__on_event(AGlActor* actor, GdkEvent* event, AGliPt xy)
	{
		return NOT_HANDLED;
	}

	AGlActor* actor = AYYI_NEW(AGlActor,
		.name     = "Fader",
		.region   = (AGliRegion){0, 40, 20, 160},
		.paint    = agl->use_shaders ? gl_fader__paint : agl_actor__null_painter,
		.set_size = gl_fader_set_size,
		.on_event = gl_fader__on_event
	);
	return actor;
}


static AGlActor*
gl_aux_new(GtkWidget* panel)
{
	bool gl_aux__paint(AGlActor* actor)
	{
		agl->shaders.plain->uniform.colour = 0xff00007f;
		agl_use_program((AGlShader*)agl->shaders.plain);
		agl_rect(0, 0, agl_actor__width(actor), agl_actor__height(actor));

		return true;
	}

	void gl_aux_set_size(AGlActor* actor)
	{
		actor->region.x2 = agl_actor__width(actor->parent);
	}

	bool gl_aux__on_event(AGlActor* actor, GdkEvent* event, AGliPt xy)
	{
		return NOT_HANDLED;
	}

	AGlActor* actor = (AGlActor*)AYYI_NEW(AGlActor,
		.name     = "Aux",
		.region   = (AGliRegion){0, 10, 0, 30},
		.paint    = gl_aux__paint,
		.set_size = gl_aux_set_size,
		.on_event = gl_aux__on_event
	);
	return actor;
}


