/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#define __transport_window_c__
#include "global.h"
#include <math.h>
#include <gdk/gdkkeysyms.h>

#include <model/am_time.h>
#include "model/am_message.h"
#include "widgets/svgtogglebutton.h"
#include "windows.h"
#include "shell.h"
#include "transport.h"
#include "song.h"
#include "icon.h"
#include "support.h"
#include "toolbar.h"
#include "transport/transport_locators_box.h"
#include "text_fx.h"
#include <image/text_render.h>

//#define BIG_FONT "Verdana Bold 40"
//#define BIG_FONT "Vera Sans Bold 40"
#define BIG_FONT "Goha-Tiheb Zemen Bold 40"
#undef SET_FONT_USING_STYLE

extern SMConfig* config;

static GObject* transport_construct        (GType, guint n_construct_properties, GObjectConstructParam*);
static void     transport_on_ready         (AyyiPanel*);
static gboolean transport_spp_on_edit_start(GtkWidget*, gpointer);
static gboolean transport_on_key_press     (GtkWidget*, GdkEventKey*, gpointer);
static void     transport_on_resize        (GtkWidget*, GtkAllocation*);
static void     transport_on_realize       (GtkWidget*);
static gboolean transport_play             (GtkWidget*, gpointer);
static gboolean transport_stop             (GtkWidget*, gpointer);
static gboolean transport_rec              (GtkWidget*, gpointer);
static gboolean transport_rew              (GtkWidget*, gpointer);
static void     transport_ff               (GtkWidget*, gpointer);
static void     transport_cycle_toggle     (GtkWidget*, gpointer);
static gboolean transport_spp_on_edit_done (GtkWidget*, gpointer);
static void     transport_locator_update   ();
static void     transport_sync             (GObject*, gpointer);
static void     transport_on_locators      (GObject*, gpointer);
static void     transport_periodic_update  (TransportWin*);


static gboolean transport_button_enter     (GtkWidget*, GdkEventCrossing*, gpointer);

G_DEFINE_TYPE (AyyiTransportWin, ayyi_transport_win, AYYI_TYPE_PANEL)


void
transport_init()
{
	g_type_class_unref (g_type_class_ref (AYYI_TYPE_TRANSPORT_WIN));

	// Fill the global tr_array with info on icons and callbacks for transport buttons.

	// TODO these are not part of the transport window!
	int i = 0;
	TransportButton* transport_struct;
	tr_array = g_array_sized_new(ZERO_TERMINATED, FALSE/*clear*/, sizeof(TransportButton), MAX_TRANSPORT_BUT);

	transport_struct           = &g_array_index(tr_array, TransportButton, i++);
	*transport_struct = (TransportButton){
		.callback = G_CALLBACK(transport_rec),
		.active   = ayyi_transport_is_rec_enabled,
		.svgpath  = "record",
		.mouseover= transport_button_enter
	};
	strcpy(transport_struct->desc, "Record");

	transport_struct           = &g_array_index(tr_array, TransportButton, i++);
	transport_struct->callback = G_CALLBACK(transport_rew);
	transport_struct->active   = ayyi_transport_is_rew;
	transport_struct->svgpath  = "tr_rew";
	transport_struct->mouseover= transport_button_enter;
	strcpy(transport_struct->desc, "Rewind Rewind");

	transport_struct           = &g_array_index(tr_array, TransportButton, i++);
	transport_struct->callback = G_CALLBACK(transport_stop);
	transport_struct->active   = ayyi_transport_is_stopped;
	transport_struct->svgpath  = "tr_stop";
	transport_struct->mouseover= transport_button_enter;
	strcpy(transport_struct->desc, "Stop");

	transport_struct           = &g_array_index(tr_array, TransportButton, i++);
	transport_struct->callback = G_CALLBACK(transport_play);
	transport_struct->active   = ayyi_transport_is_playing;
	transport_struct->svgpath  = "tr_play";
	transport_struct->mouseover= transport_button_enter;
	strcpy(transport_struct->desc, "Play");

	transport_struct           = &g_array_index(tr_array, TransportButton, i++);
	transport_struct->callback = G_CALLBACK(transport_ff);
	transport_struct->active   = ayyi_transport_is_ff;
	transport_struct->svgpath  = "tr_ff";
	transport_struct->mouseover= transport_button_enter;
	strcpy(transport_struct->desc, "ff");

	transport_struct           = &g_array_index(tr_array, TransportButton, i++);
	transport_struct->callback = G_CALLBACK(transport_cycle_toggle);
	transport_struct->active   = ayyi_transport_is_cycling;
	transport_struct->svgpath  = "refresh";
	transport_struct->mouseover= transport_button_enter;
	strcpy(transport_struct->desc, "cycle");

	g_array_set_size(tr_array, i);

	//--------------------------------------------------------

	//create pre-rendered text glyphs:

	//init_prerendered_glyphs(app.text_buf);
	char digit[4];
	PangoFontDescription* font_desc = pango_font_description_from_string("Verdana Bold 56");
	int width, height;
	text_get_max_char_size("0123456789", font_desc, &width, &height);
	int fx_width = 25; //filtering the image makes it bigger
	//warning: oversized for unexplained gandalf edge behaviour and segfaults. Does gandalf prefer square buffer?
	width += fx_width + 24;
	height += fx_width + 24;

	ImageCache* images = app->image_cache = g_new0(ImageCache, 1);

	#define N_RENDERED_CHARS 11 //10 numbers, and ":"
	//Glyph (*mem)[N_RENDERED_CHARS] = (Glyph (*)[N_RENDERED_CHARS])g_new0(Glyph, N_RENDERED_CHARS);
	char* mem = g_malloc0(sizeof(Glyph) * N_RENDERED_CHARS);

	for(i=0;i<N_RENDERED_CHARS;i++){
		//dbg(0, "%i:  %p %p", i, app.text_buf[i], mem + i * sizeof(Glyph));
		//Glyph* glyph = app.text_buf[i] = app.text_buf[0] + i * sizeof(Glyph);
		Glyph* glyph = images->text_buf[i] = (Glyph*)(mem + i * sizeof(Glyph));
		//Glyph* glyph = app.text_buf[i] = mem + i * sizeof(Glyph);
		//Glyph* glyph = app.text_buf[i] = mem[i];
		//Glyph* glyph = app.text_buf[i] = g_new0(Glyph, 1);
		glyph->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, HAS_ALPHA_FALSE, BITS_PER_PIXEL, width, height);
		glyph->rect = g_new0(GdkRectangle, 1);
		sprintf(digit, "%c", 48+i);
		int fx_width2 = 25;
		render_text_fx(digit, images->text_buf[i], &fx_width2, font_desc);
		if(fx_width2 != fx_width) gwarn("fx_width changed. may crash.");

		glyph->rect->x      = fx_width;
		glyph->rect->width  = MAX(width  - fx_width - 24, 0); //FIXME 24
		glyph->rect->height = MAX(height - fx_width - 24, 0);
	}
	images->kernel_size = fx_width;
}


static void
ayyi_transport_win_class_init (AyyiTransportWinClass *klass)
{
	AyyiPanelClass* panel = AYYI_PANEL_CLASS(klass);
	GtkWidgetClass* widget_class = GTK_WIDGET_CLASS (klass);
	GObjectClass* g_object_class = G_OBJECT_CLASS (klass);

	panel->name         = "Transport";
	panel->has_link     = false;
	panel->has_follow   = false;
	panel->accel        = GDK_F3;
	panel->new          = transport_new;
	panel->on_ready     = transport_on_ready;

	widget_class->size_allocate = transport_on_resize;
	widget_class->realize       = transport_on_realize;

	g_object_class->constructor = transport_construct;

	am_song__connect("locators-change", G_CALLBACK(transport_on_locators), NULL);
	am_song__connect("transport-sync", G_CALLBACK(transport_sync), NULL);

	void _transport_periodic_update(gpointer _){ transport_foreach { transport_periodic_update(transport); } end_transport_foreach }
	am_song__connect("periodic-update", (GCallback)_transport_periodic_update, NULL);
}


static void
ayyi_transport_win_init(AyyiTransportWin* object)
{
}


static GObject*
transport_construct (GType type, guint n_construct_properties, GObjectConstructParam* construct_param)
{
	GObject* object = G_OBJECT_CLASS(ayyi_transport_win_parent_class)->constructor(type, n_construct_properties, construct_param);

	ayyi_panel_set_name((AyyiPanel*)object, AYYI_TYPE_TRANSPORT_WIN);

	return object;
}


GtkWidget*
transport_new(AyyiShell* shell, Size* config_size, GtkOrientation orientation)
{
	/*

	window
	+--hbox
	   +--lhs vbox
	   |  +--main spp bbst
	   |  +--button hbox
	   |     +--transport button1 or pixmap
	   |     +--transport button2
	   |     +--more transport buttons...
	   |
	   +--vbox
	      +--loc1 label
	      +--loc1 bbst
	      +--loc2 label
	      +--loc2 bbst

	*/

	if (!tr_array) { gerr ("transport data not initialised."); return NULL; }

	TransportWin* transport_win = AYYI_TRANSPORT_WIN (g_object_new (AYYI_TYPE_TRANSPORT_WIN, "behavior", GDL_DOCK_ITEM_BEH_NORMAL, "orientation", orientation, NULL));
	ayyi_panel_on_new((AyyiPanel*)transport_win, config_size, orientation);

	return (GtkWidget*)transport_win;
}


static void
transport_on_ready(AyyiPanel* panel)
{
	TransportWin* transport_win = (TransportWin*)panel;

	if(transport_win->spp_entry) return;

	GtkWidget* hbox1 = transport_win->hbox = gtk_hbox_new(NON_HOMOGENOUS, 0);
	gtk_widget_show(hbox1);
	//if parent is GtkPaned, set the height to be 100%
	void hbox_on_size_request(GtkWidget* widget, GtkRequisition* req, gpointer data)
	{
		PF;
		req->height = widget_is_in_paned(widget) ? gtk_widget_get_parent(widget)->allocation.height : 100;
	}
	g_signal_connect(transport_win->hbox, "size-request", G_CALLBACK(hbox_on_size_request), NULL);
	panel_pack(hbox1, EXPAND_FALSE);

	//-----------------------------------------

	PangoFontDescription* font_desc = pango_font_description_from_string (BIG_FONT);
	pango_font_description_set_size(font_desc, PANGO_SCALE * (app->font_size * 2 + 8));

	//-----------------------------------------

	//lhs vbox:
	GtkWidget* vbox_lhs = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox_lhs);
	gtk_container_set_border_width(GTK_CONTAINER(vbox_lhs), 2);
	gtk_box_pack_start(GTK_BOX(hbox1), vbox_lhs, FALSE, FALSE, 0);

	//an hbox for the spp and label so they can be padded on lhs: ? needed?

	//label for bar spp:
#ifdef SHOW_SPP_LABEL
	GtkWidget* label1 = gtk_label_new ("spp");
	gtk_misc_set_alignment(GTK_MISC(label1), 0,0);
	gtk_widget_show (label1);
	gtk_box_pack_start(GTK_BOX(vbox_lhs), label1, FALSE, FALSE, 0);
#endif

	// main spp display
//#undef USE_GANDALF //double free error on window close!
#if USE_GANDALF
	GtkWidget* text_fx = transport_win->textfx = text_fx_new(app.image_cache->text_buf);
	//text_fx_set_string(text_fx, "000:00:00:00"); //crashes.
	gtk_widget_show(text_fx);
	gtk_box_pack_start(GTK_BOX(vbox_lhs), text_fx, FALSE, FALSE, 0);
#elif USE_TEXT_VIEW
	// this uses a multiline text widget - see below for single line widget...
	GtkWidget* spp_view = gtk_text_view_new();
	GtkTextBuffer* buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(spp_view));
	gtk_text_buffer_set_text(buf, "05:04:00:00", -1);
	gtk_widget_set_name(spp_view, "transport");
	gtk_text_view_set_accepts_tab(GTK_TEXT_VIEW(spp_view), FALSE);
	//even with all borders set to 0, there is still a lot of vert space inside the textview.
	gtk_text_view_set_pixels_above_lines(GTK_TEXT_VIEW(spp_view), -1);
	gtk_text_view_set_pixels_below_lines(GTK_TEXT_VIEW(spp_view), -1);
	gtk_container_set_border_width(GTK_CONTAINER(spp_view), 1); //this border is 'outside' the textview.

	// set left margin
	GValue gval = {0,};
	g_value_init(&gval, G_TYPE_CHAR);
	g_value_set_char(&gval, 4);
	g_object_set_property(G_OBJECT(spp_view), "left-margin", &gval);
	//free(gval); //is this sufficient? http://10.0.0.151/man/gobject/gobject-Type-Information.html#GTypeValueTable

	gtk_widget_show(spp_view);
	gtk_widget_modify_font(spp_view, font_desc);
	gtk_box_pack_start(GTK_BOX(vbox_lhs), spp_view, FALSE, FALSE, 0);
#else

	// single line spp
	GtkWidget* spp_entry = transport_win->spp_entry = gtk_entry_new_with_max_length(13);
	gtk_entry_set_text(GTK_ENTRY(spp_entry), "001:02:03:470");
	gtk_entry_set_has_frame(GTK_ENTRY(spp_entry), FALSE);
	gtk_widget_modify_font(spp_entry, font_desc);
	gtk_widget_show(spp_entry);
	gtk_box_pack_start(GTK_BOX(vbox_lhs), spp_entry, FALSE, FALSE, 0);
	g_signal_connect((gpointer)spp_entry, "focus-in-event", G_CALLBACK(transport_spp_on_edit_start), NULL);
	g_signal_connect((gpointer)spp_entry, "focus-out-event", G_CALLBACK(transport_spp_on_edit_done), NULL);
#endif

	// buttons

	GtkWidget* hbox_b = transport_win->button_box = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox_b);
	gtk_box_pack_end(GTK_BOX(vbox_lhs), hbox_b, FALSE, FALSE, 0);

	int transport_style = TRANSTYLE_PLAIN;
	TransportButton* button;
	#define DEFAULT_TR_BUT_SIZE 48
	int size = DEFAULT_TR_BUT_SIZE;

	int i = 0;
	if(transport_style == TRANSTYLE_PLAIN){
		for(i=0;i<tr_array->len;i++){
			button = &g_array_index(tr_array, TransportButton, i);
			GtkWidget* b = transport_win->b[i] = (GtkWidget*)svg_toggle_button_new(button->svgpath);
			svg_toggle_button_set_size((SvgToggleButton*)b, size);
			gtk_box_pack_start(GTK_BOX(hbox_b), b, FALSE, FALSE, 0);
			g_signal_connect(G_OBJECT(b), "clicked", button->callback, NULL);
		}
	} else {
		GtkWidget* image3 = gtk_image_new_from_file("pixmaps/buttons1.xpm");
		gtk_widget_show(image3);
		gtk_box_pack_start(GTK_BOX(hbox_b), image3, TRUE, TRUE, 0);
	}

	//-----------------------------------------

	// locators

	transport_win->locators = transport_locators_box_new(transport_win);
	gtk_box_pack_start(GTK_BOX(transport_win->hbox), GTK_WIDGET(((TransportLocatorsBox*)transport_win->locators)->locators_bg), FALSE, FALSE, 0);

	pango_font_description_free(font_desc); //needed?
	//key-press is connected to the toplevel window as connecting to other widgets doesnt always work.
	g_signal_connect((gpointer)panel->shell->widget, "key-press-event", G_CALLBACK(transport_on_key_press), transport_win);

	//focus the play button:
	gtk_widget_grab_focus(transport_win->b[3]);

	ayyi_panel_on_open(panel);
	transport_sync(NULL, panel);
}


static gboolean
transport_spp_on_edit_start(GtkWidget* widget, gpointer user_data)
{
	PF;
	gtk_window_remove_accel_group(GTK_WINDOW(gtk_widget_get_toplevel(widget)), app->global_accel_group);
	return FALSE;
}


static gboolean
transport_spp_on_edit_done(GtkWidget* widget, gpointer user_data)
{
	PF;
	gtk_window_add_accel_group(GTK_WINDOW(gtk_widget_get_toplevel(widget)), app->global_accel_group);
	return FALSE;
}


/*
 *  Change the size of the widgets so they fit in the window.
 *  Note: there is no scrollwin on this window.
 */
static void
transport_on_resize(GtkWidget* widget, GtkAllocation* allocation)
{
	// TODO this needs some work. Sizes are only approx, and are not stable.

	if(!GTK_WIDGET_REALIZED(widget)) return;
	PF;
	((GtkWidgetClass*)ayyi_transport_win_parent_class)->size_allocate(widget, allocation);

	TransportWin* panel = (TransportWin*)widget;

	if(panel->old_allocation.height == allocation->height && panel->old_allocation.width == allocation->width && !panel->button_resize_pending){
		dbg(1, "unchanged");
		return;
	}

	// a trap to prevent endless resize attempts
	static int count = 0;
	gboolean resize__on_timeout(gpointer data)
	{
		int* count = data;
		(*count)--;
		if(*count < -10){ *count = 0; return TIMER_STOP; }
		return TIMER_CONTINUE;
	}
	if(!count) g_timeout_add(100, resize__on_timeout, &count);
	if(count++ > 50){ gwarn("too many resize attempts!"); return; }

	if(!panel->spp_entry) return; // deal with this later...
	g_return_if_fail(GTK_ENTRY(panel->spp_entry));

	//FIXME we cannot get the width of the right hand column because it is being allocated space outside of the visible area!
	//      ...however, maybe this doesnt matter - if there isnt room for it, dont show it.
	//GtkWidget* parent = gtk_widget_get_parent(widget);
	int total_width = /*(G_OBJECT_TYPE(parent) == GTK_TYPE_VPANED) ? parent->allocation.width :*/ widget->allocation.width;
	if(panel->locators->allocation.width > panel->locators->requisition.width) gwarn("wider!");
	int right_hand_column_width = panel->locators->allocation.width > 10 ? panel->locators->allocation.width : 150;
	int left_hand_column_width = MAX(130, total_width - right_hand_column_width - 2);
	dbg(1, "panel_width=%i locators_width=%i left_width=%i", widget->allocation.width, panel->locators->allocation.width, left_hand_column_width);

	if(!panel->button_resize_pending) { //main spp size
		dbg(1, "1");

		transport_locators_box_update_n_visible(panel->locators);

#ifdef SET_FONT_USING_STYLE
		//because the gtk_entry has a style applied to it, we cannot use gtk_widget_modify_font(). We set the font_desc directly on the style.
#endif

  PangoFontDescription* font_desc = pango_font_description_from_string (BIG_FONT);
		PangoLayout* layout = gtk_entry_get_layout(GTK_ENTRY(panel->spp_entry));
		//GtkStyle* style = gtk_widget_get_style(panel->spp_entry);
		GtkStyle* style = gtk_style_copy(gtk_widget_get_style(panel->spp_entry));
		//			GtkRcStyle* rcstyle = gtk_widget_get_modifier_style(panel->spp_entry);
		PangoFontDescription* font_desc_o = style->font_desc;
		PangoRectangle logical_rect;
		pango_layout_get_pixel_extents(layout, NULL, &logical_rect);
#ifdef SET_FONT_USING_STYLE
#else
		g_object_unref(style);
#endif

		int spp_alloc_height = panel->spp_entry->allocation.height;
		int spp_text_height = logical_rect.height - logical_rect.y;//ink_rect.height;
		int spp_padding_height = spp_alloc_height - spp_text_height;
		int spp_font_size = pango_font_description_get_size(font_desc_o);
		gboolean absolute = pango_font_description_get_size_is_absolute(font_desc_o);

		//how tall do we want to be?
		//if the box is not tall enough, use all the available height.
		#define SMALL_HEIGHT 50
		int spp_target_height = (widget->allocation.height < SMALL_HEIGHT) ? widget->allocation.height : widget->allocation.height / 2 - 2 - spp_padding_height;

		int spp_text_width = logical_rect.width - logical_rect.x;
		int spp_padding_width = 5; //total guess
		int spp_target_font_width = left_hand_column_width - spp_padding_width;

		g_return_if_fail(spp_text_height);
		g_return_if_fail(spp_text_width);
		float w1 = spp_text_width;
		dbg(1, "width: current=%.0f target=%i --> %i", w1, spp_target_font_width, (int)(((spp_font_size * spp_target_font_width) / spp_text_width)) / PANGO_SCALE);
		dbg(1, "height: current=%i target=%i --> %i", spp_text_height, spp_target_height, (int)(((spp_font_size * spp_target_height) / spp_text_height) / PANGO_SCALE));
		int spp_target_font_size = MIN(
			(spp_font_size * spp_target_height) / spp_text_height,
			(spp_font_size * spp_target_font_width) / spp_text_width
		);
		spp_target_font_size = 2 * PANGO_SCALE * (spp_target_font_size / (2 * PANGO_SCALE));
		dbg (2, "spp: target: %i x %i", left_hand_column_width, spp_target_height);
		dbg (1, "vpadding=%i font_size=%i-->%i", spp_padding_height, spp_font_size/PANGO_SCALE, spp_target_font_size/PANGO_SCALE);
		if(ABS(spp_target_font_size - spp_font_size) >= 2){
			if(absolute) pango_font_description_set_absolute_size(font_desc, spp_target_font_size);
			else         pango_font_description_set_size(font_desc, spp_target_font_size);

			pango_font_description_set_weight(font_desc, PANGO_WEIGHT_HEAVY);

			dbg(2, "  %s", pango_font_description_to_string(font_desc));
#ifdef SET_FONT_USING_STYLE
			gtk_widget_set_style(panel->spp_entry, style);
#else

			GtkRcStyle* rcstyle = gtk_widget_get_modifier_style(panel->spp_entry);
			rcstyle->font_desc = font_desc;
			gtk_widget_modify_font(panel->spp_entry, font_desc);
#endif

			gtk_widget_set_size_request(panel->spp_entry, spp_target_font_width - 4, -1); // why does the widget do this itself?

			panel->button_resize_pending = true;
		}
		else if(_debug_) gwarn("style not applied: too similar.");
	}

	else{
		dbg(1, "2");
		int aw = widget->allocation.width;
		int rw = widget->requisition.width;

		if(aw < rw){
			if(panel->button_resize_pending){
				//handle the (eg, tabbed wm) simple case where the alloc may be smaller than the requisition:

				if(widget->allocation.height >= SMALL_HEIGHT){
					gtk_widget_show(panel->button_box);

					int available_height = widget->allocation.height - panel->spp_entry->allocation.height;

					// change button size
					int n_buttons = tr_array->len;

					int old_button_width = panel->b[0]->allocation.width;
					static int previous_button_req = 0;
					#define PADDING_BODGE 8 // There is a difference in the way we are getting and setting the button size.
					int button_req = MAX(
						16,
						MIN(left_hand_column_width / n_buttons - PADDING_BODGE, available_height) // set depending on whether horizontally or vertically constrained.
					);
					dbg(3, "reducing button size: win_wid=%i win_req=%i old_button_width=%i new_button_req=%i", aw, rw, old_button_width, button_req);
					if(button_req > 3 && button_req != previous_button_req){
						int i;
						for(i=0;i<tr_array->len;i++){
							svg_toggle_button_set_size((SvgToggleButton*)panel->b[i], button_req);
						}
					}
					previous_button_req = button_req;
				}else{
					gtk_widget_hide(panel->button_box);
				}
			}
		} else {

#if 0
			//do the buttton widgets need to be increased in size?
			int tot_widget_height = panel->spp_entry->allocation.height + panel->b[0]->allocation.height;
			//printf("%s(): widget_height=%i\n", __func__, widget_height);
			#define SLACK 10
			if(tot_widget_height < widget->allocation.height - SLACK){
				dbg (4, "widgets need to be taller!");
				//so is the widget width also too small?
				int widget_width = widget->allocation.width;
				if(widget->requisition.width < widget_width - SLACK){
					//printf("%s(): widgets need to be wider!\n", __func__);

					int new_button_size = svgbutton_get_size(panel->b[0]) + 6; //FIXME set the needed size!

					int i; for(i=0;i<tr_array->len;i++){
						TransportButton* tr_struct = &g_array_index(tr_array, TransportButton, i);
						svg_toggle_button_set_size((SvgToggleButton*)panel->b[i], button_req);
					}
				}
			}
#endif
			int available_height = widget->allocation.height - panel->spp_entry->allocation.height;
			int n_buttons = tr_array->len;
			int button_req = MAX(
				16,
				MIN(
					left_hand_column_width / n_buttons - PADDING_BODGE,
					available_height
				) //set depending on whether horizontally or vertically constrained.
			);

//this is the same as above
				dbg(3, "button size: win_wid=%i win_req=%i new_button_req=%i", aw, rw, button_req);
				if(button_req > 3/* && button_req != previous_button_req*/){
					int i;
					for(i=0;i<tr_array->len;i++){
						//TransportButton* tr_struct = &g_array_index(tr_array, TransportButton, i);
						svg_toggle_button_set_size((SvgToggleButton*)panel->b[i], button_req);
					}
				}
		}
		panel->button_resize_pending = false;
	} //end phase 2.

	//transport_locators_box_update_n_visible(panel->locators);

	// reduce the min window size so users can resize
	gtk_widget_set_size_request(widget, 10, 10);

	panel->old_allocation = *allocation;
}


static void
transport_on_realize(GtkWidget* widget)
{
  //do window creation stuff that requires the widgets to be already realised.

  TransportWin* window = (TransportWin*)widget;
  window->button_resize_pending = false;

  ((GtkWidgetClass*)ayyi_transport_win_parent_class)->realize(widget);

  //set the background colour on the main spp display (we use the ACTIVE colour not the NORMAL):
  if(window->spp_entry){
    GtkStyle* style = gtk_style_copy(gtk_widget_get_style(window->spp_entry));
    //GtkStyle* style = gtk_style_copy(gtk_widget_get_style(widget));
    //GtkStyle* style = gtk_widget_get_default_style();
    GdkColor colour = style->base[GTK_STATE_NORMAL];
    style->base[GTK_STATE_NORMAL] = style->text[GTK_STATE_NORMAL];
    style->text[GTK_STATE_NORMAL] = colour;
    dbg (2, "colour=#%x", am_gdk_to_rgba(&colour));
    //colour_get_frame(&colour);
#if 0
    colour.red   = style->base[GTK_STATE_ACTIVE].red;
    colour.green = style->base[GTK_STATE_ACTIVE].green;
    colour.blue  = style->base[GTK_STATE_ACTIVE].blue;
#endif
#ifdef TEMP
    gtk_widget_set_style(window->spp_entry, style);
#endif
    g_object_unref(style);
    //printf("  colour: base-active: %x %x %x.\n", color.red, color.green, color.blue);
    //gtk_widget_modify_base(window->spp_entry, GTK_STATE_NORMAL, &colour);
  }
}


static gboolean
transport_rec(GtkWidget* widget, gpointer user_data)
{
  printf("%s%s%s()...\n", red, __func__, white);
  ayyi_transport_rec();
  return true;
}


static gboolean
transport_rew(GtkWidget* widget, gpointer user_data)
{
  PF;
  ayyi_transport_rew();
  return true;
}


static gboolean
transport_stop(GtkWidget* widget, gpointer not_used)
{
  ayyi_transport_stop();
  return true;
}


static gboolean
transport_play(GtkWidget* widget, gpointer user_data)
{
  //this fn may be called from a button press or a keyboard shortcut.

  ayyi_transport_play();

  return TRUE; //stop further propogation of the signal - needed when using keyboard shortcut.
}


static void
transport_ff(GtkWidget* widget, gpointer user_data)
{
	PF;

	ayyi_transport_ff();

	//TODO check these are handled by the observer
	transport_clear_buttons();
	svg_toggle_button_set_active((SvgToggleButton*)widget, TRUE);

	//cant get Ardour to notify us when in FF
	am_song__emit("transport-ff");
}


static void
transport_cycle_toggle(GtkWidget* widget, gpointer user_data)
{
	PF;
	am_transport_cycle_toggle();
}


void
transport_clear_buttons()
{
	//reset visual status of all transport buttons in transport windows, and toolbars.

	panel_foreach {
		if(G_OBJECT_TYPE(panel) == AYYI_TYPE_TRANSPORT_WIN){
			GtkWidget* gtkwin = shell->widget;
			g_return_if_fail(gtkwin);

			TransportWin* transport_win = (TransportWin*)panel;

			if(!transport_win->b[TR_PLAY]){ gerr ("bad button widget reference."); continue; }
			int i;
			for(i=0;i<tr_array->len;i++){
				gtk_widget_set_style(transport_win->b[i],  NULL);
			}

		}else{
			//reset any toolbar transport buttons this window may have:
			GtkWidget** toolbar = panel->toolbar->b_transport;
			if(toolbar && toolbar[0]){
				int i; for(i=0;i<TR_SIZE;i++){
					gtk_widget_set_style(toolbar[i],  NULL);
				}
			}
		}
	} end_panel_foreach
}


static void
transport_locator_update()
{
	//update displayed locator values.

	static char str[16];

	transport_foreach {
		ayyi_pos2bbst(&song->loc[1].vals[0].val.sp, str);
		dbg(3, "loc1=%s", str);
		transport_locators_box_set_text(transport->locators, 0, str);
		ayyi_pos2bbst_1(&song->loc[2].vals[0].val.sp, str);
		transport_locators_box_set_text(transport->locators, 1, str);
	} end_transport_foreach
}


gboolean
transport_is_cycling()
{
	return ayyi_transport_is_cycling();
}


gboolean
transport_button_enter(GtkWidget* widget, GdkEventCrossing* event, gpointer _data)
{
	struct _cb_data {
		AyyiPanel* panel;
		TransportButton* button;
	};
	struct _cb_data* data = (struct _cb_data*)_data;
	AyyiPanel* panel = data->panel;
	if(panel->shell){
		shell__statusbar_print(panel->shell, 2, data->button->desc);
	}
	return NOT_HANDLED;
}


static gboolean
transport_on_key_press(GtkWidget* window, GdkEventKey* event, gpointer user_data)
{
  //handle RETURN key for locator boxes.

  int handled = FALSE;

  TransportWin* transport = user_data;
  g_return_val_if_fail(transport, false);

  switch(event->keyval){
    case GDK_Return:
      {
        GtkWidget* focused = gtk_window_get_focus(GTK_WINDOW(window));
        TransportLocatorsBox* locators = TRANSPORT_LOCATORS_BOX(transport->locators);
        g_return_val_if_fail(locators, false);
        int i; for(i=0;i<5;i++){
          if(focused == locators->text_view[i]){
            dbg(0, " RETURN pressed for locate box %i.", i + 1);
            gboolean sig_ret = 1;
            g_signal_emit_by_name(focused, "focus-out-event", locators, &sig_ret);
            //restore the correct accel state:
            g_signal_emit_by_name(focused, "focus-in-event", locators, &sig_ret);
            break;
          }
        }
      }
      handled = TRUE;
      break;
    default:
      dbg(2, "keyval=%i %x", event->keyval, event->keyval);
      break;
  }
  return handled;
}


static void
transport_sync(GObject* _song, gpointer _transport)
{
	TransportWin* tw = _transport;

	transport_locator_update();

	void a(TransportWin* transport)
	{
		svg_toggle_button_set_active((SvgToggleButton*)transport->b[TR_REW],  ayyi_transport_is_rew());
		svg_toggle_button_set_active((SvgToggleButton*)transport->b[TR_STOP], ayyi_transport_is_stopped());
		svg_toggle_button_set_active((SvgToggleButton*)transport->b[TR_FF],   ayyi_transport_is_ff());
		svg_toggle_button_set_active((SvgToggleButton*)transport->b[TR_PLAY], ayyi_transport_is_playing());
		svg_toggle_button_set_active((SvgToggleButton*)transport->b[TR_CYC],  ayyi_transport_is_cycling());
	}

	if(tw)
		a(tw);
	else {
		transport_foreach {
			a(transport);
		} end_transport_foreach;
	}
}


static void
transport_on_locators(GObject* _song, gpointer user_data)
{
	transport_locator_update();
}


static void
transport_periodic_update(TransportWin* transport)
{
	// update spp text readouts, etc

	char spp_str[16];
	ayyi_transport_get_spp_string(spp_str);

	if(transport->spp_entry && !GTK_WIDGET_HAS_FOCUS(transport->spp_entry)) gtk_entry_set_text(GTK_ENTRY(transport->spp_entry), spp_str);
	//if(transport->textfx) text_fx_set_value(transport->textfx, pos.beat/4);
	if(transport->textfx) text_fx_set_string(transport->textfx, spp_str);
}


