/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#ifndef __shell_h__
#define __shell_h__
#include "panels/panel.h"
#include "menu.h"

#define shell_foreach GList* shells = shell_list; if(shells){ for(;shells;shells=shells->next){ AyyiShell* shell = shells->data;
#define end_shell_foreach }}

struct _statusbar_id
{
	guint                cid;                  //the statusbar context_id.
	guint                mid;                  //the statusbar message_id.
};

enum {
	STATUSBAR_ICON = 0,
	STATUSBAR_MAIN,
	STATUSBAR_SELECTION,
	STATUSBAR_HOVER,
	STATUSBAR_MAX
};

struct _shell
{
	GtkWidget*           widget;               //toplevel gtk window.
	GtkWidget*           vbox;

	GdlDock*             dock;
	GList*               panels;               //GdlDockItem* / AyyiPanel*

	struct {
		GtkWidget*       widget[STATUSBAR_MAX];
		StatusbarId      id_partsel;           //statusbar id for part selections
		StatusbarId      drag_id;              //statusbar id for drag operations
		GtkWidget*       progress;
	}                    statusbar;

	GtkWidget*           window_menu;          //dynamically generated Window menu on the main menubar.
	struct {
		GtkWidget*       history_item;
		GtkWidget*       history;
		GtkWidget*       recent_files[MAX_RECENT_SONGS];
		GtkWidget*       file;
		GtkWidget*       undo;
		EditMenu         canvas_type[4];       //MAX_CANVAS_TYPES
	}                    menu;

	ShellMenus*          shell_menus;
#ifndef DEBUG_DISABLE_SHORTCUTS
	GtkWidget*           context_menu;
#endif

	AyyiPanel*           focussed_panel;

	int                  current_width;        //used to check for changes.
};

AyyiShell*   shell__new                 ();
AyyiShell*   shell__new_floating        (AyyiPanelClass*);
AyyiShell*   shell__new_from_dock       (GdlDock*);
void         shell__free                (AyyiShell*);
AyyiShell*   shell__lookup_by_dock      (GdlDock*);
void         shell__add_panel           (AyyiShell*, AyyiPanel*);
void         shell__destroy_panel       (AyyiShell*, AyyiPanel*);
void         shell__remove_panel        (AyyiShell*, AyyiPanel*);
GList*       shell__get_dock_items      (AyyiShell*);
GList*       shell__get_panels          (AyyiShell*);
int          shell__get_n_panels        (AyyiShell*);
bool         shell__has_panel_type      (AyyiShell*, PanelType);
AyyiPanel*   shell__find_panel          (AyyiShell*, PanelType);
void         shell__menu_popup          (AyyiShell*, GdkEvent*);
void         shell__statusbar_print     (AyyiShell*, int n, const char* fmt, ...);
void         shell__statusbar_print_if_panel
                                        (AyyiShell*, int n, PanelType, const char* fmt, ...);
void         shell__statusbar_print_bbs (AyyiShell*, int n, GPos*);

void         print_shelllist_items      ();
#endif
