/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#define __transport_c__
#include "global.h"
#include "model/am_transport.h"
#include "panels/transport.h"
#include "model/am_song.h"
#include "widgets/svgtogglebutton.h"
#include "windows.h"
#include "icon.h"
#include "toolbar.h"
#include "transport.h"

static void transport_on_sync        (GObject*, gpointer);
static void transport_on_stop        (GObject*, gpointer);
static void transport_on_play        (GObject*, gpointer);
static void transport_on_rew         (GObject*, gpointer);
static void transport_on_ff          (GObject*, gpointer);
static void transport_on_cycle       (GObject*, gpointer);
static void transport_on_rec         (GObject*, bool enabled, gpointer);
static void check_record_status      (gpointer);


void
transport__init()
{
	am_song__connect("transport-sync",  G_CALLBACK(transport_on_sync), NULL);
	am_song__connect("transport-stop",  G_CALLBACK(transport_on_stop), NULL);
	am_song__connect("transport-play",  G_CALLBACK(transport_on_play), NULL);
	am_song__connect("transport-rew",   G_CALLBACK(transport_on_rew), NULL);
	am_song__connect("transport-ff",    G_CALLBACK(transport_on_ff), NULL);
	am_song__connect("transport-cycle", G_CALLBACK(transport_on_cycle), NULL);
	am_song__connect("transport-rec",   G_CALLBACK(transport_on_rec), NULL);
	am_song__connect("periodic-update", (GCallback)check_record_status, NULL);
}


static void
transport_on_sync(GObject* _song, gpointer user_data)
{
	//sets the correct state of all open transport windows following aquisition of rectran_shm, or a general change in transport state.

	if(ayyi_transport_is_rec_enabled())  transport_on_rec(NULL, TRUE, NULL);
}


static void
transport_on_stop(GObject* _song, gpointer user_data)
{
	//currently we leave the spp timer running for meter decay.

	transport_clear_buttons();

	transport_foreach {
		svg_toggle_button_set_active((SvgToggleButton*)transport->b[TR_STOP], true);
	} end_transport_foreach;

	app->recording = FALSE;
	request_toolbar_update();
}


static void
transport_on_play(GObject* _song, gpointer user_data)
{
	transport_clear_buttons();

	transport_foreach {
		svg_toggle_button_set_active((SvgToggleButton*)transport->b[TR_PLAY], true);
	} end_transport_foreach;

	request_toolbar_update();//FIXME we should possibly make this transport_update instead to include tr windows.
}


static void
transport_on_rew(GObject* _song, gpointer user_data)
{
	PF;
	transport_clear_buttons();
	request_toolbar_update();

	transport_foreach {
		svg_toggle_button_set_active((SvgToggleButton*)transport->b[TR_REW], true);
	} end_transport_foreach;
}


static void
transport_on_ff(GObject* _song, gpointer user_data)
{
	PF;

	transport_foreach {
		svg_toggle_button_set_active((SvgToggleButton*)transport->b[TR_FF], true);
	} end_transport_foreach;

	request_toolbar_update();
}


static void
transport_on_cycle(GObject* _song, gpointer user_data)
{
	PF;
	request_toolbar_update();
}


static void
transport_on_rec(GObject* _song, bool enabled, gpointer user_data)
{
	//currently everything neccesary is updated in transport_play_cb or transport_stop_cb
	PF;
}


static void
check_record_status(gpointer data)
{
	static int filter = 0;
	if(!filter){
		if(ayyi_transport_is_recording()){
			app->recording = TRUE;
		}
	}
	if(filter++ > 6) filter = 0;
}


void
transport_enable()
{
	if(_debug_ && song->periodic) gwarn ("timer already started.");

	am_song__enable_meters(true);

	g_signal_emit_by_name (app, "app-start-meters");

	am_transport_sync();
}


void
transport_disable()
{
	//stop spp timer:

	if(!song->periodic){ dbg(2, "timer not running?"); return; }

	am_song__enable_meters(false);
	g_signal_emit_by_name (app, "app-stop-meters", NULL);
}


