/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef __fdr_label_h__
#define __fdr_label_h__


#include <gdk/gdk.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkwidget.h>


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define GTK_DIAL2(obj)          GTK_CHECK_CAST (obj, gtk_dial2_get_type (), GtkDial2)
#define GTK_DIAL2_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, gtk_dial2_get_type (), GtkDial2Class)
#define GTK_IS_DIAL2(obj)       GTK_CHECK_TYPE (obj, gtk_dial2_get_type ())

typedef struct _GtkDial2        GtkDial2;
typedef struct _GtkDial2Class   GtkDial2Class;

struct _GtkDial2
{
  GtkWidget widget;

  //update policy (GTK_UPDATE_[CONTINUOUS/DELAYED/DISCONTINUOUS]):
  guint policy : 2;

  //button currently pressed or 0 if none:
  guint8 button;

  //dimensions of dial components:
  gint radius;
  gint pointer_width;

  /* ID of update timer, or 0 if none */
  guint32 timer;

  /* Current angle */
  gfloat angle;

  /* Old values from adjustment stored so we know when something changes */
  gfloat old_value;
  gfloat old_lower;
  gfloat old_upper;

  /* The adjustment object that stores the data for this dial */
  GtkAdjustment* adjustment;
};

struct _GtkDial2Class
{
  GtkWidgetClass parent_class;
};


GtkWidget*     fdr_label_new                    (GtkAdjustment*);
guint          gtk_dial2_get_type               (void);
void           gtk_dial2_set_update_policy      (GtkDial2*, GtkUpdateType);
void           gtk_dial2_set_adjustment         (GtkDial2*, GtkAdjustment*);
#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __fdr_label_H__ */
