/*
  svgbutton.h - multi-state button with svg images.

  This file is part of the Ayyi Project. http://www.ayyi.org
  copyright (C) 2007-2016 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#ifndef __svgbutton_h__
#define __svgbutton_h__

#include <gdk/gdk.h>
#include <gtk/gtkvbox.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define SVGBUTTON(obj)          GTK_CHECK_CAST (obj, svgbutton_get_type (), SvgButton)
#define SVGBUTTON_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, svgbutton_get_type (), SvgButtonClass)
#define IS_SVGBUTTON(obj)       GTK_CHECK_TYPE (obj, svgbutton_get_type ())

typedef struct _SvgButton       SvgButton;
typedef struct _SvgButtonClass  SvgButtonClass;

struct _SvgButton
{
  GtkButton           button;
  int                 size;
  
  GList*              state_list;
  struct _state_item* current_state;

  char*               mouseover_text;
};

struct _SvgButtonClass
{
  GtkButtonClass parent_class;

  void (* svgbutton) (SvgButton *svgb);
};

typedef struct _state_item
{
  char       name[256];
  char       svg_path[256];
  GdkPixbuf* pixbuf;
  GtkWidget* image;
} SvgButtonState;

guint          svgbutton_get_type        (void);
GtkWidget*     svgbutton_new             (void);
void           svgbutton_add_state       (SvgButton*, const char* state_name, const char* svg_path);
char*          svgbutton_get_state       (SvgButton*);
void           svgbutton_set_state_n     (SvgButton*, int n);
void           svgbutton_set_size        (SvgButton*, int);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
