/*
  This file is part of the Ayyi Project. http://www.ayyi.org
  copyright (C) 2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <glib.h>
#include <glib-object.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include "gtk/gtkbutton.h"
#include "icon.h"
#include "ayyi/ayyi_utils.h"
#include "svgtogglebutton.h"

struct _SvgToggleButtonPrivate
{
	gchar*     svgpath;
	gchar*     svgpath2;
	GdkPixbuf* pixbuf;
	GtkImage*  image;
};

static gpointer svg_toggle_button_parent_class = NULL;

GType svg_toggle_button_get_type (void) G_GNUC_CONST;
#define SVG_TOGGLE_BUTTON_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), TYPE_SVG_TOGGLE_BUTTON, SvgToggleButtonPrivate))
enum  {
	SVG_TOGGLE_BUTTON_DUMMY_PROPERTY
};
static void     svg_toggle_button_class_init                (SvgToggleButtonClass*);
static void     svg_toggle_button_instance_init             (SvgToggleButton*);
static void     svg_toggle_button_clicked                   (GtkButton*);
//static gboolean svg_toggle_button_expose_event              (GtkWidget*, GdkEventExpose*);
//static gboolean svg_toggle_button_real_button_press_event   (GtkWidget*, GdkEventButton*);
//static gboolean svg_toggle_button_real_button_release_event (GtkWidget*, GdkEventButton*);
//static gboolean svg_toggle_button_real_motion_notify_event  (GtkWidget*, GdkEventMotion*);
static void     svg_toggle_button_finalize                  (GObject*);

#ifdef OVERRIDE_DEFAULT_PAINT //TODO
static void __gtk_button_paint (GtkButton* button, const GdkRectangle* area, GtkStateType state_type, GtkShadowType shadow_type, const gchar* main_detail);
#endif


GType
svg_toggle_button_get_type ()
{
	static volatile gsize svg_toggle_button_type_id__volatile = 0;
	if (g_once_init_enter (&svg_toggle_button_type_id__volatile)) {
		static const GTypeInfo g_define_type_info = { sizeof (SvgToggleButtonClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) svg_toggle_button_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (SvgToggleButton), 0, (GInstanceInitFunc) svg_toggle_button_instance_init, NULL };
		GType svg_toggle_button_type_id;
		svg_toggle_button_type_id = g_type_register_static (GTK_TYPE_TOGGLE_BUTTON, "SvgToggleButton", &g_define_type_info, 0);
		g_once_init_leave (&svg_toggle_button_type_id__volatile, svg_toggle_button_type_id);
	}
	return svg_toggle_button_type_id__volatile;
}


static void
svg_toggle_button_class_init (SvgToggleButtonClass* klass)
{
	svg_toggle_button_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (SvgToggleButtonPrivate));

	GtkButtonClass* button_class = (GtkButtonClass*) klass;

	//GTK_WIDGET_CLASS (klass)->expose_event         = svg_toggle_button_expose_event;
	//GTK_WIDGET_CLASS (klass)->button_press_event   = svg_toggle_button_real_button_press_event;
	//GTK_WIDGET_CLASS (klass)->button_release_event = svg_toggle_button_real_button_release_event;
	//GTK_WIDGET_CLASS (klass)->motion_notify_event  = svg_toggle_button_real_motion_notify_event;
	button_class->clicked                          = svg_toggle_button_clicked;
	G_OBJECT_CLASS   (klass)->finalize             = svg_toggle_button_finalize;
}


static void
svg_toggle_button_instance_init (SvgToggleButton* self)
{
	self->size = 0;
	self->priv = SVG_TOGGLE_BUTTON_GET_PRIVATE (self);
	self->priv->svgpath = NULL;
	self->priv->svgpath2 = NULL;
	self->priv->image = NULL;
	self->priv->pixbuf = NULL;
}


SvgToggleButton*
svg_toggle_button_construct (GType object_type)
{
	return (SvgToggleButton*) g_object_new (object_type, NULL);
}


SvgToggleButton*
svg_toggle_button_new (const char* svgpath)
{
	SvgToggleButton* self = svg_toggle_button_construct (TYPE_SVG_TOGGLE_BUTTON);

	g_return_val_if_fail(svgpath, self);

	SvgToggleButtonPrivate* _btn = self->priv;

	_btn->svgpath = (char*)svgpath;
	_btn->image = (GtkImage*)gtk_image_new_from_pixbuf(NULL);
	gtk_container_add(GTK_CONTAINER(self), (GtkWidget*)_btn->image);

	svg_toggle_button_set_size(self, 16);

#if 0
	if((_btn->pixbuf = get_icon(svgpath, self->size))){
		if(gdk_pixbuf_get_height(_btn->pixbuf) != self->size) gwarn("svg rendered at wrong size: %i.", gdk_pixbuf_get_height(_btn->pixbuf));
		//put the pixbuf into an image, then a button:
		GtkWidget* image = gtk_image_new_from_pixbuf(_btn->pixbuf);
		gtk_container_add(GTK_CONTAINER(self), image);
	}
#endif
	return self;
}


GtkWidget*
svg_toggle_button_new_full (const char* svgpath, const char* svgpath2, int size)
{
	SvgToggleButton* self = svg_toggle_button_new (svgpath);
	self->priv->svgpath2 = (char*)svgpath2;
	svg_toggle_button_set_size(self, size);
	return (GtkWidget*)self;
}


static void
svg_toggle_button_finalize(GObject* obj)
{
	SvgToggleButton* self = SVG_TOGGLE_BUTTON (obj);
	SvgToggleButtonPrivate* _btn = self->priv;

	if(_btn->pixbuf){ g_object_unref(_btn->pixbuf); _btn->pixbuf = NULL; }

	G_OBJECT_CLASS (svg_toggle_button_parent_class)->finalize (obj);
}


void
svg_toggle_button_set_size(SvgToggleButton* self, int size)
{
	SvgToggleButtonPrivate* _btn = self->priv;

	if(size == self->size) return;

	if(_btn->pixbuf) g_object_unref(_btn->pixbuf);

	self->size = size;

	if((_btn->pixbuf = get_icon(_btn->svgpath, self->size))){
		if(gdk_pixbuf_get_height(_btn->pixbuf) != self->size) gwarn("svg rendered at wrong size: %i.", gdk_pixbuf_get_height(_btn->pixbuf));
		gtk_image_set_from_pixbuf(_btn->image, _btn->pixbuf);
	}
}


void
svg_toggle_button_set_active(SvgToggleButton* self, gboolean active)
{
	// like gtk_toggle_button_set_active but without emitting a signal

	//can use gtk_action_block_activate ?
	//GtkToggleButton implements GtkActivatableIface

	SvgToggleButtonPrivate* _btn = self->priv;

	if(active == ((GtkToggleButton*)self)->active) return;

	//gboolean old_state = gtk_widget_get_state ((GtkWidget*)self);

	((GtkToggleButton*)self)->active = active;
	if(active){
		//dbg(0, "active! path2=%s", _btn->svgpath2);
		if(_btn->svgpath2){
			if((_btn->pixbuf = get_icon(_btn->svgpath2, self->size))){
				//dbg(0, "  setting...");
				gtk_image_set_from_pixbuf(_btn->image, _btn->pixbuf);
			}
		}
	}else{
		if((_btn->pixbuf = get_icon(_btn->svgpath, self->size))){
			//dbg(0, "  inactive. setting icon...");
			gtk_image_set_from_pixbuf(_btn->image, _btn->pixbuf);
		}
	}
	gtk_widget_set_state ((GtkWidget*)self, active ? GTK_STATE_ACTIVE : GTK_STATE_NORMAL);
	gtk_widget_queue_draw((GtkWidget*)self);
}


#ifdef OVERRIDE_DEFAULT_PAINT
static gboolean
svg_toggle_button_expose_event (GtkWidget* widget, GdkEventExpose* event)
{
	//SvgToggleButton* self = (SvgToggleButton*) base;

	if (GTK_WIDGET_DRAWABLE (widget)) {
		GtkWidget* child = GTK_BIN (widget)->child;
		GtkButton* button = GTK_BUTTON (widget);

		GtkStateType state_type = GTK_WIDGET_STATE (widget);

		GtkShadowType shadow_type = button->depressed ? GTK_SHADOW_IN : GTK_SHADOW_OUT;

		__gtk_button_paint (button, &event->area, state_type, shadow_type, "togglebutton");

		if (child) gtk_container_propagate_expose (GTK_CONTAINER (widget), child, event);
	}
	return FALSE;
}
#endif


static void
svg_toggle_button_clicked (GtkButton* button)
{
	dbg(1, "...");
	GtkToggleButton* toggle_button = GTK_TOGGLE_BUTTON (button);

	svg_toggle_button_set_active((SvgToggleButton*)button, !toggle_button->active);

	gtk_toggle_button_toggled (toggle_button);

	g_object_notify (G_OBJECT (toggle_button), "active");

	//there is nothing to chain up to;
	//we have re-implemented gtktogglebutton, and gtkbutton does not have a clicked fn (it connects to the signal instead).
}


#if 0
static gboolean
svg_toggle_button_real_button_press_event (GtkWidget* base, GdkEventButton* event)
{
	gboolean result = FALSE;
	//SvgToggleButton* self = (SvgToggleButton*) base;
	return result;
}


static gboolean
svg_toggle_button_real_button_release_event (GtkWidget* base, GdkEventButton* event)
{
	gboolean result = FALSE;
	//SvgToggleButton* self = (SvgToggleButton*) base;
	return result;
}


static gboolean
svg_toggle_button_real_motion_notify_event (GtkWidget* base, GdkEventMotion* event)
{
	gboolean result = FALSE;
	//SvgToggleButton* self = (SvgToggleButton*) base;
	return result;
}
#endif


#ifdef OVERRIDE_DEFAULT_PAINT
//tidied up copy of private function from gtkbutton.c - useful for debugging
static void
__gtk_button_paint (GtkButton* button, const GdkRectangle* area, GtkStateType state_type, GtkShadowType shadow_type, const gchar* main_detail)
{
	GtkBorder default_border;
	GtkBorder default_outside_border;
	gint focus_width;
	gint focus_pad;
	//if(state_type == GTK_STATE_ACTIVE) dbg(0, "state=active"); else dbg(0, "state=%i", state_type);

	if (GTK_WIDGET_DRAWABLE (button)) {
		GtkWidget* widget = GTK_WIDGET (button);
		gint border_width = GTK_CONTAINER (widget)->border_width;

		gboolean interior_focus = false;
		//gtk_button_get_props (button, &default_border, &default_outside_border, NULL, &interior_focus);
		gtk_widget_style_get (GTK_WIDGET (widget), "focus-line-width", &focus_width, "focus-padding", &focus_pad, NULL); 

		gint x = widget->allocation.x + border_width;
		gint y = widget->allocation.y + border_width;
		gint width = widget->allocation.width - border_width * 2;
		gint height = widget->allocation.height - border_width * 2;

		if (GTK_WIDGET_HAS_DEFAULT (widget) && GTK_BUTTON (widget)->relief == GTK_RELIEF_NORMAL) {
			gtk_paint_box (widget->style, widget->window, GTK_STATE_NORMAL, GTK_SHADOW_IN, area, widget, "buttondefault", x, y, width, height);

			x += default_border.left;
			y += default_border.top;
			width -= default_border.left + default_border.right;
			height -= default_border.top + default_border.bottom;
		}
		else if (GTK_WIDGET_CAN_DEFAULT (widget)) {
			x += default_outside_border.left;
			y += default_outside_border.top;
			width -= default_outside_border.left + default_outside_border.right;
			height -= default_outside_border.top + default_outside_border.bottom;
		}

		if (!interior_focus && GTK_WIDGET_HAS_FOCUS (widget)) {
			x += focus_width + focus_pad;
			y += focus_width + focus_pad;
			width -= 2 * (focus_width + focus_pad);
			height -= 2 * (focus_width + focus_pad);
		}

		if (button->relief != GTK_RELIEF_NONE || button->depressed || GTK_WIDGET_STATE(widget) == GTK_STATE_PRELIGHT)
			gtk_paint_box (widget->style, widget->window, state_type, shadow_type, area, widget, "button", x, y, width, height);

		if (GTK_WIDGET_HAS_FOCUS (widget)) {
			gint child_displacement_x;
			gint child_displacement_y;
			gboolean displace_focus;

			gtk_widget_style_get (GTK_WIDGET (widget),
				"child-displacement-y", &child_displacement_y,
				"child-displacement-x", &child_displacement_x,
				"displace-focus", &displace_focus,
				NULL);

			if (interior_focus) {
				x += widget->style->xthickness + focus_pad;
				y += widget->style->ythickness + focus_pad;
				width -= 2 * (widget->style->xthickness + focus_pad);
				height -=  2 * (widget->style->ythickness + focus_pad);
			} else {
				x -= focus_width + focus_pad;
				y -= focus_width + focus_pad;
				width += 2 * (focus_width + focus_pad);
				height += 2 * (focus_width + focus_pad);
			}

			if (button->depressed && displace_focus) {
				x += child_displacement_x;
				y += child_displacement_y;
			}

			gtk_paint_focus (widget->style, widget->window, GTK_WIDGET_STATE (widget), area, widget, "button", x, y, width, height);
		}
	}
}
#endif

