/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#define __window_c__
#include "global.h"
#include <gdk/gdkkeysyms.h>

#include "model/am_utils.h"
#include "support.h"
#include "windows.h"
#include "panels/arrange.h"
#include "toolbar.h"
#include "settings.h"
#include "statusbar.h"
#include "shell.h"
#include "icon.h"

extern SMConfig* config;

extern void       register_panels            ();

static void       windows__count_table_print (GHashTable*);

typedef struct
{
	GType         gtype;
	void          (*init)();
} DefaultPanelClass;

#include "register.c"

AyyiPanelClass window_classes[WINDOW_TYPE_MAX];

GList* windowlist; // list of AyyiPanel*, one for each open window panel. DEPRECATED in favour of shell_list->panels
                   // To iterate over panels, use panel_foreach


void
windows_init()
{
	windowlist = NULL; // initialise the list of open windows
	shell_list = NULL;

	// list of panel types hashed by GType.
	app->panel_classes = g_hash_table_new(g_int_hash, g_int_equal);

	register_panels();

	int i; for(i=1;i<G_N_ELEMENTS(default_panel_class_types);i++){
		DefaultPanelClass* dp = &default_panel_class_types[i];
		if(!dp->gtype) gerr("gtype not set! %i", i);
		if(dp->init) dp->init();

		g_hash_table_insert(app->panel_classes, &dp->gtype, &window_classes[i]);
	}

#if 0
	GHashTableIter iter;
	gpointer key, value;
	g_hash_table_iter_init (&iter, app->panel_classes);
	while (g_hash_table_iter_next (&iter, &key, &value)){
		WindowClass* window_class = value;
		PanelType panel_type = *((int*)key);
		dbg(0, "  %i=%p", *((int*)key), value);
	}
#endif
#if 0
	void each(gpointer key, gpointer value, gpointer user_data)
	{
		PanelType type = *((int*)key);
		dbg(0, "  %s", g_type_name(*((int*)key)));
	}
	g_hash_table_foreach(app->panel_classes, each, NULL);
#endif

	windows__add_type(default_panel_class_types[TYPE_METER].gtype);

	toolbar_init();
}


void
windows__add_type(PanelType type)
{
	// only needed for dynamic panel types (eg plugins)

	dbg(2, "%i", type);
	g_type_class_unref (g_type_class_ref (type));

	PanelType* dtype = g_malloc(sizeof(int));
	*dtype = type;
	g_hash_table_insert(app->panel_classes, dtype, NULL);
}


static char*
dock_object_get_depth()
{
	static char str[][3] = {" * "};
	return str[0];
}


#define DOCK_NAME(A) (A ? (A->name ? A->name : G_OBJECT_CLASS_NAME(G_OBJECT_GET_CLASS(A))) : "null")
struct _done
{
    ConfigWindow*  cw;
    GdlDockObject* obj;
};

#if 0
  static int** docktree_lookup(GdlDockObject* dock_object, GList* done_list)
  {
    GList* l = done_list;
    for(;l;l=l->next){
      struct _done* d = l->data;
      //dbg(0, "  d=%p", d);
      if(d->obj == dock_object) return (int**)d->cw->docktree;
    }
    return NULL;
  }
#endif

	static gboolean dockpath_is_same(int** path1, int** path2, int n)
	{
		gboolean same = true;
		int i=0; for(;i<16;i++){
			if(i == n) break;
			if(path1[i] != path2[i]) return false;
		}
		return same;
	}

	struct _aa {
		int*           depth;
		int            idx;
		ConfigWindow*  panel;
		gboolean*      finished;
		GdlDockObject* target;
		GList*         done_list;
		int**          pathinfo;
	};
	static void get_children(GdlDockObject* object, ConfigWindow* panel, gboolean* finished, int** pathinfo, int* depth, int idx, GdlDockObject* target, GList* done_list);

	static void children_foreach(GdlDockObject* object, struct _aa* a)
	{
		a->pathinfo[*(a->depth)]++;
		gboolean path_match = dockpath_is_same((int**)a->panel->docktree, (int**)a->pathinfo, a->idx + 1);
		if(path_match){
			*(a->finished) = true;
			a->target = object;
			dbg(0, "found!");
			if(*(a->depth) > a->idx - 1){ //cough!
				gwarn("found item is at wrong depth. using parent...");
				a->target = gdl_dock_object_get_parent_object(object);
				g_return_if_fail(GDL_IS_DOCK_OBJECT(a->target));
			}
		}
		else if(1 /*|| pathinfo[*depth] == panel->docktree[*depth]*/) get_children(object, a->panel, a->finished, a->pathinfo, a->depth, a->idx, a->target, a->done_list);
	}

    static void get_children(GdlDockObject* object, ConfigWindow* panel, gboolean* finished, int** pathinfo, int* depth, int idx, GdlDockObject* target, GList* done_list)
    {
      // this fn generates a "generating trampoline in object (requires executable stack)" warning with gcc 4.1
      // because of the recursive fn call
      // which indicates security implications and non-portability.
      // however apparently the warning isnt given by gcc 4.3, so the code is not being changed for now.
      // ** actually warning is still there with 4.3.2

      g_return_if_fail(!*finished);

      (*depth)++;
      if(*depth == idx){
        //if target is GdlDockPaned, it won't be in the done_list, and won't have a docktree.
        //-using the parent doesnt always work because there isnt always one with a docktree.
#if 0
        int** dt = docktree_lookup(target, done_list);
        if(!dt){
        }
#endif
        //dbg(0, "using target=%s %s", DOCK_NAME(object), dock_info_to_string((int**)pathinfo));
      }

      if(!*finished){
        if(gdl_dock_object_is_compound(object)){
          struct _aa aa = {depth, idx, panel, finished, target, done_list, pathinfo};
          gtk_container_foreach(GTK_CONTAINER (object), (GtkCallback)children_foreach, &aa);
        }
        else dbg(2, "leaf. depth=%i.", *depth);
      }
      (*depth)--;
    }

#ifndef GDLDOCK_XML
#if 0
AyyiShell*
window__open(int shell_instance, ConfigWindow* config_window)
{
  //create a panel/window, or reuse an existing one.
  //
  //@param shell_instance - the shell_list position of an existing shell to add to.
  //                        Set this to zero at application start, or you if dont want to 'close' existing windows.
  //@param config_window  - nested list of list of Panels to add to the shell.
  //
  //-the window may not yet be shown.
  //-fn is subject to change for docking panes, and may be merged with shell__add_panel_....

  static GList* done_list; done_list = NULL; //list of "struct _done*"

  void add_done(ConfigWindow* cw, GdlDockObject* obj)
  {
    struct _done* d = g_new(struct _done, 1);
    d->cw = cw;
    d->obj = obj;
    done_list = g_list_append(done_list, d);
  }

  void free_done(){
    GList* l = done_list;
    for(;l;l=l->next) g_free(l->data);
    g_list_clear(done_list);
  }

  GdlDockObject* find_previous_panel(GList* configlist, int panel_depth)
  {
    //find already-restored panel with given panel_depth.
    //!!deprecated?

    ConfigWindow* cw = NULL;
    GList* l = configlist;
    for(;l;l=l->prev){
      if(((ConfigWindow*)l->data)->depth == panel_depth){ cw = (ConfigWindow*)l->data; break; }
    }
    if(!cw) dbg(2, "configWindow not found for depth=%i", panel_depth);
    if(!cw) return NULL;

    //now find the configwindow in the dock...
    GList* k = done_list;
    for(;k;k=k->next){
      struct _done* d = k->data;
      dbg(2, "  d=%p", d);
      if(d->cw == cw) return d->obj;
    }
    gwarn("type%i not found in done_list!", cw->type);
    return NULL;
  }
  GList* panels = config_window->child_panels;

  ConfigWindow* find_cw_in_donelist(GdlDockObject* object)
  {
    GList* l = done_list;
    for(;l;l=l->next){
      struct _done* d = l->data;
      if(d->obj == object) return d->cw;
    }

    //object is probably not a panel and doesnt have a ConfigWindow. Find the most appropriate:
    static GList* k; k = NULL;
    void foreach(GdlDockObject* object, gpointer user_data)
    {
      k = g_list_append(k, object);
    }
    gtk_container_foreach(GTK_CONTAINER (object), (GtkCallback)foreach, NULL);
    if(k){
      ConfigWindow* c = find_cw_in_donelist(k->data);
      if(c) return c;
    }

    return NULL;
  }

  int get_docktree_size(int** docktree)
  {
    g_return_val_if_fail(docktree, 0);

    int i=0; for(;i<16;i++){
      if(!docktree[i]) return i;
    }
    return -1;
  }

  gboolean has_sibling(ConfigWindow* root, GdlDockObject* dock_object, int** docktree)
  //gboolean has_sibling(struct _AyyiDockInfo* nfo)
  {
    int size = get_docktree_size(docktree);
    GList* p = root->child_panels;
    for(;p;p=p->next){
      ConfigWindow* panel = p->data;
      if((int**)panel->docktree == docktree) continue;
      if(get_docktree_size((int**)panel->docktree) == size){
        int i=0; for(;i<size-1;i++){
          if(panel->docktree[i] != (int)docktree[i]) return false;
        }
        return true;
      }
    }
    return false;
  }

  int find_split_point(ConfigWindow* panel, ConfigWindow* previous)
  {
    //find the level at which the index goes from 1 to 2

    int i=15; for(;i>0;i--){
      if(panel->docktree[i] == 2){
        if(previous->docktree[i] == 1) return i;
        break;
      }
    }
    return -1;
  }

  GdlDockObject* get_dock_object_from_docktree(ConfigWindow* panel, int idx)
  {
    //find the dockitem at the depth given by idx that 'matches' the panel docktree.

    static int depth; depth = 0;
    static gboolean finished; finished = false;
    static int pathinfo[16]; memset(pathinfo, 0, 16 * sizeof(int));
    static GdlDockObject* target; target = NULL;
    pathinfo[depth]++;

    if(done_list){
      struct _done* d = done_list->data;
      g_return_val_if_fail(d->obj, NULL);
      GdlDock* dock = gdl_dock_object_get_toplevel(d->obj);
      GdlDockObject* root = dock->root;
      if(!gdl_dock_object_is_compound(root)) dbg(2, "root not compound dockobject. (%s)", root->name);

      get_children(root, panel, &finished, (int**)pathinfo, &depth, idx, target, done_list);

      if(!finished) dbg(2, "didnt finish.");
    }

    dbg(2, "%s idx=%i target=%s", dock_info_to_string((int**)pathinfo), idx, DOCK_NAME(target));
    return target;
  }

  AyyiShell* shell = NULL;
  //check if window already exists:
#ifdef WHAT_IS_THIS_FOR___FIX_OR_REMOVE
  //(1st window is always an Arrange window, and 2nd is sometimes a Log window. Both of these are always unaffected by a song load)
  int n_reserved_instances = window_classes[TYPE_LOG].instances ? 2 : 1;
#endif
  if(/*shell_instance > n_reserved_instances && */shell_instance <= g_list_length(shell_list)){
    //shell = windows__get_from_instance(type, instance)->shell;
    shell = g_list_nth_data(shell_list, shell_instance - 1);
    dbg(1, "using existing shell. shell_instance=%i n_shells=%i", shell_instance, g_list_length(shell_list));
  }
  else {
    dbg(1, "existing shell not found. shell_instance=%i n_shells=%i", shell_instance, g_list_length(shell_list));
    shell = shell__new(NULL);

    //TODO when is the optimum time to maximise a window?
    gboolean maximise(gpointer _window)
    {
      gtk_window_maximize((GtkWindow*)_window);
      return false;
    }
    if(config_window->maximised) g_idle_add(maximise, shell->widget);
  }
  g_return_val_if_fail(shell, false);
  //gtk_widget_set_size_request(shell->widget, config_window->width, config_window->height);

  //GHashTable* instances = windows__new_count_table();
  GHashTable* instances = windows__new_count_table_from_shell(shell);

  /*
  GList* existing_panels = shell__get_dock_items(shell);
  GList* p = existing_panels;
  for(;p;p=p->next){
    GdlDockObject* object = p->data;
    windows__count_table_add(instances, G_OBJECT_TYPE(object));
    //windows__print_dock_object(object, "  ");
  }
  g_list_free(existing_panels);
  */

  //windows__count_table_print (instances);

  if(!panels){
    //probably using a config parser that doesnt support nesting.
    //Make a single child panel.

    ConfigWindow* window = &g_array_index(config->windows, ConfigWindow, shell_instance - 1);

    dbg(2, "creating default panel...");
    ConfigWindow* panel = g_new0(ConfigWindow, 1);
    *panel = *window;
    panels = g_list_prepend(NULL, panel);
  }

  //check for child panels:
  if(panels){
    dbg(2, "child panels to make: %i", g_list_length(panels));
    int child_num = 0;
    //GtkOrientation previous_orientation = 255;
    GtkWidget* previous_panel = (GtkWidget*)shell->dock;
    int previous_depth = -1;
    ConfigWindow* previous_cw = NULL;
    GList* l = panels;
    for(;l;l=l->next, child_num++){
      ConfigWindow* panel = l->data;
      if(!panel->type) continue;
      AyyiPanelClass* klass = g_type_class_peek(panel->type);
      int n_instances = GPOINTER_TO_INT(g_hash_table_lookup(instances, &panel->type));
      if(n_instances){
        dbg(1, "%i: have instance. %s. will not add.", child_num, g_type_name(panel->type));
/*
        windows__count_table_remove_instance(instances, panel->type); //why?
*/
      } else {

        if(_debug_) printf("%i\n", child_num);
        dbg(1, "%i: creating panel: %s%s%s", child_num, bold, g_type_name(panel->type), white);

        //should we re-use an existing panel?
		AyyiPanel* existing = windows__get_from_instance(panel->type, n_instances);
        dbg(1, "existing(%i)=%p instance=%i", n_instances, existing, n_instances);

        //dbg(0, "depth=%i panlorientation=%s", panel->depth, ((panel->orientation == GTK_ORIENTATION_VERTICAL) ? "GTK_ORIENTATION_VERTICAL" : "GTK_ORIENTATION_HORIZONTAL"));
        //dbg(0, "depth=%i prevorientation=%s", panel->depth, ((previous_orientation == GTK_ORIENTATION_VERTICAL) ? "GTK_ORIENTATION_VERTICAL" : "GTK_ORIENTATION_HORIZONTAL"));
#if 0
        GtkOrientation orientation = ((previous_orientation < 255) ? previous_orientation : panel->orientation);
#endif
        GtkOrientation orientation = panel->orientation;
        //dbg(0, "previous=%i orientation=%i", previous_orientation, orientation);
        GdlDockObject* parent_object = NULL;
        parent_object = GDL_DOCK_OBJECT(previous_panel);
        dbg(2, "%s has_sibling: %s", dock_info_to_string((int**)panel->docktree), has_sibling(config_window, parent_object, (int**)panel->docktree) ? "true" : "false");
        if(panel->depth < previous_depth){
          parent_object = find_previous_panel(l->prev, panel->depth);
          if(!parent_object) parent_object = (GdlDockObject*)shell->dock;
          dbg(2, "previous_parent=%s", parent_object ? GDL_DOCK_OBJECT(previous_panel)->name: "<null>");
        }
        //dbg(0, "request position: x=%i y=%i", panel->x, panel->y);

        int split_pt = find_split_point(panel, previous_cw);
        dbg(3, "split_pt=%i", split_pt);
        if(split_pt > 0){
          //split the dock at this point.
          GdlDockObject* splitter = get_dock_object_from_docktree(panel, split_pt);
          if(splitter) parent_object = splitter;
        }

        //get the object to be split:
        //case 1: previous_panel can be split directly.
        //case 2: previous_panel may have a sibling, and we need to split the parent of previous_panel.
        //        -we need to inspect the nfo array to determine this. TODO save the nfo array.

        //get the geometry of the object to be split:

        if(parent_object){
          if(GDL_IS_DOCK_ITEM(parent_object)){
            if(panel->depth != previous_depth){
              orientation = GDL_DOCK_ITEM(parent_object)->orientation;
              dbg(2, "using parent orientation: %s - %s", parent_object->name, get_orientation_string(orientation));
            }
          }
          else if(false && panel->depth > 1) gwarn("parent is not dockitem. dont have parent orientation.");
        }
        else gwarn("using child panel orientation - this is likely to be wrong.");

        //ignore orientation calculations above and get them by inspecting saved geometry.
        ConfigWindow* parent_config = find_cw_in_donelist(parent_object);
        if(child_num){
          //ConfigWindow* c = find_cw_in_donelist(parent_object);
          if(parent_config){
            //FIXME still some work to do here. Currently only works for splits in one dimension.
            //int child_x = panel->x;
            int child_y = panel->y;
            //int parent_x = parent_config->x;
            int parent_y = parent_config->y;
            orientation = (child_y != parent_y) ? GTK_ORIENTATION_VERTICAL : GTK_ORIENTATION_HORIZONTAL;
          } else gerr("failed to find configwindow for parent panel!");
        }

        Size g_config_size = {0, 0};
        g_config_size.width = panel->width;
        g_config_size.height = panel->height;
        //TODO have to increase panel size before we split. Add dimensions of new panel in direction of split. tidy up and copy for GTK_ORIENTATION_VERTICAL
        // (needed because if the new panel is not smaller than its parent object, gdl will not be able to set the position of the divider)
        if(orientation == GTK_ORIENTATION_HORIZONTAL){
          int pw = -1;
          if(G_OBJECT_TYPE(parent_object) == GDL_TYPE_DOCK){
            if(GDL_DOCK(parent_object)->root)
              g_object_get(GDL_DOCK(parent_object)->root, "preferred-width", &pw, NULL);
          }else
            g_object_get(parent_object, "preferred-width", &pw, NULL);
          dbg(3, "preferred-width=%i", pw);
          GtkRequisition r = ((GtkWidget*)parent_object)->requisition;
          if(G_OBJECT_TYPE(parent_object) == GDL_TYPE_DOCK_PANED){
            GList* children = gtk_container_get_children(GTK_CONTAINER(parent_object));
            if(children){
              GtkWidget* child = children->data;
              r = child->requisition;
              dbg(2, "using child requistion: %i", r.width);
            }
          }
          else dbg(2, "parent_object=%s", G_OBJECT_CLASS_NAME(G_OBJECT_GET_CLASS(parent_object)));
          int new_width = g_config_size.width + MAX(pw, (parent_config ? parent_config->width : r.width));
          //new_width = MAX(pw, new_width);
          dbg(2, "setting width: old_parent_width=%i parentc=%i new_req=%i <--------------------------", r.width, parent_config ? parent_config->width : -1, new_width);
          //gtk_widget_set_size_request((GtkWidget*)parent_object, new_width, -1);
          if(G_OBJECT_TYPE(parent_object) != GDL_TYPE_DOCK){
            g_object_set (parent_object, "preferred-width", new_width, NULL);
            GTK_WIDGET(parent_object)->requisition.width = new_width; //TODO check this is really needed. ideally only preferred-width would be used.
            GdlDockObject* pp = gdl_dock_object_get_parent_object(GDL_DOCK_OBJECT(parent_object));
            if(pp){
              dbg(2, "object='%s' (%s)", pp->name, G_OBJECT_CLASS_NAME(G_OBJECT_GET_CLASS(pp)));
              GTK_WIDGET(pp)->allocation.width = new_width; //used to notify gdl of the required size.
              //gtk_widget_set_size_request(GTK_WIDGET(pp), new_width, -1);
            }
          }else{
            GObject* root = G_OBJECT(GDL_DOCK(parent_object)->root);
            if(root) g_object_set (root, "preferred-width", new_width, NULL);
          }
        }

        dbg(2, "depth=%i-->%i orientation=%s", previous_depth, panel->depth, get_orientation_string(orientation));

        //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        if(klass->new) previous_panel = klass->new(shell, &g_config_size, orientation); else gwarn ("class->new is not set (%p) type=%i", klass->new, panel->type);
        //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

        add_done(panel, GDL_DOCK_OBJECT(previous_panel));

        //shell isnt yet named, so give it a name:
        char shell_name[64];
        config_make_group_name(klass, klass->instances-1, shell_name);
        strcpy(shell_name + strlen(shell_name), " Shell");
        char** name = &GDL_DOCK_OBJECT(shell->dock)->name;
        if(strstr(*name, "hell ")){
          dbg(3, "renaming dock to %s%s%s", bold, shell_name, white);
          g_free(*name);
          *name = g_strdup(shell_name);
        }
      }
      //previous_orientation = panel->orientation;
      previous_depth = panel->depth;
      previous_cw = panel;
    }
  }

  free_done();
  //gdl_dock_print_recursive((GdlDockMaster*)((GdlDockObject*)shell->dock)->master);
  return shell;
}
#endif
#endif


AyyiPanel*
windows__get_panel_from_widget(GtkWidget* widget)
{
	// Return the AyyiPanel for the window containing the given widget.

	// note that this doesnt work for context menus.

	g_return_val_if_fail(widget, NULL);
	g_return_val_if_fail(GTK_IS_WIDGET(widget), NULL);

	GtkWidget* window = gtk_widget_get_toplevel(widget);

	/*
	GtkWindow* t_parent = gtk_window_get_transient_for((GtkWindow*)window);
	if(t_parent) dbg(0, "has transient parent!"); else dbg(0, "transient not set"); //should it be set?
	*/
	GtkWindowType type;
	g_object_get(window, "type", &type, NULL);
	if(type == GTK_WINDOW_POPUP) return NULL;

	// new code that works for windows with multiple panels
	// TODO better to use gtk_widget_get_ancestor() ?
	{
		GtkWidget* parent = widget;
		do {
			dbg(3, "  widget=%s %i", G_OBJECT_CLASS_NAME(G_OBJECT_GET_CLASS(parent)), AYYI_IS_PANEL(parent));
			if(AYYI_IS_PANEL(parent)){
				AyyiPanel* panel = (AyyiPanel*)parent;
				char type[AYYI_SHORT_NAME_MAX]; ayyi_panel_print_type(panel, type);
				dbg(3, "found panel '%s %i' %s", type, ayyi_panel_get_instance_num(panel), GDL_DOCK_OBJECT(parent)->name);
				return panel;
			}
		}
	    while((parent = gtk_widget_get_parent(parent)));
	}
	dbg(0, "falling back to (unreliable) old method...");

//#ifdef USE_OLD_CODE_THAT_DOESNT_WORK_WITH_DOCKED_PANELS
	g_return_val_if_fail(window, NULL);
	GList* l = windowlist;
	for(;l;l=l->next){
		AyyiPanel* panel = l->data;
		if(panel->shell->widget == window){
			if(_debug_ > 1 && shell__get_n_panels(panel->shell) > 1){
				gwarn("FIXME doesnt make sense for shells with multiple panels.");
			}
			return panel;
		}
	}
//#endif

	// So far not found. Lets deal with context menus
	// (currently some menu items have the smwin pointer embedded in them to get round the toplevel problem)
	gpointer data = g_object_get_data(G_OBJECT(widget), "smwin");
	for(l=windowlist;l;l=l->next){
		AyyiPanel* panel = l->data;
		if(G_OBJECT_TYPE(panel) == ayyi_mixer_win_get_type()){
			if(data == panel) return panel;
		}
	}

	windows__print();
	gwarn ("window not found. widget=%p gtk_window=%p windowcount=%i", widget, window, g_list_length(windowlist));
	return NULL;
}


/*
 *  Return the AyyiPanel for the currently active toplevel window.
 */
AyyiPanel*
windows__get_active()
{
	//under certain circumstance, no window may be active:
	// -a dnd operation from another application.
	// -a popup menu is open.

	int active = 0;

	//how many active windows are there?
	int count=0;
	AyyiShell* active_shell = NULL;
	gboolean realised = true;
	GList* l = shell_list;
	for(;l;l=l->next){
		AyyiShell* shell = l->data;
		if(!GTK_WIDGET_REALIZED(shell->widget)){ realised = false; continue; }
		g_object_get(shell->widget, "has-toplevel-focus", &active, NULL);
		if(active){
			active_shell = shell;
			count++;
		}
	}
	if(count != 1) gwarn ("%i windows active.", count);
	if(!realised){ gwarn ("not realized!"); return NULL; }
	else g_return_val_if_fail(active_shell, NULL);

	GtkWidget* focussed = gtk_window_get_focus(GTK_WINDOW(active_shell->widget));
	if(!focussed){
		gwarn ("no widget focussed. active_shell='%s'", GDL_DOCK_OBJECT(active_shell->dock)->name);

		GList* panels = shell__get_panels(active_shell);
		if(panels){
			if(g_list_length(panels) == 1);
			AyyiPanel* panel = panels->data;
			g_list_free(panels);
			return panel; //return the shell's only panel.
		}

		return app->active_panel;
	}

	AyyiPanel* panel = windows__get_panel_from_widget(focussed);
	//find out which Panel the widget belongs to:
	/*
	AyyiPanel* panel = NULL;
	for(l=windowlist;l;l=l->next){
	panel = l->data;
	g_object_get(panel->shell->widget, "has-toplevel-focus", &active, NULL);
	if(active) break;
	}
	*/
	if(!active){ gwarn ("active window not found. panel=%p", panel);/* return NULL;*/ }
	return panel;
}


AyyiPanel*
windows__get_first(PanelType type)
{
	//return the first open windows of the given type

	GList* l = windowlist;
	for(;l;l=l->next){
		AyyiPanel* panel = l->data;
		if(G_OBJECT_TYPE(panel) == type) return panel;
	}
	dbg (2, "window not found. type=%i", type);
	return NULL;
}


GList*
windows__get_by_type(PanelType dpt)
{
	//return a newly allocating list of open windows of the given type.

	GList* ret = NULL;
	GList* l = windowlist;
	for(;l;l=l->next){
		AyyiPanel* panel = l->data;
		if(G_OBJECT_TYPE(panel) == dpt) ret = g_list_append(ret, panel);
	}
	return ret;
}


int
windows__get_n_instances(PanelType type)
{
	// this is only useful for debugging.
	// -normally you would use class->instances intead.

	int n = 0;
	GList* l = windowlist;
	for(;l;l=l->next){
		AyyiPanel* panel = l->data;
		PanelType dtype = G_OBJECT_TYPE(panel);
		if(dtype == type) n++;
	}
	return n;
}


PanelType
windows__panel_type_from_str(const char* str)
{
	// return the panel class type for the given string.
	// The string can contain multiple words, but only the first word is matched.

	g_return_val_if_fail(str, 0);

	GHashTableIter iter;
	gpointer key, value;
	g_hash_table_iter_init (&iter, app->panel_classes);
	while (g_hash_table_iter_next (&iter, &key, &value)){
		PanelType panel_type = *((int*)key);
		//dbg(0, "  type=%s", g_type_name(panel_type));
		AyyiPanelClass* k = g_type_class_peek(panel_type);
		if(!k){ gwarn("no class for type=%s", g_type_name(panel_type)); continue; }

		char* found = g_strstr_len(str, 32, k->name);
		if(found == str){
			return G_TYPE_FROM_CLASS(k);
		}
	}
	if(_debug_) gwarn("not found! %s", str);

	return 0;
}


bool
windows__verify_pointer(AyyiPanel* panel, PanelType type)
{
	//@param type - if this is non-zero, we also check the window type matches.

	g_return_val_if_fail(panel, false);

	AyyiPanelClass* panel_gclass = g_type_class_peek(type);

	bool found = false;
	AyyiPanel* target = panel;
	panel_foreach {
		if(panel == target){ found = true; break; }
	} end_panel_foreach;
	if(!found){ gerr ("window not found in list."); return false; }

	if(!type) return true;

	if(G_OBJECT_TYPE(panel) == type) return true;

	gerr ("not verified. wrong type. type=%s expected=%s", AYYI_PANEL_GET_CLASS(panel)->name, panel_gclass->name);
	return false;
}


AyyiPanel*
windows__get_from_instance_string(const char* str)
{
	//@str should be of form: "$window_class_name-$integer"

	char* separator = strstr(str, " ");
	if(separator){
		char* instance_str = separator + 1;
		int instance_num = atoi(instance_str);
		if(instance_num){
			char window_type[32];
			int len = separator - str;
			strncpy(window_type, str, len);
			window_type[len] = '\0';
			dbg(3, "type=%s", window_type);

			panel_foreach {
				int i = ayyi_panel_get_instance_num(panel);
				dbg(3, "  (%s) %i==%i %s", str, i, instance_num, AYYI_PANEL_GET_CLASS(panel)->name);
				if(i == instance_num && !strcmp(AYYI_PANEL_GET_CLASS(panel)->name, window_type)) return panel;
			} end_panel_foreach;
		}
		else dbg(0, "failed to get instance num.");
	}
	else dbg(0, "failed to find separator. str='%s'", str);
	dbg(0, "failed.");
	return NULL;
}


AyyiPanel*
windows__get_from_instance(GType type, int instance_num)
{
	panel_foreach {
		if(G_OBJECT_TYPE(panel) != type) continue;
		int i = ayyi_panel_get_instance_num(panel);
		if(i == instance_num) return panel;
	} end_panel_foreach;

	return NULL;
}


GHashTable*
windows__new_count_table()
{
	GHashTable* instances = g_hash_table_new(g_int_hash, g_int_equal);

	//add an instances entry for each window type:
	GHashTableIter iter;
	gpointer key, value;
	g_hash_table_iter_init (&iter, app->panel_classes);
	while (g_hash_table_iter_next (&iter, &key, &value)){
		g_hash_table_insert(instances, key, 0); //key is PanelType
	}
	return instances;
}


GHashTable*
windows__new_count_table_from_shell(AyyiShell* shell)
{
	GHashTable* instances = windows__new_count_table();

	GList* existing_panels = shell__get_dock_items(shell);
	GList* p = existing_panels;
	for(;p;p=p->next){
		GdlDockObject* object = p->data;
		windows__count_table_add(instances, G_OBJECT_TYPE(object));
		//windows__print_dock_object(object, "  ");
	}
	g_list_free(existing_panels);
	return instances;
}


int
windows__count_table_add(GHashTable* table, PanelType type)
{
	int n_instances = GPOINTER_TO_INT(g_hash_table_lookup(table, &type));
	n_instances++;
	g_hash_table_insert(table, &type, GINT_TO_POINTER(n_instances));
	return n_instances;
}


void
windows__count_table_remove_instance(GHashTable* instances, PanelType type)
{
	PF;
	int n_instances = GPOINTER_TO_INT(g_hash_table_lookup(instances, &type));
	if(!n_instances) return;
	n_instances--;
	g_hash_table_insert(instances, &type, &n_instances);
}


static void
windows__count_table_print(GHashTable* table)
{
	GHashTableIter iter;
	gpointer key, value;
	g_hash_table_iter_init (&iter, table);
	while (g_hash_table_iter_next (&iter, &key, &value)){
		int instances = GPOINTER_TO_INT(value);
		PanelType panel_type = *((int*)key);
		dbg(0, "  %i=%i %s", panel_type, instances, g_type_name(panel_type));
	}
}


void
windows__ensure(PanelType dtype)
{
	// make sure there is at least one window open of the given type.

	if(!windows__get_n_instances(dtype))
		shell__new_floating(g_type_class_peek(dtype));
}


AyyiShell*
lookup_shell_by_widget(GtkWindow* window)
{
	GList* l = shell_list;
	for(;l;l=l->next){
		AyyiShell* s = l->data;
		if(s->widget == (GtkWidget*)window) return s;
	}
	return NULL;
}


/*
AyyiPanelClass*
window__lookup_class_by_static_type(StaticPanelType w)
{
	GHashTableIter iter;
	gpointer key, value;
	g_hash_table_iter_init (&iter, app->panel_classes);
	while (g_hash_table_iter_next (&iter, &key, &value)){
		WindowClass* window_class = value;
		PanelType panel_type = *((int*)key);
		if(window_class){
			//dbg(0, "  key=%i", *((int*)key));
			dbg(0, "  val=%p", value);
		}
	}
}
*/

AyyiPanelClass* panel_class_lookup_by_accel(int accel)
{
	GHashTableIter iter;
	gpointer key, value;
	g_hash_table_iter_init (&iter, app->panel_classes);
	while (g_hash_table_iter_next (&iter, &key, &value)){
		PanelType panel_type = *((int*)key);
		AyyiPanelClass* class = g_type_class_peek(panel_type);
		if(class->accel == accel) return class;
	}
	gwarn("not found!");
	return NULL;
}


void
windows__print()
{
	PF;
	printf("     type                  panel shellwdgt\n");
	int w = 0;
	char type[AYYI_SHORT_NAME_MAX];

	panel_foreach {
		ayyi_panel_print_type(panel, type);
		printf("  %i: %02i %11s %p %p\n", w, (int)G_OBJECT_TYPE(panel), type, panel, panel->shell->widget);
		w++;
	} end_panel_foreach;
}


void
windows__print_dock()
{
	GList* i = gdl_dock_get_named_items(app->master_dock);
	if(i){
		UNDERLINE;
		PF;
		printf("\n");
		printf("           type container           name       longname b    orientation     ph    parent\n");
		char shell_name[64] = "";
		GList* s = shell_list;
		for(;s;s=s->next){
			AyyiShell* shell = s->data;
			if(!GDL_IS_DOCK_OBJECT(shell->dock)){ printf("    !! shell_list corrupted. shell=%p\n", shell); continue; }

			sprintf(shell_name, "%s", GDL_DOCK_OBJECT(shell->dock)->name);
			GList* items = shell__get_dock_items(shell);
			if(items){
				GList* i = items;
				printf(" shell='%s' n_items=%i\n", shell_name, g_list_length(i));
				for(;i;i=i->next){
					if(GDL_IS_DOCK_OBJECT(i->data)){
						GdlDockObject* object = GDL_DOCK_OBJECT(i->data);
						windows__print_dock_object(object, "  ");
					}
					else printf("    not dock object! shell__get_dock_items() returned bad item\n");
				}
				g_list_free(items);
			}
		}
		UNDERLINE;
		g_list_free(i);
	}
	else dbg(0, "empty.");
}


void
windows__print_dock_item(GdlDockItem* item, const char* text)
{
	// note: see also gdl_dock_item_print()

	g_return_if_fail(item);
	GdlDockObject* object = &item->object;
	g_return_if_fail(object);
	dbg(0, "%s: %s %s", text, object->name, object->long_name);
}


void
windows__print_dock_object(GdlDockObject* object, const char* text)
{
	g_return_if_fail(object);

	char* depth = dock_object_get_depth();

	char type[32] = "<unknown type>";
	char behaviour[32] = "";
	char orientation_str[64] = "";
	char placeholder_name[64] = "";
	if(GDL_IS_DOCK_ITEM(object)){
		GdlDockItem* item  = (GdlDockItem*)object;

		strcpy(type, "DockItem");
		sprintf(behaviour, "%i", item->behavior);

		GtkOrientation orientation;
		g_object_get(item, "orientation", &orientation, NULL);
		strcpy(orientation_str, get_orientation_string(orientation));

		GList* ph = gdl_dock_item_get_placeholders(item);
		if(ph){
			if(ph->data){
				GdlDockObject* placeholder_object = GDL_DOCK_OBJECT(ph->data);
				if(placeholder_object) sprintf(placeholder_name, "%s", placeholder_object->name);
				else sprintf(placeholder_name, "<none>");
			}
			g_list_free(ph);
		}
	}

	GtkContainer* container = &object->container;

	printf("%s %s %8s %p %14s %14s %s %14s %s %p\n", text, depth, type, container, object->name ? object->name : "<Unnamed>", object->long_name, behaviour, orientation_str, placeholder_name, gdl_dock_object_get_parent_object(object));
}


void
shell_never()
{
	windows__count_table_print(NULL);
}


