/*
  This file is part of the Ayyi Project. http://www.ayyi.org
  copyright (C) 2004-2017 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#ifndef __window_h__
#define __window_h__
#include <gdl/gdl.h>
#include <panels/panel.h>

GList* shell_list; // list of AyyiShell* - toplevel windows.

#define STRLEN_WIN_ID 16


struct _ConfigWindow //used for both windows and Panels
{
  PanelType      type;
  char*          config_group;
  int            width;
  int            height;
  int            x;
  int            y;
  gboolean       maximised;
  int            depth;
  int            docktree[16];
  GtkOrientation orientation;
  GList*         params;       // ConfigParamValue*
  GList*         child_panels; // recursive list of ConfigWindow*
};


void             windows_init                   ();
void             windows__add_type              (PanelType);
AyyiPanel*       windows__get_first             (PanelType);
GList*           windows__get_by_type           (PanelType);

GHashTable*      windows__new_count_table       ();
GHashTable*      windows__new_count_table_from_shell(AyyiShell*);
int              windows__count_table_add       (GHashTable*, PanelType);

int              windows__get_n_instances       (PanelType);
AyyiPanel*       windows__get_panel_from_widget (GtkWidget*);
AyyiPanel*       windows__get_active            ();
PanelType        windows__panel_type_from_str   (const char*);
gboolean         windows__verify_pointer        (AyyiPanel*, PanelType);
AyyiPanel*       windows__get_from_instance_string(const char* str);
AyyiPanel*       windows__get_from_instance     (GType, int instance);
void             windows__ensure                (PanelType);

AyyiShell*       lookup_shell_by_widget         (GtkWindow*);

AyyiPanelClass*  panel_class_lookup_by_accel    (int);

void             windows__print                 ();
void             windows__print_dock            ();
void             windows__print_dock_item       (GdlDockItem*, const char* text);
void             windows__print_dock_object     (GdlDockObject*, const char* text);

#ifndef GDLDOCK_XML
#if 0
AyyiShell*       window__open               (int instance, ConfigWindow*);
#endif
#endif

#endif //__window_h__
