/**
* +----------------------------------------------------------------------+
* | This file is part of the Ayyi project. http://www.ayyi.org           |
* | copyright (C) 2004-2017 Tim Orford <tim@orford.org>                  |
* +----------------------------------------------------------------------+
* | This program is free software; you can redistribute it and/or modify |
* | it under the terms of the GNU General Public License version 3       |
* | as published by the Free Software Foundation.                        |
* +----------------------------------------------------------------------+
*
*/
#include <yaml.h>

#define end_map \
		if(!yaml_mapping_end_event_initialize(&event)) goto error; \
		if(!yaml_emitter_emit(&emitter, &event)) goto error;

#define map_open(A) \
		if(!yaml_scalar_event_initialize(&event, NULL, str_tag, (guchar*)A, -1, plain_implicit, 0, YAML_PLAIN_SCALAR_STYLE)) goto error; \
		if(!yaml_emitter_emit(&emitter, &event)) goto error; \
		if(!yaml_mapping_start_event_initialize(&event, NULL, map_tag, 1, YAML_BLOCK_MAPPING_STYLE)) goto error; \
		if(!yaml_emitter_emit(&emitter, &event)) goto error; \

#define end_document \
	yaml_document_end_event_initialize(&event, 0); \
	yaml_emitter_emit(&emitter, &event); \
	yaml_stream_end_event_initialize(&event); \
	if(!yaml_emitter_emit(&emitter, &event)) goto error;

gboolean yaml_add_key_value_pair       (const char* key, const char*);
gboolean yaml_add_key_value_pair_int   (const char* key, int);
gboolean yaml_add_key_value_pair_float (const char* key, float);
gboolean yaml_add_key_value_pair_bool  (const char* key, gboolean);

typedef void   (*YamlCallback)         (yaml_parser_t*, const char* key, gpointer);
typedef void   (*YamlValueCallback)    (yaml_event_t*, gpointer);

typedef struct _yh
{
	char*        key;
	YamlCallback callback;
	gpointer     data;
} YamlHandler;

typedef struct _yvh
{
	char*             key;
	YamlValueCallback callback;
	gpointer          value;
} YamlValueHandler;

gboolean yaml_load                     (const char* filename, YamlHandler handlers[]);
gboolean yaml_load_section             (yaml_parser_t*, YamlHandler handlers[], YamlValueHandler vhandlers[]);
bool     yaml_load_section2            (yaml_parser_t*, YamlHandler**, YamlValueHandler**);
void     yaml_set_string               (yaml_event_t*, gpointer);
void     yaml_set_int                  (yaml_event_t*, gpointer);
void     yaml_set_uint64               (yaml_event_t*, gpointer);
void     yaml_set_position             (yaml_event_t*, gpointer);


